﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***   System.Security.Cryptography.DES                           ***//
//***     Authors:                                                 ***//
//***     Sergey Chaban (serge@wildwestsoftware.com)               ***//
//***     Sebastien Pouliot (spouliot@motus.com)                   ***//
//*** Get get the original SymmetricAlgorithm class, please visit  ***//
//*** the mono project at http://www.go-mono.com.                  ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** DES.cs:                                                      ***//
//*** 	Class for encrypting and decrypting using the DES          ***//
//***   cipher.                                                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    01.06.2003: Daniel Fabian - To fit SharpPrivacy's needs   ***//
//***    31.10.2013  James R. Stembridge - To fit the needs of     ***//
//***                Hermes Crypto Library                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;

// References:
// a.	FIPS PUB 46-3: Data Encryption Standard
//	http://csrc.nist.gov/publications/fips/fips46-3/fips46-3.pdf

namespace Hermes.Crypto.Cipher {

public abstract class DES : SymmetricAlgorithm {

	private static int blockSizeByte = 8;

	public DES () {
		KeySizeValue = 64; 
		BlockSizeValue = 64; 
		FeedbackSizeValue = 64;

		LegalKeySizesValue = new KeySizes[1];
		LegalKeySizesValue[0] = new KeySizes(64, 64, 0);

		LegalBlockSizesValue = new KeySizes[1];
		LegalBlockSizesValue[0] = new KeySizes(64, 64, 0);
	}

	public static new DES Create () {
		return (DES)Activator.CreateInstance(Type.GetType("Hermes.Crypto.Cipher.DESCryptoServiceProvider"), null);
	}

	internal static ulong PackKey (byte [] key) {
		ulong res = 0;
		for (int i = 0, sh = 8*blockSizeByte; (sh = sh - 8) >= 0; i++) {
			res |= (ulong) key [i] << sh;
		}
		return res;
	}

	internal static byte [] UnpackKey (ulong key) {
		byte [] res = new byte [blockSizeByte];
		for (int i = 0, sh = 8*blockSizeByte; (sh = sh - 8) >= 0; i++) {
			res [i] = (byte) (key >> sh);
		}
		return res;
	}

	// Ek(Ek(m)) = m
	internal static ulong [] weakKeys = {
		0x0101010101010101, /* 0000000 0000000 */
		0xFEFEFEFEFEFEFEFE, /* FFFFFFF FFFFFFF */
		0x1F1F1F1FE0E0E0E0, /* 0000000 FFFFFFF */
		0xE0E0E0E01F1F1F1F  /* FFFFFFF 0000000 */
	};

	// Ek1(Ek2(m)) = m
	internal static ulong [] semiweakKeys = {
		0x01FE01FE01FE01FE, 0xFE01FE01FE01FE01,
		0x1FE01FE00EF10EF1, 0xE01FE01FF10EF10E,
		0x01E001E001F101F1, 0xE001E001F101F101,
		0x1FFE1FFE0EFE0EFE, 0xFE1FFE1FFE0EFE0E,
		0x011F011F010E010E, 0x1F011F010E010E01,
		0xE0FEE0FEF1FEF1FE, 0xFEE0FEE0FEF1FEF1
	};

	public static bool IsWeakKey (byte [] rgbKey) {
        var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
        if (rgbKey.Length == (blockSizeByte >> 3))
            //throw new System.Security.Cryptography.CryptographicException ("Wrong Key Length");
            throw new Exception(messageLoader.GetString("DESKeyLengthError"));
            //throw new Exception("DES: Wrong Key Length");

		ulong lk = PackKey (rgbKey);
		foreach (ulong wk in weakKeys) {
			if (lk == wk) return true;
		}
		return false;
	}

	public static bool IsSemiWeakKey (byte [] rgbKey) {
        var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
		if (rgbKey.Length == (blockSizeByte >> 3))
			//throw new System.Security.Cryptography.CryptographicException ("Wrong Key Length");
            throw new Exception(messageLoader.GetString("DESKeyLengthError"));
            //throw new Exception("DES: Wrong Key Length");

		ulong lk = PackKey (rgbKey);
		foreach (ulong swk in semiweakKeys) {
			if (lk == swk) return true;
		}
		return false;
	}

	public override byte[] Key {
        
		get { return base.Key; }
		set {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (value == null)
				throw new ArgumentNullException ();
			if (IsWeakKey (value) || IsSemiWeakKey (value))
				//throw new System.Security.Cryptography.CryptographicException ();
                throw new Exception(messageLoader.GetString("DESNotStrongKey"));
                //throw new Exception("DES: Not a strong key");
			base.Key = value;
		}
	}

} // DES

} // System.Security.Cryptography
