//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***   System.Security.Cryptography.DES                           ***//
//***                                                              ***//
//*** Get get the original SymmetricAlgorithm class, please visit  ***//
//*** the mono project at http://www.go-mono.com.                  ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** DSA.cs:                                                      ***//
//*** 	Class for encrypting and decrypting using the DSA          ***//
//***   cipher.                                                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    01.06.2003: Daniel Fabian - To fit SharpPrivacy's needs   ***//
//***    31.10.2013  James R. Stembridge - To fit the needs of     ***//
//***                Hermes Crypto Library                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Windows.Storage.Streams;
using Windows.Security.Cryptography;
using Windows.Security.Cryptography.Core;
using System.Runtime.InteropServices.WindowsRuntime;
using Hermes.Crypto.OpenPGP;
using Hermes.Crypto.Cipher.Math;

namespace Hermes.Crypto.Cipher {
	
	
	/// <summary>
	/// The DSA class is there for signing and verifying messages
	/// with the DSA signature standard.
	/// </summary>
	/// <remarks>No remarks</remarks>
	public class DSA : AsymmetricCipher {

		private struct DSA_Public_Key {
			public BigInteger p;	    // prime 
			public BigInteger q;	    // group order 
			public BigInteger g;	    // group generator 
			public BigInteger y;	    // g^x mod p 
		}
		
		private struct DSA_Secret_Key {
			public BigInteger p;	    // prime 
			public BigInteger q;	    // group order 
			public BigInteger g;	    // group generator 
			public BigInteger y;	    // g^x mod p 
			public BigInteger x;	    // secret exponent
		}
		
		private BigInteger[][] biGeneratedKey;
		public BigInteger[][] GetGeneratedKey() {
			return this.biGeneratedKey;
		}

		
		/// <summary>
		/// Creates a new DES secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.
		/// </summary>
		/// <remarks>
		/// Creates a new DSA secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.<br></br>
		/// The order of the public components is p, q, g, y
		/// The order of the secret components is x.
		/// </remarks>
		/// <param name="nbits">The size of the key in bits.</param>
		/// <returns> a new DSA secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.<br></br>
		/// The order of the public components is p, q, g, y
		/// The order of the secret components is x.</returns>
		public override BigInteger[][] Generate(int keyLength) {
			DSA_Secret_Key dskKey = new DSA_Secret_Key();
			dskKey = GenerateParams(keyLength);
			dskKey = GenerateKeyPair(dskKey);
			
			biGeneratedKey = new BigInteger[2][];
			
			biGeneratedKey[0] = new BigInteger[4];
			biGeneratedKey[0][0] = dskKey.p;
			biGeneratedKey[0][1] = dskKey.q;
			biGeneratedKey[0][2] = dskKey.g;
			biGeneratedKey[0][3] = dskKey.y;
			
			biGeneratedKey[1] = new BigInteger[1];
			biGeneratedKey[1][0] = dskKey.x;
			
			return biGeneratedKey;
		}
		
		/// <summary>
		/// Encryption is not supported for DSA. If you call this function,
		/// an Exception will be thrown.
		/// </summary>
		/// <remarks>
		/// Encryption is not supported for DSA. If you call this function,
		/// an Exception will be thrown.
		/// </remarks>
		public override BigInteger[] Encrypt(BigInteger biPlain, PublicKeyPacket pkpKey) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("DSACannotEncrypt"));
			//throw(new Exception("The DSA cipher cannot be used for encryption"));
		}
		
		/// <summary>
		/// Decryption is not supported for DSA. If you call this function,
		/// an Exception will be thrown.
		/// </summary>
		/// <remarks>
		/// Decryption is not supported for DSA. If you call this function,
		/// an Exception will be thrown.
		/// </remarks>
		public override BigInteger Decrypt(BigInteger[] biCipher, SecretKeyPacket spkKey, string strPassphrase) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("DSACannotEncrypt"));
			//throw(new Exception("The DSA cipher cannot be used for encryption/decryption"));
		}
		
		
		private BigInteger[,] ParseSecretKey(DSA_Secret_Key dskKey) {
			BigInteger[,] biReturn = new BigInteger[2,4];
			
			biReturn[0,0] = dskKey.p;
			biReturn[0,1] = dskKey.q;
			biReturn[0,2] = dskKey.g;
			biReturn[0,3] = dskKey.y;
			biReturn[1,0] = dskKey.x;
			
			return biReturn;
		}
		
		private bool CheckKey(DSA_Secret_Key dskKey) {
			// - p, q, g, x, y > 0
			if (dskKey.p <= 0 || dskKey.q <= 0 || dskKey.g <= 0 ||
			    dskKey.y <= 0 || dskKey.x <= 0) {
				return false;
			}
			
			// - p is odd, q is odd
			if (dskKey.p % 2 == 0 || dskKey.q % 2 == 0)
				return false;
			
			
			// - 2^159 < q < 2^160
			if (dskKey.q.bitCount() < 159)
				return false;
			
			// - 1 < g < p
			if (dskKey.g >= dskKey.p || dskKey.g <= 1) 
				return false;
			
			// - 1 < y < p
			if (dskKey.y >= dskKey.p || dskKey.y <= 1) 
				return false;
			
			// - x < q
			if (dskKey.x >= dskKey.q)
				return false;
			
			// - g^q mod p = 1
			if (dskKey.g.modPow(dskKey.q, dskKey.p) != 1)
				return false;
			
			// - g^x mod p = y
			if (dskKey.g.modPow(dskKey.x, dskKey.p) != dskKey.y)
				return false;
			
			return true;
			
		}
		
		private DSA_Secret_Key ParseSecretKey(SecretKeyPacket skpKey, string strPassphrase) {
			DSA_Secret_Key dskKey = new DSA_Secret_Key();
			
			dskKey.p = skpKey.PublicKey.KeyMaterial[0];
			dskKey.q = skpKey.PublicKey.KeyMaterial[1];
			dskKey.g = skpKey.PublicKey.KeyMaterial[2];
			dskKey.y = skpKey.PublicKey.KeyMaterial[3];

			BigInteger[] biSecretKeyMaterial = skpKey.GetDecryptedKeyMaterial(strPassphrase);
			
			dskKey.x = biSecretKeyMaterial[0];
			
			return dskKey;
			
		}
		
		private DSA_Public_Key ParsePublicKey(PublicKeyPacket pkpKey) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			DSA_Public_Key dpkKey = new DSA_Public_Key();

            if (pkpKey.Algorithm != AsymAlgorithms.DSA)
                throw new Exception(messageLoader.GetString("KeyNotForUseWithDSA"));
				//throw(new System.ArgumentException("The given key is not supposed to be used with DSA!"));
			
			if (pkpKey.KeyMaterial.Length != 4)
                throw new Exception(messageLoader.GetString("KeyNotValidWithDSA"));
                //throw(new System.ArgumentException("The given key is not a valid key for DSA!"));
			
			dpkKey.p = pkpKey.KeyMaterial[0];
			dpkKey.q = pkpKey.KeyMaterial[1];
			dpkKey.g = pkpKey.KeyMaterial[2];
			dpkKey.y = pkpKey.KeyMaterial[3];
			
			return dpkKey;
		}
		
		private BigInteger[] ParsePublicKey(DSA_Public_Key dpkKey) {
			BigInteger[] biReturn = new BigInteger[4];
			
			biReturn[0] = dpkKey.p;
			biReturn[1] = dpkKey.q;
			biReturn[2] = dpkKey.g;
			biReturn[3] = dpkKey.y;
			
			return biReturn;
		}
	
		// this part is quite fast
		private DSA_Secret_Key GenerateKeyPair(DSA_Secret_Key dskKey) {
			dskKey.x = new BigInteger();
			do {
				// size of x (private key) isn't affected by the keysize (512-1024)
				dskKey.x = BigInteger.genRandom(160);
				BigInteger xx = new BigInteger();
			} while ((dskKey.x == 0) || (dskKey.x >= dskKey.q));
	
			// calculate the public key y = g^x % p
			dskKey.y = dskKey.g.modPow(dskKey.x, dskKey.p);
			
			return dskKey;
		}
	
		private DSA_Secret_Key GenerateParams(int keyLength) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			byte[] seed = new byte[20];
			byte[] part1 = new byte[20];
			byte[] part2 = new byte[20];
			byte[] u = new byte[20];
			//RandomNumberGenerator rng = RandomNumberGenerator.Create();
            IBuffer buffMsg;
            IBuffer buffHash;
			
			BigInteger p = new BigInteger();	    // prime 
			BigInteger q = new BigInteger();	    // group order 
			BigInteger g;	    // group generator 
			DSA_Secret_Key dskKey = new DSA_Secret_Key();
			
			//SHA1 sha = SHA1.Create();

            string hashAlg = HashAlgorithmNames.Sha1;
            // Create a HashAlgorithmProvider object.
            HashAlgorithmProvider objAlgProv = HashAlgorithmProvider.OpenAlgorithm(hashAlg);

			int n = (keyLength - 1) / 160;
			byte[] w = new byte [keyLength / 8];
			bool primesFound = false;
			
			while (!primesFound) {
				do {
					//rng.GetBytes(seed);                //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                    UInt32 seedLen = 20;
                    // Generate random data and copy it to a buffer.
                    IBuffer buffer = CryptographicBuffer.GenerateRandom(seedLen);
                    seed = buffer.ToArray();

					//part1 = sha.ComputeHash(seed);

                    buffMsg = seed.AsBuffer();
                    // Hash the message.
                    buffHash = objAlgProv.HashData(buffMsg);

                    // Verify that the hash length equals the length specified for the algorithm.
                    if (buffHash.Length != objAlgProv.HashLength)
                    {
                        throw new Exception(messageLoader.GetString("DSAErrorCreatingHash"));
                        //throw new Exception("There was an error creating the hash");
                    }

                    // Convert the IBuffer (hash) to a Byte Array
                    part1 = buffHash.ToArray();
                    


					Array.Copy(seed, 0, part2, 0, seed.Length);
					
					add(part2, seed, 1);
					
					//part2 = sha.ComputeHash(part2);

                    buffMsg = part2.AsBuffer();
                    // Hash the message.
                    buffHash = objAlgProv.HashData(buffMsg);

                    // Verify that the hash length equals the length specified for the algorithm.
                    if (buffHash.Length != objAlgProv.HashLength)
                    {
                        throw new Exception(messageLoader.GetString("DSAErrorCreatingHash"));
                        //throw new Exception("There was an error creating the hash");
                    }

                    // Convert the IBuffer (hash) to a Byte Array
                    part2 = buffHash.ToArray();





					
					for (int i = 0; i != u.Length; i++) 
						u[i] = (byte)(part1[i] ^ part2[i]);
					
					// first bit must be set (to respect key length)
					u[0] |= (byte)0x80;
					// last bit must be set (prime are all odds - except 2)
					u[19] |= (byte)0x01;
					
					q = new BigInteger(u);
				} while (!q.isProbablePrime());
				
				int counter = 0;
				int offset = 2;
				while (counter < 4096) {
					for (int k = 0; k < n; k++) {
						add(part1, seed, offset + k);

						//part1 = sha.ComputeHash(part1);


                        buffMsg = part1.AsBuffer();
                        // Hash the message.
                        buffHash = objAlgProv.HashData(buffMsg);

                        // Verify that the hash length equals the length specified for the algorithm.
                        if (buffHash.Length != objAlgProv.HashLength)
                        {
                            throw new Exception(messageLoader.GetString("DSAErrorCreatingHash"));
                            //throw new Exception("There was an error creating the hash");
                        }

                        // Convert the IBuffer (hash) to a Byte Array
                        part1 = buffHash.ToArray();


						Array.Copy(part1, 0, w, w.Length - (k + 1) * part1.Length, part1.Length);
					}
	
					add(part1, seed, offset + n);

					//part1 = sha.ComputeHash(part1);

                    buffMsg = part1.AsBuffer();
                    // Hash the message.
                    buffHash = objAlgProv.HashData(buffMsg);

                    // Verify that the hash length equals the length specified for the algorithm.
                    if (buffHash.Length != objAlgProv.HashLength)
                    {
                        throw new Exception(messageLoader.GetString("DSAErrorCreatingHash"));
                        //throw new Exception("There was an error creating the hash");
                    }

                    // Convert the IBuffer (hash) to a Byte Array
                    part1 = buffHash.ToArray();





					Array.Copy(part1, part1.Length - ((w.Length - (n) * part1.Length)), w, 0, w.Length - n * part1.Length);
	
					w[0] |= (byte)0x80;
					BigInteger xx = new BigInteger (w);
	
					BigInteger c = xx % (q * 2);
	
					p = xx - (c - 1);
	
					if (p.testBit((uint)(keyLength - 1))) {
						if (p.isProbablePrime()) {
							primesFound = true;
							break;
						}
					}
	
					counter += 1;
					offset += n + 1;
				}
				
			}
		
			// calculate the generator g
			BigInteger pMinusOneOverQ = (p - 1) / q;
			for (;;) {
				BigInteger h = new BigInteger();
				h = BigInteger.genRandom(keyLength);
				if ((h <= 1) || (h >= (p - 1)))
					continue;
	
				g = h.modPow(pMinusOneOverQ, p);
				if (g <= 1)
					continue;
				break;
			}
			
		    dskKey.p = p;
		    dskKey.q = q;
		    dskKey.g = g;
			
			return dskKey;
			
		}
		
		private void add(byte[] a, byte[] b, int value) {
			uint x = (uint)((b [b.Length - 1] & 0xff) + value);
	
			a[b.Length - 1] = (byte)x;
			x >>= 8;
	
			for (int i = b.Length - 2; i >= 0; i--) {
				x += (uint)(b[i] & 0xff);
				a[i] = (byte)x;
				x >>= 8;
			}
		}
		
		/// <summary>
		/// Public key operation. Verifies biSignature with the keydata
		/// in the given public key packet and returns true if the signature
		/// is valid.
		/// </summary>
		/// <param name="biSignature">The signature that is about to
		/// be verified</param>
		/// <param name="biHash">The hash value of the signed message.</param>
		/// <param name="pkpKey">The public key packet with the key
		/// material for the verification.</param>
		/// <returns>True if the signature is valid, otherwise 
		/// false</returns>
		/// <remarks>No remarks</remarks>
		public override bool Verify(BigInteger[] biSignature, BigInteger biHash, PublicKeyPacket pkpKey) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (biSignature == null)
				throw new ArgumentNullException("rgbSignature");
			
			DSA_Public_Key dpkKey = new DSA_Public_Key();
			dpkKey = ParsePublicKey(pkpKey);
			
			try {
				BigInteger m = biHash;
				BigInteger r = biSignature[0];
				BigInteger s = biSignature[1];
	
				if ((r < 0) || (dpkKey.q <= r))
					return false;
	
				if ((s < 0) || (dpkKey.q <= s))
					return false;
	
				BigInteger w = s.modInverse(dpkKey.q);
				BigInteger u1 = m * w % dpkKey.q;
				BigInteger u2 = r * w % dpkKey.q;
	
				u1 = dpkKey.g.modPow(u1, dpkKey.p);
				u2 = dpkKey.y.modPow(u2, dpkKey.p);
	
				BigInteger v = ((u1 * u2 % dpkKey.p) % dpkKey.q);
				return (v == r);
			} catch {
				//throw new CryptographicException();
                throw new Exception(messageLoader.GetString("DSAErrorValidatingSignature"));
                //throw new Exception("Error validating signature");
			}
		}
		
		/// <summary>
		/// Secret key operation. Signs biHash with the keydata
		/// in the given secret key packet.
		/// </summary>
		/// <param name="biHash">The hash value of a message that is about to
		/// be signed</param>
		/// <param name="skpKey">The secret key packet with the key
		/// material for the signature</param>
		/// <param name="strPassphrase">The passphrase for the 
		/// keymaterial</param>
		/// <returns>The signed hash as array of biginteger. Only return[0]
		/// contains a value: the signed hash.</returns>
		/// <remarks>No remarks</remarks>
		public override BigInteger[] Sign(BigInteger biHash, SecretKeyPacket skpKey, string strPassphase) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			DSA_Secret_Key dskKey = new DSA_Secret_Key();
			
			dskKey = ParseSecretKey(skpKey, strPassphase);
			
			//check if the key has been mangled with
            if (!CheckKey(dskKey))
                throw new Exception(messageLoader.GetString("DSABadKey"));
				//throw(new Exception("This key does not fullfill the requirements of a valid DSA key. Please check if someone messed with your keys!"));
			
			//if (biHash == null)
			//	throw new ArgumentNullException();
			
			// (a) Select a random secret integer k; 0 < k < q.
			BigInteger k = new BigInteger();
			k = BigInteger.genRandom(160);
			while (k >= dskKey.q)
				k = BigInteger.genRandom(160);
			
			// (b) Compute r = ( k mod p) mod q
			BigInteger r = (dskKey.g.modPow (k, dskKey.p)) % dskKey.q;
			// (c) Compute k -1 mod q (e.g., using Algorithm 2.142).
			// (d) Compute s = k -1 fh(m) +arg mod q.
			BigInteger s = (k.modInverse (dskKey.q) * (biHash + dskKey.x * r)) % dskKey.q;
			
			BigInteger[] biReturn = new BigInteger[2];
			
			biReturn[0] = r;
			biReturn[1] = s;
			return biReturn;
		}
		
	
		
		
	}
}
		
