//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** ElGamal.cs:                                                  ***//
//*** 	Class for encrypting and decrypting using the El Gamal     ***//
//***   cipher.                                                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    11.01.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header informaionation  ***//
//***    14.06.2003: Daniel Fabian - Changed Namespace from        ***//
//***                SharpPrivacy.Cipher to                        ***//
//***                SharpPrivacy.SharpPrivacyLib.Cipher           ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace from  ***//
//***                SharpPrivacy.SharpPrivacyLib.Cipher to        ***//
//***                Hermes.Crypto.Cipher.                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.OpenPGP;
using Hermes.Crypto.Cipher.Math;

namespace Hermes.Crypto.Cipher {
	
	
	/// <summary>
	/// This class is used for creating ElGamal keys, encrypting and decrypting,
	/// and signing and verifying with the ElGamal algorithm.
	/// </summary>
	/// <remarks>
	/// This class is used for creating ElGamal keys, encrypting and decrypting,
	/// and signing and verifying with the ElGamal algorithm.
	/// </remarks>
	public class ElGamal : AsymmetricCipher {

		private struct EG_Public_Key {
			public BigInteger p;	    /* prime */
			public BigInteger g;	    /* group generator */
			public BigInteger y;	    /* g^x mod p */
		}
		
		private struct EG_Secret_Key {
			public BigInteger p;	    /* prime */
			public BigInteger g;	    /* group generator */
			public BigInteger y;	    /* g^x mod p */
			public BigInteger x;	    /* secret exponent */
		}
		
		private BigInteger[][] biGeneratedKey;
		public BigInteger[][] GetGeneratedKey() {
			return this.biGeneratedKey;
		}
		
		
		/// <summary>
		/// Creates a new ElGamal secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.
		/// </summary>
		/// <remarks>
		/// Creates a new ElGamal secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.<br></br>
		/// The order of the public components is p, g, y
		/// The order of the secret components is x.
		/// </remarks>
		/// <param name="nbits">The size of the key in bits.</param>
		/// <returns>A new ElGamal secret key and returns it as a
		/// 2 dimensional array of biginteger. return[0] holds
		/// the public values of the key and return[1] all the
		/// secret values.<br></br>
		/// The order of the public components is p, g, y
		/// The order of the secret components is x.
		/// </returns>
		public override BigInteger[][] Generate(int nBits) {
			EG_Secret_Key eskKey = GenerateKey(nBits);
			
			biGeneratedKey = new BigInteger[2][];
			biGeneratedKey[0] = new BigInteger[3];
			biGeneratedKey[0][0] = eskKey.p;
			biGeneratedKey[0][1] = eskKey.g;
			biGeneratedKey[0][2] = eskKey.y;
			
			biGeneratedKey[1] = new BigInteger[1];
			biGeneratedKey[1][0] = eskKey.x;
			
			return biGeneratedKey;
		}
		
		private EG_Secret_Key GenerateKey(int nBits){
			BigInteger q = new BigInteger();
			BigInteger p;
			BigInteger g;
			BigInteger gPowTwo;
			BigInteger gPowQ;
			EG_Secret_Key eskKey = new EG_Secret_Key();
			
			/*
			// construct a prime p = 2q + 1
			do {
				q = BigInteger.genRandom(nBits - 1);
				System.Windows.Forms.Application.DoEvents();
				p = (2*q) + 1;
			} while ((!p.isProbablePrime()) || (!q.isProbablePrime()));
			*/
			
			q = BigInteger.genPseudoPrime(nBits - 1);
			p = BigInteger.genPseudoPrime(nBits);
			
			// find a generator
			do {
				g = new BigInteger();
				g = BigInteger.genRandom(nBits - 1);
				gPowTwo = g.modPow(new BigInteger(2), p);
				gPowQ = g.modPow(q, p);
			} while ((gPowTwo == 1) || (gPowQ == 1));
			
			BigInteger x;
			
			do {
				x = new BigInteger();
				x = BigInteger.genRandom(nBits);
			} while (x >= p-1);
			
			BigInteger y = g.modPow(x, p);
			
			eskKey.p = p;
			eskKey.g = g;
			eskKey.x = x;
			eskKey.y = y;
			
			return eskKey;
			
		}
		
		/// <summary>
		/// Secret key operation. Decrypts biCipher with the keydata
		/// in the given secret key packet.
		/// </summary>
		/// <param name="biInput">The ciphertext that is about to
		/// be decrypted</param>
		/// <param name="skpKey">The secret key packet with the key
		/// material for the decryption</param>
		/// <param name="strPassphrase">The passphrase for the 
		/// keymaterial</param>
		/// <returns>The decrypted ciphertext.</returns>
		/// <remarks>No remarks.</remarks>
		public override BigInteger Decrypt(BigInteger[] biInput, SecretKeyPacket skpKey, string strPassphrase) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			BigInteger[] biKeyMaterial = skpKey.GetDecryptedKeyMaterial(strPassphrase);
			EG_Secret_Key eskKey = new EG_Secret_Key();
			eskKey.x = biKeyMaterial[0];
			eskKey.p = skpKey.PublicKey.KeyMaterial[0];
			eskKey.g = skpKey.PublicKey.KeyMaterial[1];
			eskKey.y = skpKey.PublicKey.KeyMaterial[2];

            if (biInput.Length != 2)
                throw new Exception(messageLoader.GetString("NotElgamalEncrypted"));
				//throw new ArgumentException("biInput is not an ElGamal encrypted Packet");
			
			BigInteger B = biInput[0];
			BigInteger c = biInput[1];
			
			BigInteger z = B.modPow(eskKey.x, eskKey.p).modInverse(eskKey.p);
			
			BigInteger output = (z * c) % eskKey.p;
			
			return output;
			
		}
		
		/// <summary>
		/// Public key operation. Encrypts biInput with the keydata
		/// in the given public key packet.
		/// </summary>
		/// <param name="biInput">The plaintext that is about to
		/// be encrypted</param>
		/// <param name="pkpKey">The public key packet with the key
		/// material for the encryption</param>
		/// <returns>The encrypted ciphertext.</returns>
		/// <remarks>No remarks.</remarks>
		public override BigInteger[] Encrypt(BigInteger biInput, PublicKeyPacket pkpKey) {
			EG_Public_Key epkKey = new EG_Public_Key();
			epkKey.p = pkpKey.KeyMaterial[0];
			epkKey.g = pkpKey.KeyMaterial[1];
			epkKey.y = pkpKey.KeyMaterial[2];
			
			BigInteger k = new BigInteger();
			
			//Random number needed for encryption
			k = BigInteger.genRandom(epkKey.p.bitCount()-1);
			
			while (k > (epkKey.p-1)) {
				k = new BigInteger();
				k = BigInteger.genRandom(epkKey.p.bitCount()-1);
			}
			
			BigInteger B = epkKey.g.modPow(k, epkKey.p);
			BigInteger c = epkKey.y.modPow(k, epkKey.p);
			c = (biInput * c) % epkKey.p;
			//BigInteger c = (biInput * epkKey.y.modPow(k, epkKey.p)) % epkKey.p;
			
			BigInteger[] biOutput = new BigInteger[2];
			
			biOutput[0] = B;
			biOutput[1] = c;
			
			return biOutput;
			
		}
		
		public override BigInteger[] Sign(BigInteger biHash, SecretKeyPacket spkKey, string strPassphrase) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("ElGamalSigNotImplemented"));

			//throw new System.NotImplementedException("Signatures with ElGamal keys are not yet implemented!");
		}
		
		public override bool Verify(BigInteger[] biSignature, BigInteger biHash, PublicKeyPacket pkpKey) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("ElGamalSigNotImplemented"));
            //throw new System.NotImplementedException("Signatures with ElGamal keys are not yet implemented!");
		}
	}
	
}
