﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***   System.Security.Cryptography SymmetricAlgorithm            ***//
//***     Authors:                                                 ***//
//***     Thomas Neidhart (tome@sbox.tugraz.at)                    ***//
//***     Sebastien Pouliot (spouliot@motus.com)                   ***//
//*** Portions (C) 2002 Motus Technologies Inc.                    ***//
//*** (http://www.motus.com)                                       ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***                                                              ***//
//*** Get get the original SymmetricAlgorithm class, please visit  ***//
//*** the mono project at http://www.go-mono.com.                  ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** SymmetricAlgorithm.cs:                                       ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    01.06.2003: Daniel Fabian - To fit SharpPrivacy's needs   ***//
//***    31.10.2013  James R. Stembridge - To fit the needs of     ***//
//***                Hermes Crypto Library                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Windows.Storage.Streams;
using Windows.Security.Cryptography;
using Windows.Security.Cryptography.Core;
using System.Runtime.InteropServices.WindowsRuntime;

namespace Hermes.Crypto.Cipher {

	// This class implement most of the common code required for symmetric
	// algorithm transforms, like:
	// - CipherMode: Builds CBC and CFB on top of (descendant supplied) ECB
	// - PaddingMode, transform properties, multiple blocks, reuse...
	//
	// Descendants MUST:
	// - intialize themselves (like key expansion, ...)
	// - override the ECB (Electronic Code Book) method which will only be
	//   called using BlockSize byte[] array.
	internal abstract class SymmetricTransform : ICryptoTransform {
		protected SymmetricAlgorithm algo;
		protected bool encrypt;
		private int BlockSizeByte;
		private byte[] temp;
        private byte[] holdBuffer;
		private byte[] temp2;
		private byte[] workBuff;
		private byte[] workout;
		private byte[] openPGPStart; // needed for openpgp's CFB
		private int FeedBackByte;
		private int FeedBackIter;
		private bool m_disposed = false;
        

		public SymmetricTransform (SymmetricAlgorithm symmAlgo, bool encryption, byte[] rgbIV) {
			algo = symmAlgo;
			encrypt = encryption;
			BlockSizeByte = (algo.BlockSize >> 3);
			// mode buffers
			temp = new byte [BlockSizeByte];
			Array.Copy (rgbIV, 0, temp, 0, BlockSizeByte);
			temp2 = new byte [BlockSizeByte];
			FeedBackByte = (algo.FeedbackSize >> 3);
			FeedBackIter = (int) BlockSizeByte / FeedBackByte;
			//FeedBackIter = 1;
			// transform buffers
			workBuff = new byte [BlockSizeByte];
			workout =  new byte [BlockSizeByte];
			
			//OpenPGP CFB needs an all zero IV
			if (algo.Mode == CipherMode.OpenPGP_CFB) {
				algo.IV = new byte[algo.BlockSize >> 3];
			}
			
			// needed for openpgp's ciphertext feedback mode
			openPGPStart = new byte[(algo.BlockSize >> 3) + 2];
		}
		
		~SymmetricTransform () {
			Dispose (false);
		}

		void IDisposable.Dispose () {
			Dispose (true);
			GC.SuppressFinalize (this);  // Finalization is now unnecessary
		}

		// MUST be overriden by classes using unmanaged ressources
		// the override method must call the base class
		protected void Dispose (bool disposing) {
			if (!m_disposed) {
				if (disposing) {
					// dispose managed object: zeroize and free
					Array.Clear (temp, 0, BlockSizeByte);
					temp = null;
					Array.Clear (temp2, 0, BlockSizeByte);
					temp2 = null;
				}
				m_disposed = true;
			}
		}

		public virtual bool CanTransformMultipleBlocks {
			get { return true; }
		}

		public bool CanReuseTransform {
			get { return false; }
		}

		public virtual int InputBlockSize {
			get { return BlockSizeByte; }
		}

		public virtual int OutputBlockSize {
			get { return BlockSizeByte; }
		}

		// note: Each block MUST be BlockSizeValue in size!!!
		// i.e. Any padding must be done before calling this method
        protected void Transform(byte[] input, byte[] output)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            switch (algo.Mode)
            {
                case CipherMode.ECB:
                    ECB(input, output);
                    break;
                case CipherMode.CBC:
                    CBC(input, output);
                    break;
                case CipherMode.CFB:
                    CFB(input, output);
                    break;
                case CipherMode.OFB:
                    OFB(input, output);
                    break;
                case CipherMode.CTS:
                    CTS(input, output);
                    break;
                case CipherMode.OpenPGP_CFB:
                    CFB(input, output);
                    break;
                default:
                    string msg = messageLoader.GetString("UnknownCipherMode") + algo.Mode.ToString();
                    throw new Exception(msg);
                //throw new NotImplementedException("Unkown CipherMode" + algo.Mode.ToString ());
            }
        }

		// Electronic Code Book (ECB)
		protected abstract void ECB (byte[] input, byte[] output); 

		// Cipher-Block-Chaining (CBC)
		protected virtual void CBC (byte[] input, byte[] output) {
			if (encrypt) {
				for (int i = 0; i < BlockSizeByte; i++)
					temp[i] ^= input[i];
				ECB (temp, output);
				Array.Copy (output, 0, temp, 0, BlockSizeByte);
			}
			else {
				Array.Copy (input, 0, temp2, 0, BlockSizeByte);
				ECB (input, output);
				for (int i = 0; i < BlockSizeByte; i++)
					output[i] ^= temp[i];
				Array.Copy (temp2, 0, temp, 0, BlockSizeByte);
			}
		}
		
		// Cipher-FeedBack (CFB)
		protected virtual void CFB (byte[] input, byte[] output) {
            if (encrypt)
            {
                for (int x = 0; x < FeedBackIter; x++)
                {
                    // temp is first initialized with the IV
                    ECB(temp, temp2);
                    for (int i = 0; i < FeedBackByte; i++)
                        output[i + x] = (byte)(temp2[i] ^ input[i + x]);
                    Array.Copy(temp, FeedBackByte, temp, 0, BlockSizeByte - FeedBackByte);
                    Array.Copy(output, x, temp, BlockSizeByte - FeedBackByte, FeedBackByte);
                }
            }
            else
            {
                for (int x = 0; x < FeedBackIter; x++)
                {
                    // we do not really decrypt this data!
                    encrypt = true;
                    // temp is first initialized with the IV
                    ECB(temp, temp2);
                    encrypt = false;

                    for (int i = 0; i < FeedBackByte; i++)
                        output[i + x] = (byte)(temp2[i] ^ input[i + x]);

                    Array.Copy(temp, FeedBackByte, temp, 0, BlockSizeByte - FeedBackByte);
                    Array.Copy(input, x, temp, BlockSizeByte - FeedBackByte, FeedBackByte);
                }
            }
		}
		
		// Output-FeedBack (OFB)
        protected virtual void OFB(byte[] input, byte[] output)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("OFBNotYetSupported"));
            //throw new NotImplementedException ("OFB not yet supported");
        }

		// Cipher Text Stealing (CTS)
		protected virtual void CTS(byte[] input, byte[] output) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            throw new Exception(messageLoader.GetString("CTSNotYetSupported"));
            //throw new NotImplementedException ("CTS not yet supported");
		}

		// this method may get called MANY times so this is the one to optimize
		public virtual int TransformBlock(byte[] inputBuffer, int inputOffset, int inputCount, ref byte[] outputBuffer, int outputOffset) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            int originalInputCount = inputCount;

            if (m_disposed)
                throw new Exception(messageLoader.GetString("SymAlgoObjectDisposed"));
                //throw new ObjectDisposedException("Object is disposed");

			if (outputOffset + inputCount > outputBuffer.Length)
                throw new Exception(messageLoader.GetString("SymAlgoInsufficientOutputBuffer"));
				//throw new System.Security.Cryptography.CryptographicException("Insufficient output buffer size.");
                //throw new System.Exception("SymmetricAlgorithm: Insufficient output buffer size.");
            
			int offs = inputOffset;
			int full;



			// this way we don't do a modulo every time we're called
			// and we may save a division
            if (inputCount != BlockSizeByte)
            {

                // fill the last block with 0's. this doesn't matter for CFB
                if (((inputCount % BlockSizeByte) != 0) && ((algo.Mode == CipherMode.CFB) || (algo.Mode == CipherMode.OpenPGP_CFB)))
                {
                    byte[] oldInputBuffer = new byte[inputCount];
                    Array.Copy(inputBuffer, oldInputBuffer, inputCount);
                    if (algo.Mode == CipherMode.CFB)
                        inputBuffer = new byte[inputCount + (BlockSizeByte - (inputCount % BlockSizeByte))];
                    else
                        inputBuffer = new byte[inputCount + (BlockSizeByte - (inputCount % BlockSizeByte)) + 2];
                    outputBuffer = new byte[inputBuffer.Length];
                    Array.Copy(oldInputBuffer, 0, inputBuffer, 0, inputCount);
                    inputCount = inputBuffer.Length;
                }
                else if ((inputCount % BlockSizeByte) != 0)
                {
                    //throw new System.Security.Cryptography.CryptographicException("Invalid input block size.");
                    throw new Exception(messageLoader.GetString("SymAlgoInvalidInputBlocksize"));
                    //throw new System.Exception("SymmetricAlgorithm: Invalid input block size.");
                }

            }
            else
                full = 1;
			
			// OpenPGP needs some special treatment
            if (offs == 0 && algo.Mode == CipherMode.OpenPGP_CFB && !algo.IntegrityProtected)
            {
				if (encrypt) {
					// at first we have an encrypted random block
					byte[] tmpInput = new byte[BlockSizeByte];
					byte[] tmp2Input = new byte[BlockSizeByte];
					outputBuffer = new byte[outputBuffer.Length + BlockSizeByte + 2];

					//System.Security.Cryptography.RandomNumberGenerator rngIntro = System.Security.Cryptography.RandomNumberGenerator.Create();
					//rngIntro.GetBytes(tmpInput);

                    //rng.GetBytes (key);
                    UInt32 targetLen = (UInt32)tmpInput.Length;
                    // Generate random data and copy it to a buffer.
                    IBuffer buffer = CryptographicBuffer.GenerateRandom(targetLen);
                    tmpInput = buffer.ToArray();

					tmp2Input[0] = tmpInput[tmpInput.Length - 2];
					tmp2Input[1] = tmpInput[tmpInput.Length - 1];
					Transform(tmpInput, workout);
					Array.Copy(workout, 0, outputBuffer, 0, BlockSizeByte);
					outputOffset += BlockSizeByte;
                    
					byte[] tmpWorkout = new byte[BlockSizeByte];
					
					Transform(tmp2Input, tmpWorkout);
					Array.Copy(tmpWorkout, 0, outputBuffer, outputOffset, 2);
					outputOffset += 2;
					
					// Load feedback-register with c3 - c10
                    if (!algo.IntegrityProtected)
                        Array.Copy(outputBuffer, 2, temp, 0, BlockSizeByte);
                    
				} else {
					Array.Copy(inputBuffer, offs, workBuff, 0, BlockSizeByte);
					Transform(workBuff, workout);
					Array.Copy(workout, 0, openPGPStart, 0, BlockSizeByte);
					offs += BlockSizeByte;
                   

					byte[] tmpInput = new byte[BlockSizeByte];
					tmpInput.Initialize();
					Array.Copy(inputBuffer, offs, tmpInput, 0, 2);
					Transform(tmpInput, workout);
                   
					Array.Copy(workout, 0, openPGPStart, BlockSizeByte, 2);
					int x = openPGPStart.Length;
					if ((openPGPStart[x-1] != openPGPStart[x-3]) || (openPGPStart[x-2] != openPGPStart[x-4]))
						//throw new System.Security.Cryptography.CryptographicException("Wrong Key!!");
                        throw new Exception(messageLoader.GetString("SymAlgoWrongKey"));
                        //throw new System.Exception("SymmetricAlgorithm: Wrong Key!!");
					
					// Load feedback-register with c3 - c[BS+2]
                    if (!algo.IntegrityProtected)
                        Array.Copy(inputBuffer, 2, temp, 0, BlockSizeByte);
                    
					offs += 2;
					
				}
				
			}
            else if (offs == 0 && algo.Mode == CipherMode.OpenPGP_CFB && encrypt)
            {
                // First, we create a block of random data for the OpenPGP_CFB IV
                byte[] tmpInput = new byte[BlockSizeByte];
                //System.Security.Cryptography.RandomNumberGenerator rngIntro = System.Security.Cryptography.RandomNumberGenerator.Create();
                //rngIntro.GetBytes(tmpInput);
                //rng.GetBytes (key);
                UInt32 targetLen = (UInt32)tmpInput.Length;
                IBuffer buffer = CryptographicBuffer.GenerateRandom(targetLen);
                tmpInput = buffer.ToArray();

                // Create a work buffer large enough to handle the random data 
                // plus the two repeat bytes plus the input data 
                byte[] wkBuffer = new byte[tmpInput.Length + inputCount + 2];
                // Copy the random data to the work buffer
                Array.Copy(tmpInput, 0, wkBuffer, 0, tmpInput.Length);
                // Copy the two repeat bytes to the next two positions of the work buffer
                Array.Copy(tmpInput, tmpInput.Length - 2, wkBuffer, tmpInput.Length, 2);
                // Copy the input data to the work data following the random data and repeat bytes
                Array.Copy(inputBuffer, 0, wkBuffer, tmpInput.Length + 2, inputCount);
                

                // Determine the actual length of data in the input buffer
                // NOTE: I am not sure if this is needed or not.  I put it
                // here so that I can play with the data that needs to be 
                // passed to the MDC calculation routine.  As soon as I 
                // can determine how GnuGPG handles the calculation I'll
                // correct this.
                int x = 0;
                int actDataLen = 0;
                bool keepGoing = true;
                for (actDataLen = wkBuffer.Length; keepGoing; actDataLen--)
                {
                    if (wkBuffer[actDataLen - 1] != 0)
                        keepGoing = false;
                    if (actDataLen < 5)
                        keepGoing = false;
                }
                actDataLen++;
                
                x = actDataLen;
                x = originalInputCount;
                int copyLen = originalInputCount + BlockSizeByte + 2;
                int testLen = actDataLen ;


                //copyLen = testLen;

                // Here we are going to calculate the MDC hash and
                // build the MDC packet.
                byte[] hData;
                byte[] prefix = new byte[2];                            // Create a buffer for the MDC 
                prefix[0] = 0xD3;                                       // prefix and load the prefix
                prefix[1] = 0x14;
                if (copyLen == originalInputCount + BlockSizeByte + 2)
                {
                    hData = new byte[copyLen + 2];                   // Create a hash buffer
                    Array.Copy(wkBuffer, 0, hData, 0, copyLen);             // Copy the data to it
                                                           // data to it.
                    Array.Copy(prefix, 0, hData, copyLen, prefix.Length);   // Copy prefix data to hash buffer
                }
                else
                {
                    // scenario -> Random, Headerbyte[0], cData, prefix
                    int hDataLength = actDataLen - 2 +3;
                    hData = new byte[hDataLength];
                    Array.Copy(wkBuffer, 0, hData, 0, BlockSizeByte+2);
                    byte wkByte = inputBuffer[0];
                    hData[BlockSizeByte + 2] = wkByte;
                    Array.Copy(inputBuffer, 3, hData, BlockSizeByte + 2+1, originalInputCount-3);
                    Array.Copy(prefix, 0, hData, hData.Length-2, prefix.Length);
                }



                string hashAlg = HashAlgorithmNames.Sha1;               // set hash algo to SHA1
                // Create a HashAlgorithmProvider object.
                HashAlgorithmProvider objAlgProv = HashAlgorithmProvider.OpenAlgorithm(hashAlg);
                IBuffer buffMsg = hData.AsBuffer();                     // Load hash data to I buffer
                IBuffer buffHash = objAlgProv.HashData(buffMsg);        // Hash the data
                // Verify that the hash length equals the length specified for the algorithm.
                if (buffHash.Length != objAlgProv.HashLength)
                {
                    throw new Exception(messageLoader.GetString("SymAlgoErrorCreatingHash"));
                    //throw new Exception("SymmetricAlgorithm: Error creating the MDC hash");
                  
                }
                // Convert the IBuffer (hash) to a Byte Array
                byte[] myHash = buffHash.ToArray();
                byte[] MDCPacket = new byte[22];                        // create buffer for MDC packet 
                Array.Copy(prefix, 0, MDCPacket, 0, 2);                 // MDC packet header
                Array.Copy(myHash, 0, MDCPacket, 2, 20);                // MDC packet data



                copyLen = originalInputCount + BlockSizeByte + 2;       // for MDC calc testing JRS

                // Prepare the data to be encrypted. We encrypt the MDC packet too
                // Reallocate the input buffer so that it is large enough to
                // handle all of the data we have added. It should also be
                // a multiple of the blocksize.
                int buffSize = wkBuffer.Length + 22;                    // The +22 is for the MDC packet
                if ((buffSize % BlockSizeByte) != 0) 
                {
                    buffSize = buffSize + (BlockSizeByte - (buffSize % BlockSizeByte));
                }
                inputBuffer = new byte[buffSize];
                Array.Copy(wkBuffer, 0, inputBuffer, 0, wkBuffer.Length);   // Copy the data to it
                Array.Copy(MDCPacket, 0, inputBuffer, copyLen, 22);         // Copy the MDC packet to it
                // Reset the input count to allow for all of the new data
                inputCount = inputBuffer.Length;
                // Reallocate the output buffer so that it is large enough to handle all
                // of the data we added to the input buffer (random data and MDC packet)
                outputBuffer = new byte[buffSize];
            }
			
			full = (inputCount - offs) / BlockSizeByte;
			
			int total = 0;
			for (int i = 0; i < full; i++) {
				Array.Copy(inputBuffer, offs, workBuff, 0, BlockSizeByte);
                
				Transform(workBuff, workout);
				Array.Copy (workout, 0, outputBuffer, outputOffset, BlockSizeByte);
				offs += BlockSizeByte;
				outputOffset += BlockSizeByte;
				total += BlockSizeByte;
			}
			return total;
		}
		
		private byte[] FinalEncrypt(byte [] inputBuffer, int inputOffset, int inputCount) {
			// are there still full block to process ?
			int full = (inputCount / BlockSizeByte) * BlockSizeByte;
			int rem = inputCount - full;
			int total = full;

			// we need to add an extra block if...
			// a. the last block isn't complate (partial);
			// b. the last block is complete but we use padding
			if ((rem > 0) || (algo.Padding != PaddingMode.None))
				total += BlockSizeByte;
			byte[] res = new byte [total];

			// process all blocks except the last (final) block
			while (total > BlockSizeByte) {
				TransformBlock(inputBuffer, inputOffset, BlockSizeByte, ref res, inputOffset);
				inputOffset += BlockSizeByte;
				total -= BlockSizeByte;
			}

			// now we only have a single last block to encrypt
			int padding = BlockSizeByte - rem;
			switch (algo.Padding) {
				case PaddingMode.None:
					break;
				case PaddingMode.PKCS7:
					for (int i = BlockSizeByte; --i >= (BlockSizeByte - padding);) 
						res [i] = (byte) padding;
					break;
				case PaddingMode.Zeros:
					for (int i = BlockSizeByte; --i >= (BlockSizeByte - padding);)
						res [i] = 0;
					break;
			}
			Array.Copy (inputBuffer, inputOffset, res, full, rem);

			// the last padded block will be transformed in-place
			TransformBlock(res, full, BlockSizeByte, ref res, full);
			return res;
		}

		private byte[] FinalDecrypt(byte [] inputBuffer, int inputOffset, int inputCount) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if ((inputCount % BlockSizeByte) > 0)
				//throw new System.Security.Cryptography.CryptographicException ("Invalid input block size.");
                throw new Exception(messageLoader.GetString("SymAlgoInvalidInputBlocksize"));
                //throw new System.Exception("SymmetricAlgorithm: Invalid input block size.");

			int total = inputCount;
			byte[] res = new byte [total];
			while (inputCount > 0) {
				TransformBlock (inputBuffer, inputOffset, BlockSizeByte, ref res, inputOffset);
				inputOffset += BlockSizeByte;
				inputCount -= BlockSizeByte;
			}

			switch (algo.Padding) {
				case PaddingMode.None:
					break;
				case PaddingMode.PKCS7:
					total -= res [total - 1];
					break;
				case PaddingMode.Zeros:
					// TODO
					break;
			}

			// return output without padding
			byte[] data = new byte [total];
			Array.Copy (res, 0, data, 0, total);
			// zeroize decrypted data (copy with padding)
			Array.Clear (res, 0, res.Length);
			return data;
		}

		public virtual byte [] TransformFinalBlock (byte [] inputBuffer, int inputOffset, int inputCount) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (m_disposed)
                throw new Exception(messageLoader.GetString("SymAlgoObjectDisposed"));
                //throw new ObjectDisposedException ("Object is disposed");

			if (encrypt)
				return FinalEncrypt (inputBuffer, inputOffset, inputCount);
			else
				return FinalDecrypt (inputBuffer, inputOffset, inputCount);
		}
	}

	/// <summary>
	/// Abstract base class for all cryptographic symmetric algorithms.
	/// Available algorithms include:
	/// DES, RC2, Rijndael, TripleDES
	/// </summary>
	public abstract class SymmetricAlgorithm : IDisposable {
		protected int BlockSizeValue; // The block size of the cryptographic operation in bits. 
		protected int FeedbackSizeValue; // The feedback size of the cryptographic operation in bits. 
		protected byte[] IVValue; // The initialization vector ( IV) for the symmetric algorithm. 
		protected int KeySizeValue; // The size of the secret key used by the symmetric algorithm in bits. 
		protected byte[] KeyValue; // The secret key for the symmetric algorithm. 
		protected KeySizes[] LegalBlockSizesValue; // Specifies the block sizes that are supported by the symmetric algorithm. 
		protected KeySizes[] LegalKeySizesValue; // Specifies the key sizes that are supported by the symmetric algorithm. 
		protected CipherMode ModeValue; // Represents the cipher mode used in the symmetric algorithm. 
		protected PaddingMode PaddingValue; // Represents the padding mode used in the symmetric algorithm. 
		private bool m_disposed;
        private bool integrity = false;  // Denotes using Symmetrically Encrypted Integrity Protected Packet
        private bool _signed = false; // Indicates if this message is signed or not. It matters for the MDC Packet

		/// <summary>
		/// Called from constructor of derived class.
		/// </summary>
		public SymmetricAlgorithm () {
			ModeValue = CipherMode.CBC;
			PaddingValue = PaddingMode.PKCS7;
			m_disposed = false;
		}
		
		/// <summary>
		/// Called from constructor of derived class.
		/// </summary>
		~SymmetricAlgorithm () {
			Dispose (false);
		}

		public void Clear() {
			Dispose (true);
		}

		void IDisposable.Dispose () {
			Dispose (true);
			GC.SuppressFinalize (this);  // Finalization is now unnecessary
		}

		protected virtual void Dispose (bool disposing) {
			if (!m_disposed) {
				// always zeroize keys
				if (KeyValue != null) {
					// Zeroize the secret key and free
					Array.Clear (KeyValue, 0, KeyValue.Length);
					KeyValue = null;
				}
				// dispose unmanaged managed objects
				if (disposing) {
					// dispose managed objects
				}
				m_disposed = true;
			}
		}

		/// <summary>
		/// Gets or sets the actual BlockSize
		/// </summary>
		public virtual int BlockSize {
			get { return this.BlockSizeValue; }
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
                if (IsLegalKeySize(this.LegalBlockSizesValue, value))
                    this.BlockSizeValue = value;
                else
                    //throw new System.Security.Cryptography.CryptographicException("block size not supported by algorithm");
                    throw new Exception(messageLoader.GetString("SymAlgoBlocksizeUnsupported"));
                    //throw new System.Exception("SymmetricAlgorithm: block size not supported by algorithm");
			}
		}

		/// <summary>
		/// Gets or sets the actual FeedbackSize
		/// </summary>
		public virtual int FeedbackSize {
			get { return this.FeedbackSizeValue; }
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (value > this.BlockSizeValue)
                    throw new Exception(messageLoader.GetString("SymAlgoFeedbackSize"));
					//throw new System.Security.Cryptography.CryptographicException("feedback size larger than block size");
                    //throw new System.Exception("SymmetricAlgorithm: feedback size larger than block size");
				else
					this.FeedbackSizeValue = value;
			}
		}
		
		/// <summary>
		/// Gets or sets the actual Initial Vector
		/// </summary>
		public virtual byte[] IV {
			get {
				if (this.IVValue == null)
					GenerateIV();

				return this.IVValue;
			}
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (value == null)
                    throw new Exception(messageLoader.GetString("SymAlgoIVNull"));
                    //throw new ArgumentNullException("SymmetricAlgorithm: tried setting initial vector to null");
					
				if (value.Length * 8 != this.BlockSizeValue)
                    throw new Exception(messageLoader.GetString("SymAlgoIVLengthBlockSize"));
					//throw new System.Security.Cryptography.CryptographicException ("IV length must match block size");
                    //throw new System.Exception("SymmetricAlgorithm: IV length must match block size");
				
				this.IVValue = new byte[value.Length];
				Array.Copy (value, 0, this.IVValue, 0, value.Length);
			}
		}

		/// <summary>
		/// Gets or sets the actual key
		/// </summary>
		public virtual byte[] Key {
			get {
				if (this.KeyValue == null)
					GenerateKey();

				return this.KeyValue;
			}
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (value == null)
                    throw new Exception(messageLoader.GetString("SymAlgoKeyNull"));
                    //throw new ArgumentNullException ("tried setting key to null");

				if (!IsLegalKeySize (this.LegalKeySizesValue, value.Length * 8))
                    throw new Exception(messageLoader.GetString("SymAlgoKeysizeUnsupported"));
					//throw new System.Security.Cryptography.CryptographicException ("key size not supported by algorithm");
                    //throw new System.Exception("SymmetricAlgorithm: key size not supported by algorithm");

				this.KeySizeValue = value.Length * 8;
				this.KeyValue = new byte [value.Length];
				Array.Copy (value, 0, this.KeyValue, 0, value.Length);
			}
		}

        /// <summary>
        /// Gets or sets Symmetrically Encrypted Integrity Protected Data Packet Indicator
        /// </summary>
        public virtual bool IntegrityProtected
        {
            get
            {
                return this.integrity;
            }
            set
            {
                this.integrity = value;
            }
        }

        /// <summary>
        /// Gets or sets Message is Signed Indicator
        /// </summary>
        //public virtual bool Signed
        //{
        //    get
        //    {
        //        return this._signed;
        //    }
        //    set
         //   {
         //       this._signed = value;
          //  }
        //}
        
        /// <summary>
		/// Gets or sets the actual key size in bits
		/// </summary>
		public virtual int KeySize {
			get { return this.KeySizeValue; }
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (!IsLegalKeySize (this.LegalKeySizesValue, value))
                    throw new Exception(messageLoader.GetString("SymAlgoKeysizeUnsupported"));
					//throw new System.Security.Cryptography.CryptographicException ("key size not supported by algorithm");
                    //throw new System.Exception("SymmetricAlgorithm: key size not supported by algorithm");
				
				this.KeyValue = null;
				this.KeySizeValue = value;
			}
		}

		/// <summary>
		/// Gets all legal block sizes
		/// </summary>
		public virtual KeySizes[] LegalBlockSizes {
			get { return this.LegalBlockSizesValue; }
		}

		/// <summary>
		/// Gets all legal key sizes
		/// </summary>
		public virtual KeySizes[] LegalKeySizes {
			get { return this.LegalKeySizesValue; }
		}

		/// <summary>
		/// Gets or sets the actual cipher mode
		/// </summary>
		public virtual CipherMode Mode {
			get { 
				return this.ModeValue;
			}
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (Enum.IsDefined( ModeValue.GetType (), value))
					this.ModeValue = value;
				else
                    throw new Exception(messageLoader.GetString("SymAlgoPaddingMode"));
					//throw new System.Security.Cryptography.CryptographicException ("padding mode not available");
                    //throw new System.Exception("SymmetricAlgorithm: padding mode not available");
				
				if (value == CipherMode.OpenPGP_CFB) {
					IV = new byte[BlockSize >> 3];
					this.IVValue.Initialize();
				}
			}
		}

		/// <summary>
		/// Gets or sets the actual padding
		/// </summary>
		public virtual PaddingMode Padding {
			get { return this.PaddingValue; }
			set {
                var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
				if (Enum.IsDefined (PaddingValue.GetType (), value))
					this.PaddingValue = value;
				else
                    throw new Exception(messageLoader.GetString("SymAlgoPaddingMode"));
					//throw new System.Security.Cryptography.CryptographicException ("padding mode not available");
                    //throw new System.Exception("SymmetricAlgorithm: padding mode not available");
			}
		}

		/// <summary>
		/// Gets an Decryptor transform object to work with a CryptoStream
		/// </summary>
		public virtual ICryptoTransform CreateDecryptor () {
			return CreateDecryptor (Key, IV);
		}

		/// <summary>
		/// Gets an Decryptor transform object to work with a CryptoStream
		/// </summary>
		public abstract ICryptoTransform CreateDecryptor (byte[] rgbKey, byte[] rgbIV);

		/// <summary>
		/// Gets an Encryptor transform object to work with a CryptoStream
		/// </summary>
		public virtual ICryptoTransform CreateEncryptor() {
			return CreateEncryptor (Key, IV);
		}

		/// <summary>
		/// Gets an Encryptor transform object to work with a CryptoStream
		/// </summary>
		public abstract ICryptoTransform CreateEncryptor (byte[] rgbKey, byte[] rgbIV);

		/// <summary>
		/// used to generate an inital vector if none is specified
		/// </summary>
		public abstract void GenerateIV ();

		/// <summary>
		/// used to generate a random key if none is specified
		/// </summary>
		public abstract void GenerateKey ();

		internal bool IsLegalKeySize (KeySizes[] LegalKeys, int Size) {
			foreach (KeySizes LegalKeySize in LegalKeys) {
				for (int i=LegalKeySize.MinSize; i<=LegalKeySize.MaxSize; i+=LegalKeySize.SkipSize) {
					if (i == Size)
						return true;
				}
			}
			return false;
		}
		
		/// <summary>
		/// Checks wether the given keyLength is valid for the current algorithm
		/// </summary>
		/// <param name="bitLength">the given keyLength</param>
		public bool ValidKeySize (int bitLength) {
			return IsLegalKeySize (LegalKeySizesValue, bitLength);
		}
		
		/// <summary>
		/// Creates the default implementation of the default symmetric algorithm (Rijndael).
		/// </summary>
		// LAMESPEC: Default is Rijndael - not TripleDES
		public static SymmetricAlgorithm Create () {
			return (SymmetricAlgorithm)Activator.CreateInstance(Type.GetType("Hermes.Crypto.Cipher.SymmetricAlgorithm"), null);
		}
		
	}
}

