﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***   TripleDES.cs: Handles TripleDES (abstract class)           ***//
//***     Authors:                                                 ***//
//***     Sebastien Pouliot (spouliot@motus.com)                   ***//
//***     (C) 2002, 2003 Motus Technologies Inc.                   ***//
//***     (http://www.motus.com)                                   ***//
//***                                                              ***//
//*** Parts of this code rely on portions of the go-mono           ***//
//*** implemenation of the cipher. These parts are copyright of    ***//
//*** their respective owners:                                     ***//
//***                                                              ***//
//*** Get get the original SymmetricAlgorithm class, please visit  ***//
//*** the mono project at http://www.go-mono.com.                  ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** TripleDES.cs:                                                ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    01.06.2003: Daniel Fabian - To fit SharpPrivacy's needs   ***//
//***    31.10.2013  James R. Stembridge - To fit the needs of     ***//
//***                Hermes Crypto Library                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;

namespace Hermes.Crypto.Cipher {

// References:
// a.	FIPS PUB 46-3: TripleDES
//	http://csrc.nist.gov/publications/fips/fips46-3/fips46-3.pdf
// b.	ANSI X9.52
//	not free :-(
//	http://webstore.ansi.org/ansidocstore/product.asp?sku=ANSI+X9%2E52%2D1998

public abstract class TripleDES : SymmetricAlgorithm {

	public TripleDES() {
		// from SymmetricAlgorithm
		KeySizeValue = 192;
		BlockSizeValue = 64;
		FeedbackSizeValue = 64;

		LegalKeySizesValue = new KeySizes [1];
		LegalKeySizesValue [0] = new KeySizes (128, 192, 64);

		LegalBlockSizesValue = new KeySizes [1];
		LegalBlockSizesValue [0] = new KeySizes (64, 64, 0);
	}

	public override byte[] Key {
		get {
			if (KeyValue == null) {
				// generate keys as long as we get weak keys
				GenerateKey ();
				while (IsWeakKey (KeyValue))
					GenerateKey ();
			}
			return KeyValue;
		}
		set { 
			if (value == null)
				throw new ArgumentNullException ();
			// this will check for both key size and weak keys
			if (IsWeakKey (value))
				throw new Exception ();
			KeyValue = (byte[]) value.Clone (); 
		}
	}

	// Triple DES is DES in EDE = Encrypt - Decrypt - Encrypt
	// with 2 keys (a,b)
	//	EDE = Encrypt (a) - Decrypt (b) - Encrypt (a)
        //	if a == b then TripleDES == DES(a) (hence weak key)
        // with 3 keys (a,b,c)
	//	EDE = Encrypt (a) - Decrypt (b) - Encrypt (c)
	//	if ( a == b ) then TripleDES == DES(c) (hence weak key)
	//	if ( b == c ) then TripleDES == DES(a) (hence weak key)
	public static bool IsWeakKey (byte[] rgbKey) {
		// 128 bits (16 bytes) is 3 DES with 2 keys
		if (rgbKey.Length == 16) {
			// weak if first half == second half
			for (int i = 0; i < 8; i++)
				if (rgbKey [i] != rgbKey [i+8])
					return false;
		}
		// 192 bits (24 bytes) is 3 DES with 3 keys
		else if (rgbKey.Length == 24) {
			bool bFirstCase = true;	
			// weak if first third == second third
			for (int i = 0; i < 8; i++) {
				if (rgbKey [i] != rgbKey [i+8]) {
					bFirstCase = false;
					break;
				}
			}
			// weak if second third == third third 
			if (!bFirstCase) {
				for (int i = 8; i < 16; i++)
					if (rgbKey [i] != rgbKey [i+8])
						return false;
			}
		} else
			throw new Exception ();

		return true;
	}

	public static new TripleDES Create() {
		return (TripleDES)Activator.CreateInstance(Type.GetType("Hermes.Crypto.Cipher.TripleDESCryptoServiceProvider"), null);
	}

}

}

