﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.                                       ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** KeyManagement.cs:                                            ***//
//*** 	Class to serve as an API for key management functions.     ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    01.11.2013  James R. Stembridge - Created this module.    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
using System;
using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Storage.Streams;
using Windows.Security.Cryptography;
using System.Runtime.InteropServices.WindowsRuntime;
using Hermes.Crypto.OpenPGP;
using Hermes.Crypto.OpenPGP.Messages;
using Hermes.Crypto.Cipher;
using Hermes.Crypto.Cipher.Math;

namespace Hermes.Crypto
{
    public class KeyManagement
    {
        private static BigInteger[][] bigIntEncryptionKey;
        private static BigInteger[][] bigIntSignatureKey;


        /// <summary>
        /// ChangePassPhrase - change the pass phrase in a secret key.
        /// You should reload your copy of the secret key ring after this.
        /// </summary>
        /// <param name="lKeyID"> The ulong Key ID of the key to change. 
        /// </param>
        /// <param name="oldPassPhrase"> The user's old pass phrase. 
        /// </param>
        /// <param name="newPassPhrase"> The user's new phrase. 
        /// </param>
        /// <param name="secKeyRing"> The Secret Key Ring.
        /// </param>
        /// <returns> Nothing.</returns>
        public async Task ChangePassPhrase(ulong lKeyID, string oldPassPhrase, string newPassPhrase, SecretKeyRing secretKeyRing)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            BigInteger[] bigIntEncKeys = new BigInteger[0];
            BigInteger[] bigIntSigKeys = new BigInteger[0];
            // Find the keys
            TransportableSecretKey transSecKeyChangeKey = secretKeyRing.Find(lKeyID);
            SecretKeyPacket secKeyPktSignatureKey = transSecKeyChangeKey.PrimaryKey;
            SecretKeyPacket secKeyPktEncryptKey = new SecretKeyPacket();
            string errMsg = string.Empty;
            bool keyFound = false;
            IEnumerator ieSubkeys = transSecKeyChangeKey.SubKeys.GetEnumerator();
            while (ieSubkeys.MoveNext())
            {
                if (!(ieSubkeys.Current is SecretKeyPacket))
                {
                    throw new Exception(messageLoader.GetString("SevereKeyCorrupted"));
                }
                secKeyPktEncryptKey = (SecretKeyPacket)ieSubkeys.Current;
                if (secKeyPktEncryptKey.PublicKey.KeyID == lKeyID)
                {
                    keyFound = true;
                    continue;
                }
            }
            if (!keyFound)
            {
                throw new Exception(messageLoader.GetString("SevereNoEncKey"));
            }
            // Decrypt the key material for the encryption key
            errMsg = string.Empty;
            try
            {
                bigIntEncKeys = secKeyPktEncryptKey.GetDecryptedKeyMaterial(oldPassPhrase);
            }
            catch (Exception ee)
            {
                //errMsg = "Error retrieving encryption key: " + ee.Message;
                errMsg = messageLoader.GetString("EncKeyRetrieveError") + ee.Message;
            }
            if (errMsg != string.Empty)
            {
                throw new Exception(errMsg);
            }
            // Decrypt the key material for the signature key
            errMsg = string.Empty;
            try
            {
                bigIntSigKeys = secKeyPktSignatureKey.GetDecryptedKeyMaterial(oldPassPhrase);
            }
            catch (Exception ee)
            {
                //errMsg = "Error retrieving signature key: " + ee.Message;
                errMsg = messageLoader.GetString("SigKeyRetrieveError") + ee.Message;
            }
            if (errMsg != string.Empty)
            {
                throw new Exception(errMsg);
            }
            // Remove the kay from the key ring. We do this before making changes, so there is no
            // room for confusion during the delete operation.
            secretKeyRing.Delete(transSecKeyChangeKey);
            // Encrypt the key material for the encryption key with the new passphrase
            transSecKeyChangeKey.SubKeys.Remove(secKeyPktEncryptKey);
            errMsg = string.Empty;
            try
            {
                secKeyPktEncryptKey.EncryptKeyMaterial(bigIntEncKeys, newPassPhrase);
            }
            catch (Exception ee)
            {
                //errMsg = "Error changing passphrase for encryption key: " + ee.Message;
                errMsg = messageLoader.GetString("ErrorChangingPassPhraseEncKey") + ee.Message;
            }
            if (errMsg != string.Empty)
            {
                throw new Exception(errMsg);
            }
            transSecKeyChangeKey.SubKeys.Add(secKeyPktEncryptKey);
            // Encrypt the key material for the signature key with the new passphrase
            try
            {
                secKeyPktSignatureKey.EncryptKeyMaterial(bigIntSigKeys, newPassPhrase);
            }
            catch (Exception ee)
            {
                //errMsg = "Error changing passphrase for signature key: " + ee.Message;
                errMsg = messageLoader.GetString("ErrorChangingPassPhraseSigKey") + ee.Message;
            }
            if (errMsg != string.Empty)
            {
                throw new Exception(errMsg);
            }
            transSecKeyChangeKey.PrimaryKey = secKeyPktSignatureKey;
            // Replace the Transportable Secret Key in the key ring
            secretKeyRing.AddSecretKey(transSecKeyChangeKey);
            // Sort the key ring
            secretKeyRing.Sort();
            // Save the changed key ring
            await secretKeyRing.Save("localstore");
            return;
        }

        /// <summary>
        /// ChangePassPhrase - change the pass phrase in a secret key.
        /// You should reload your copy of the public key ring after this.
        /// </summary>
        /// <param name="lKeyID"> The ulong Key ID of the key to change. 
        /// </param>
        /// <param name="pubicKeyRing"> The Public Key RIng. 
        /// </param>
        /// <param name="secKeyRing"> The Secret Key Ring.
        /// </param>
        /// <param name="strPassPhrase"> The user's pass phrase.
        /// </param>
        /// <param name="newDate"> The new expiration date. 
        /// </param>
        /// <param name="neverExpires"> Boolian indicator that the key
        /// never expires. 
        /// </param>
        /// <returns> Nothing.</returns>
        public async Task ChangeExpirationDate(ulong lKeyID, PublicKeyRing publicKeyRing, SecretKeyRing secretKeyRing, string strPassPhrase, DateTime newDate, bool neverExpires)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            TransportablePublicKey changeKey = publicKeyRing.Find(lKeyID, false);
            TransportablePublicKey holdKey = changeKey;
            // Build the userID packet
            UserIDPacket userIDPkt = new UserIDPacket();
            userIDPkt.UserID = changeKey.PrimaryUserID;
            //newCertUserID.UserID = userIDPkt;
            List<SignaturePacket> delList = new List<SignaturePacket>();
            List<SignaturePacket> addList = new List<SignaturePacket>();
            //
            // Loop throught all of the CERTIFICATES and recreate them
            //
            bool first = true;
            IEnumerator ieCertificates = changeKey.Certifications.GetEnumerator();
            while (ieCertificates.MoveNext())
            {
                if (!(ieCertificates.Current is CertifiedUserID))
                    continue;
                delList.Clear();
                addList.Clear();
                CertifiedUserID certUserID = (CertifiedUserID)ieCertificates.Current;
                IEnumerator ieSignatures = certUserID.Certificates.GetEnumerator();
                while (ieSignatures.MoveNext())
                {
                    if (!(ieSignatures.Current is SignaturePacket))
                        continue;
                    SignaturePacket sigPktCertificate = (SignaturePacket)ieSignatures.Current;
                    SignaturePacket newSigPacket = new SignaturePacket();
                    newSigPacket.Version = SignaturePacketVersionNumbers.v4;
                    newSigPacket.HashAlgorithm = HashAlgorithms.SHA1;
                    newSigPacket.KeyID = changeKey.PrimaryKey.KeyID;
                    newSigPacket.TimeCreated = changeKey.PrimaryKey.TimeCreated;
                    IEnumerator ieSubPackets = sigPktCertificate.HashedSubPackets.GetEnumerator();
                    while (ieSubPackets.MoveNext())
                    {
                        SignatureSubPacket sigSubPacket = (SignatureSubPacket)ieSubPackets.Current;
                        if (sigSubPacket.Type == SignatureSubPacketTypes.KeyExpirationTime)
                            continue;
                        if (sigSubPacket.Type == SignatureSubPacketTypes.IssuerKeyID)
                            continue;
                        if (sigSubPacket.Type == SignatureSubPacketTypes.SignatureCreationTime)
                            continue;
                        newSigPacket.AddSubPacket(sigSubPacket, true);
                    }
                    if (first && !neverExpires)
                    {
                        SignatureSubPacket expirationSubPacket = new SignatureSubPacket();
                        expirationSubPacket.Type = SignatureSubPacketTypes.KeyExpirationTime;
                        expirationSubPacket.KeyExpirationTime = new DateTime(newDate.Ticks + (new DateTime(1970, 1, 1)).Ticks - changeKey.PrimaryKey.TimeCreated.Ticks);
                        newSigPacket.AddSubPacket(expirationSubPacket, true);
                        first = false;
                    }
                    delList.Add(sigPktCertificate);
                    addList.Add(newSigPacket); // = new List<SignaturePacket>();
                }

                
                for (int i = 0; i < addList.Count; i++)
                {
                    TransportableSecretKey transSecretKey = secretKeyRing.Find(lKeyID);
                    SecretKeyPacket sigSecretKey = transSecretKey.FindKey(AsymActions.Sign);
                    string errMsg = string.Empty;
                    try
                    {
                        certUserID.Sign(addList[i], sigSecretKey, strPassPhrase, changeKey.PrimaryKey);
                    }
                    catch (Exception ee)
                    {
                        errMsg = ee.Message;
                    }
                    if (errMsg != string.Empty)
                    {
                        changeKey = holdKey;
                        throw new Exception(errMsg);
                    }
                }

                for (int i = 0; i < delList.Count; i++)
                {
                    certUserID.Certificates.Remove(delList[i]);
                }
            }
            //
            // Loop throught all of the SIGNATURES and recreate them
            //
            // Build new subkeys packet(s)

            // JRS = NOTICE - I don't think we need to do this!!! GnuPG doesn't seem to.
            // In fact, the key I just looked at from GnuPG did't have any SubKey packets.
            // Which makes me question why we created them in the first place.

            /*
            CertifiedPublicSubkey cpsEncryptionKey = new CertifiedPublicSubkey();
            cpsEncryptionKey.Subkey = pkpEncryptionKey;
            long time = enteredDate.Ticks; // -DateTime.Now.Ticks;
            if (enteredDate <= controlDate)
                time = 0;
            cpsEncryptionKey.SignKeyBindingSignature(pkpSignatureKey, secKeyPktSignatureKey, strPassphrase, new DateTime(time), true);
            */
            publicKeyRing.Delete(lKeyID);
            publicKeyRing.AddPublicKey(changeKey);
            publicKeyRing.Sort();
            await publicKeyRing.Save("localstore");
            return;
        }
         
        /// <summary>
        /// Certify Key - sign a user's key indicating a level of
        /// trust for this key.  I.e, you know and trust the person.
        /// You should reload your copy of the public key ring after this.
        /// </summary>
        /// <param name="lKeyID"> The ulong Key ID of the key to certify. 
        /// </param>
        /// <param name="pubicKeyRing"> The Public Key RIng. 
        /// </param>
        /// <param name="transSecKey"> The transportable secret key of 
        /// the person signing the key.
        /// </param>
        /// <param name="strPassPhrase"> The signer's pass phrase.
        /// </param>
        /// <param name="certUserID"> The certified userID packet to sign/certify. 
        /// </param>
        /// <param name="stType"> The signature type.
        /// </param>
        /// <param name="trustDepth"> The trust depth, how well you trust
        /// the person.
        /// </param>
        /// <param name="isExportable"> A boolian indicator designating if
        /// this sign/certify is exportable of just for internal use.
        /// </param>
        /// <returns> Nothing.</returns>
        public async Task CertifyKey(ulong lKeyID, PublicKeyRing publicKeyRing, TransportableSecretKey transSecKey, string strPassPhrase, CertifiedUserID certUserID, SignatureTypes stType, int trustDepth, bool isExportable)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            TransportablePublicKey transPubKey;
            bool rc = false;
            transPubKey = publicKeyRing.Find(lKeyID, false);
            SecretKeyPacket skpSignatureKey = transSecKey.FindKey(AsymActions.Sign);
            SignaturePacket sigPktCertificate = new SignaturePacket();
            sigPktCertificate.SignatureType = stType;
            sigPktCertificate.Version = SignaturePacketVersionNumbers.v4;
            sigPktCertificate.HashAlgorithm = HashAlgorithms.SHA1;
            sigPktCertificate.KeyID = skpSignatureKey.PublicKey.KeyID;
            sigPktCertificate.TimeCreated = DateTime.Now;
            if (isExportable)
            {
                SignatureSubPacket sigSubPktNotExportable = new SignatureSubPacket();
                sigSubPktNotExportable.Type = SignatureSubPacketTypes.ExportableSignature;
                sigSubPktNotExportable.ExportableSignature = false;
                sigPktCertificate.AddSubPacket(sigSubPktNotExportable, true);
            }
            if (trustDepth > 0)
            {
                SignatureSubPacket sigSubPktTrust = new SignatureSubPacket();
                sigSubPktTrust.Type = SignatureSubPacketTypes.TrustSignature;
                sigSubPktTrust.TrustLevel = (byte)trustDepth;
                sigSubPktTrust.TrustAmount = 120;
                sigPktCertificate.AddSubPacket(sigSubPktTrust, true);
            }
            string errMsg = string.Empty;
            try
            {
                certUserID.Sign(sigPktCertificate, skpSignatureKey, strPassPhrase, transPubKey.PrimaryKey);
            }
            catch (Exception ee)
            {
                rc = true;
                //errMsg = "Error signing key: " + ee.Message;
                errMsg = messageLoader.GetString("ErrorSigningKey") + ee.Message;
            }
            if (rc)
            {
                throw new Exception(errMsg);
            }


            transPubKey.Certifications.Remove(certUserID);
            transPubKey.Certifications.Add(certUserID);
            publicKeyRing.Delete(lKeyID);
            publicKeyRing.AddPublicKey(transPubKey);
            publicKeyRing.Sort();
            await publicKeyRing.Save("localstore");
            return;
        }

        /// <summary>
        /// AddUserID - add a userID to a user's public key.
        /// You should reload your copy of the public key ring 
        /// and the secret key ring after this.
        /// </summary>
        /// <param name="lKeyID"> The ulong Key ID of the key to certify. 
        /// </param>
        /// <param name="userID"> The Public Key RIng. 
        /// </param>
        /// <param name="strPassPhrase"> The signer's pass phrase.
        /// </param>
        /// <param name="pubicKeyRing"> The Public Key RIng. 
        /// </param>
        /// <param name="secretKeyRing"> The Secret Key RIng. 
        /// </param>
        /// <returns> Nothing.</returns>
        public async Task AddUserID(ulong lKeyID, string userID, string strPassPhrase, PublicKeyRing publicKeyRing, SecretKeyRing secretKeyRing)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            bool rc = false;
            TransportablePublicKey transPubKey;
            TransportableSecretKey transSecKey;
            transSecKey = secretKeyRing.Find(lKeyID);
            transPubKey = publicKeyRing.Find(lKeyID, false);
            CertifiedUserID certUserID = new CertifiedUserID();
            UserIDPacket userIDPkt = new UserIDPacket();
            userIDPkt.UserID = userID;
            certUserID.UserID = userIDPkt;
            string strPassphrase = strPassPhrase;
            SecretKeyPacket secKeyPktSignature = transSecKey.FindKey(AsymActions.Sign);
            SignaturePacket sigPktSelfSignature = new SignaturePacket();
            sigPktSelfSignature.Version = SignaturePacketVersionNumbers.v4;
            sigPktSelfSignature.HashAlgorithm = HashAlgorithms.SHA1;
            sigPktSelfSignature.KeyID = secKeyPktSignature.PublicKey.KeyID;
            sigPktSelfSignature.TimeCreated = DateTime.Now;
            // This is not the Primary User ID. Therefore, we do not need a Primary User ID subpacket
            // Build the preferred symmetric algorythms subpacket
            SignatureSubPacket sigSubPktPreferedSymAlgos = new SignatureSubPacket();
            sigSubPktPreferedSymAlgos.Type = SignatureSubPacketTypes.PreferedSymmetricAlgorithms;
            sigSubPktPreferedSymAlgos.PreferedSymAlgos = new SymAlgorithms[] { SymAlgorithms.AES256, SymAlgorithms.AES192, SymAlgorithms.AES256, SymAlgorithms.CAST5, SymAlgorithms.Triple_DES };
            sigPktSelfSignature.AddSubPacket(sigSubPktPreferedSymAlgos, true);
            // Build the preferred hash algorythm subpacket
            SignatureSubPacket sigSubPktPreferedHashAlgos = new SignatureSubPacket();
            sigSubPktPreferedHashAlgos.Type = SignatureSubPacketTypes.PreferedHashAlgorithms;
            sigSubPktPreferedHashAlgos.PreferedHashAlgos = new HashAlgorithms[] { HashAlgorithms.SHA1 };
            sigPktSelfSignature.AddSubPacket(sigSubPktPreferedHashAlgos, true);
            // Build the features subpacket (says we can do MDC processing)
            SignatureSubPacket sigSubPktFeatures = new SignatureSubPacket();
            sigSubPktFeatures.Type = SignatureSubPacketTypes.Features;
            sigSubPktFeatures.MDC = 0x01;                                 // Can use MDC processing
            sigPktSelfSignature.AddSubPacket(sigSubPktFeatures, true);
            certUserID.Certificates = new List<SignaturePacket>();
            string errMsg = string.Empty;
            try
            {
                certUserID.Sign(sigPktSelfSignature, secKeyPktSignature, strPassphrase, transPubKey.PrimaryKey);
            }
            catch (Exception ee)
            {
                rc = true;
                errMsg = messageLoader.GetString("AddErrorSigningKey") + ": " + ee.Message;
                //errMsg = "Error signing key: " + ee.Message;
            }
            if (rc)
            {
                throw new Exception(errMsg);
            }
            transPubKey.Certifications.Add(certUserID);
            transSecKey.UserIDs.Add(userIDPkt);
            publicKeyRing.Delete(lKeyID);
            publicKeyRing.AddPublicKey(transPubKey);
            await publicKeyRing.Save("localstore");
            secretKeyRing.Delete(lKeyID);
            secretKeyRing.AddSecretKey(transSecKey);
            await secretKeyRing.Save("localstore");
        }

        /// <summary>
        /// Create Keys - generate a public/secret key pair and add
        /// them to the key rings.
        /// You should reload your copy of the public key ring 
        /// and the secret key ring after this.
        /// </summary>
        /// <param name="keySize"> The key size to generate.
        /// </param>
        /// <param name="isRSA"> Boolian indicator for RSA encryption. 
        /// </param>
        /// <param name="Encryption"> Boolian indicator that this key 
        /// will be used for encrypting messages.
        /// </param>
        /// <param name="Signing"> Boolian indicator that this key 
        /// will be used for signing messages.
        /// </param>
        /// <param name="isElgamal"> Boolin indicator for Elgamal encryption. 
        /// </param>
        /// <param name="neverExpires"> Boolian indicator that the key 
        /// should never expire.
        /// </param>
        /// <param name="strPassPhrase"> String containing the user's
        /// pass phrase.
        /// </param>
        /// <param name="userID"> String containing the userID to
        /// use for this key pair.
        /// </param>
        /// <param name="expireDate"> DateTime expiration date. 
        /// </param>
        /// <param name="pubicKeyRing"> The Public Key RIng. 
        /// </param>
        /// <param name="secretKeyRing"> The Secret Key RIng. 
        /// </param>
        /// <returns> Nothing.</returns>
        public async Task CreateKeys(int keySize, bool isRSA, bool Encryption, bool Signing, bool isElgamal, bool neverExpires, string strPassphrase, string userID, DateTime expireDate, PublicKeyRing publicKeyRing, SecretKeyRing secretKeyRing)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            // We need to generate the key material for both the
            // encryption key and the signature keys.
            // This can be a lengthy process, so we should do it
            // in two parallel tasks. 
            Task genEncKey = GenerateEncryptionKey(isRSA, keySize);
            Task genSigKey = GenerateSignatureKey(isRSA, keySize, Encryption, Signing);
            await genEncKey;
            await genSigKey;
            // If the user selected RSA encryption, we'll use the same
            // key for both encryption and signatures.
            if (isRSA)
            {
                if (Encryption && Signing)
                {
                    bigIntSignatureKey = bigIntEncryptionKey;
                }
            }
            // Build the public signature packet
            PublicKeyPacket pubKeyPktSignatureKey = new PublicKeyPacket(false);
            if (isRSA)
            {
                pubKeyPktSignatureKey.Algorithm = AsymAlgorithms.RSA_Encrypt_Sign;
            }
            else
            {
                pubKeyPktSignatureKey.Algorithm = AsymAlgorithms.DSA;
            }
            pubKeyPktSignatureKey.KeyMaterial = bigIntSignatureKey[0];
            pubKeyPktSignatureKey.TimeCreated = DateTime.Now;
            pubKeyPktSignatureKey.Version = PublicKeyPacketVersionNumbers.v4;
            // Build the secret signature packet
            SecretKeyPacket secKeyPktSignatureKey = new SecretKeyPacket(false);
            secKeyPktSignatureKey.SymmetricalAlgorithm = SymAlgorithms.AES256;
            secKeyPktSignatureKey.PublicKey = pubKeyPktSignatureKey;
            secKeyPktSignatureKey.InitialVector = new byte[CipherHelper.CipherBlockSize(SymAlgorithms.AES256)];
            UInt32 targetLen = (UInt32)secKeyPktSignatureKey.InitialVector.Length;
            // Generate random data and copy it to a buffer.
            IBuffer buffer = CryptographicBuffer.GenerateRandom(targetLen);
            secKeyPktSignatureKey.InitialVector = buffer.ToArray();
            //The secret signature key material is encrypted with the passphrase
            secKeyPktSignatureKey.EncryptKeyMaterial(bigIntSignatureKey[1], strPassphrase);
            // The secret key contains a copy of the public key
            secKeyPktSignatureKey.PublicKey = pubKeyPktSignatureKey;
            // Build the public encrytion packet
            PublicKeyPacket pubKeyPktEncryptionKey = new PublicKeyPacket(true);
            // Set the switch that indicates what encryption method is used
            if (isRSA)
            {
                if (Encryption && Signing)
                    pubKeyPktEncryptionKey.Algorithm = AsymAlgorithms.RSA_Encrypt_Sign;
                else if (Encryption)
                    pubKeyPktEncryptionKey.Algorithm = AsymAlgorithms.RSA_Encrypt_Only;
                else if (Signing)
                    pubKeyPktEncryptionKey.Algorithm = AsymAlgorithms.RSA_Sign_Only;
            }
            else if (isElgamal)
            {
                pubKeyPktEncryptionKey.Algorithm = AsymAlgorithms.ElGamal_Encrypt_Only;
            }
            else
            {
                pubKeyPktEncryptionKey.Algorithm = AsymAlgorithms.ElGamal_Encrypt_Only;
            }
            pubKeyPktEncryptionKey.KeyMaterial = bigIntEncryptionKey[0];
            pubKeyPktEncryptionKey.TimeCreated = DateTime.Now;
            pubKeyPktEncryptionKey.Version = PublicKeyPacketVersionNumbers.v4;
            // Build the secret encryption packet
            SecretKeyPacket secKeyPktEncryptionKey = new SecretKeyPacket(true);
            secKeyPktEncryptionKey.SymmetricalAlgorithm = SymAlgorithms.AES256;
            secKeyPktEncryptionKey.PublicKey = pubKeyPktEncryptionKey;
            secKeyPktEncryptionKey.InitialVector = new byte[CipherHelper.CipherBlockSize(SymAlgorithms.AES256)];
            targetLen = (UInt32)secKeyPktEncryptionKey.InitialVector.Length;
            // Generate random data and copy it to a buffer.
            IBuffer buffer2 = CryptographicBuffer.GenerateRandom(targetLen);
            secKeyPktEncryptionKey.InitialVector = buffer2.ToArray();
            //The secret encryption key material is encrypted with the passphrase
            secKeyPktEncryptionKey.EncryptKeyMaterial(bigIntEncryptionKey[1], strPassphrase);
            secKeyPktEncryptionKey.PublicKey = pubKeyPktEncryptionKey;
            // Build the userID packet
            CertifiedUserID certUserID = new CertifiedUserID();
            UserIDPacket userIDPkt = new UserIDPacket();
            userIDPkt.UserID = userID.Trim();
            certUserID.UserID = userIDPkt;
            // Build a self-signed siganture packet & subpackets
            SignaturePacket sigPktSelfSig = new SignaturePacket();
            sigPktSelfSig.Version = SignaturePacketVersionNumbers.v4;
            sigPktSelfSig.HashAlgorithm = HashAlgorithms.SHA1;
            sigPktSelfSig.KeyID = pubKeyPktSignatureKey.KeyID;
            sigPktSelfSig.TimeCreated = DateTime.Now;
            // Build the Primary User ID subpacket
            SignatureSubPacket sigSubPktPrimaryUserID = new SignatureSubPacket();
            sigSubPktPrimaryUserID.Type = SignatureSubPacketTypes.PrimaryUserID;
            sigSubPktPrimaryUserID.PrimaryUserID = true;
            sigPktSelfSig.AddSubPacket(sigSubPktPrimaryUserID, true);
            // Build the preferred symmetric algorythms subpacket
            SignatureSubPacket sigSubPktPreferedSymAlgos = new SignatureSubPacket();
            sigSubPktPreferedSymAlgos.Type = SignatureSubPacketTypes.PreferedSymmetricAlgorithms;
            sigSubPktPreferedSymAlgos.PreferedSymAlgos = new SymAlgorithms[] { SymAlgorithms.AES256, SymAlgorithms.AES192, SymAlgorithms.AES256, SymAlgorithms.CAST5, SymAlgorithms.Triple_DES };
            sigPktSelfSig.AddSubPacket(sigSubPktPreferedSymAlgos, true);
            // Build the preferred hash algorythm subpacket
            SignatureSubPacket sigSubPktPreferedHashAlgos = new SignatureSubPacket();
            sigSubPktPreferedHashAlgos.Type = SignatureSubPacketTypes.PreferedHashAlgorithms;
            sigSubPktPreferedHashAlgos.PreferedHashAlgos = new HashAlgorithms[] { HashAlgorithms.SHA1 };
            sigPktSelfSig.AddSubPacket(sigSubPktPreferedHashAlgos, true);
            // Build the features subpacket (says we can do MDC processing)
            SignatureSubPacket sigSubPktFeatures = new SignatureSubPacket();
            sigSubPktFeatures.Type = SignatureSubPacketTypes.Features;
            sigSubPktFeatures.MDC = 0x01;                                 // Can use MDC processing
            sigPktSelfSig.AddSubPacket(sigSubPktFeatures, true);
            // IF the user entered an expiration we need to generate a key expiration sub-packet
            DateTime controlDate;
            if (!neverExpires)
            {
                SignatureSubPacket sspExpiration = new SignatureSubPacket();
                //sspExpiration.Type = SignatureSubPacketTypes.SignatureExpirationTime;
                sspExpiration.Type = SignatureSubPacketTypes.KeyExpirationTime;
                controlDate = DateTime.Now;

                sspExpiration.KeyExpirationTime = new DateTime(expireDate.Ticks + (new DateTime(1970, 1, 2)).Ticks - pubKeyPktSignatureKey.TimeCreated.Ticks);
                sigPktSelfSig.AddSubPacket(sspExpiration, true);
            }
            else
            {
                expireDate = DateTime.Now;
                controlDate = expireDate;
            }
            // Build new certificates packet(s)
            certUserID.Certificates = new List<SignaturePacket>();
            certUserID.Sign(sigPktSelfSig, secKeyPktSignatureKey, strPassphrase, pubKeyPktSignatureKey);
            // Build new subkeys packet(s)
            CertifiedPublicSubkey certPubSubkeyEncryptionKey = new CertifiedPublicSubkey();
            certPubSubkeyEncryptionKey.Subkey = pubKeyPktEncryptionKey;
            long time = expireDate.Ticks; // -DateTime.Now.Ticks;
            if (expireDate <= controlDate)
                time = 0;
            certPubSubkeyEncryptionKey.SignKeyBindingSignature(pubKeyPktSignatureKey, secKeyPktSignatureKey, strPassphrase, new DateTime(time), true);
            // Build the transportable public key
            TransportablePublicKey transPublicKey = new TransportablePublicKey();
            transPublicKey.PrimaryKey = pubKeyPktSignatureKey;
            transPublicKey.SubKeys.Add(certPubSubkeyEncryptionKey);
            transPublicKey.Certifications.Add(certUserID);
            // Build the transportable secret key
            TransportableSecretKey transSecretKey = new TransportableSecretKey();
            transSecretKey.PrimaryKey = secKeyPktSignatureKey;
            transSecretKey.SubKeys.Add(secKeyPktEncryptionKey);
            transSecretKey.UserIDs.Add(userIDPkt);
            // Add the two new transportable keys to the KeyRing(s)
            // and sort the KeyRing(s)
            publicKeyRing.AddPublicKey(transPublicKey);
            secretKeyRing.AddSecretKey(transSecretKey);
            publicKeyRing.Sort();
            secretKeyRing.Sort();
            // Save the KeyRing(s) 
            await secretKeyRing.Save("localstore");
            await publicKeyRing.Save("localstore");
            return;
        }

        /// <summary>
        /// GenerateEncryptionKey - private internal subroutine
        /// to generate the encryption key.
        /// </summary>
        /// <param name="isRSA"> Boolian indicator for RSA encryption. 
        /// </param>
        /// <param name="keySize"> The key size to generate.
        /// </param>
        /// <returns> Nothing.</returns>
        private async Task GenerateEncryptionKey(bool isRSA, int keySize)
        {
            if (isRSA)
            {
                RSA rsaKeyGenerator = new RSA();
                await Task.Factory.StartNew(() =>
                {
                    bigIntEncryptionKey = rsaKeyGenerator.Generate(keySize);
                });
            }
            else
            {
                ElGamal egKeyGenerator = new ElGamal();
                await Task.Factory.StartNew(() =>
                {
                    bigIntEncryptionKey = egKeyGenerator.Generate(keySize);
                });
            }
            return;
        }

        /// <summary>
        /// GenerateSignatureKey - private internal subroutine
        /// to generate the signature key.
        /// </summary>
        /// <param name="isRSA"> Boolian indicator for RSA encryption. 
        /// </param>
        /// <param name="keySize"> The key size to generate.
        /// </param>
        /// <param name="Encryption"> Boolian indicator that this key 
        /// will be used for encrypting messages.
        /// </param>
        /// <param name="Signing"> Boolian indicator that this key 
        /// will be used for signing messages.
        /// </param>
        /// <returns> Nothing.</returns>
        private async Task GenerateSignatureKey(bool isRSA, int keySize, bool Encryption, bool Signing)
        {
            if (isRSA)
            {
                if (Encryption && Signing)
                {

                    bigIntSignatureKey = bigIntEncryptionKey;
                }
                else
                {
                    RSA rsaKeyGenerator = new RSA();
                    await Task.Factory.StartNew(() =>
                    {
                        bigIntSignatureKey = rsaKeyGenerator.Generate(keySize);
                    });
                }
            }
            else
            {
                DSA dDSA = new DSA();
                await Task.Factory.StartNew(() =>
                {
                    bigIntSignatureKey = dDSA.Generate(1024);
                });
            }
            return;
        }



    }
}
