﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** CompressedDataPacket.cs:                                     ***//
//*** 	Class for handling and compressing data packets.           ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    07.04.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    14.06.2003: Changed Namespace from                        ***//
//***                SharpPrivacy.OpenPGP to                       ***//
//***                SharpPrivacy.SharpPrivacyLib.OpenPGP          ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp                         ***//
//***                Changed so that it works in WinRT.            ***//
//***                Changed to use SharpCompress which provides   ***//
//***                compression routines for WinRT                ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
//
using System;
using System.IO;
using System.IO.Compression;
using Hermes.Crypto.Cipher;
using System.Runtime.InteropServices.WindowsRuntime;
using SharpCompress;
using SharpCompress.Compressor.Deflate;

namespace Hermes.Crypto.OpenPGP
{

    public class CompressedDataPacket : Packet
    {

        private byte[] bCompressedData;
        private CompressionAlgorithms caAlgorithm;

        public byte[] CompressedData
        {
            get
            {
                return bCompressedData;
            }
            set
            {
                this.bIsUpdated = true;
                bCompressedData = value;
            }
        }

        public CompressionAlgorithms Algorithm
        {
            get
            {
                return caAlgorithm;
            }
            set
            {
                this.bIsUpdated = true;
                caAlgorithm = value;
            }
        }

        /// <summary>
        /// Creates a new Compressed Data Packet with 
        /// the parameters in pSource
        /// </summary>
        /// <param name="pSource">Packet from which the
        /// parameters are derived</param>
        public CompressedDataPacket(Packet pSource)
        {
            lLength = pSource.Length;
            bBody = pSource.Body;
            ctContent = pSource.Content;
            pfFormat = pSource.Format;
            bHeader = pSource.Header;
            this.bIsUpdated = false;
        }

        /// <summary>
        /// Creates a new compressed data packet. Format defaults
        /// to new packet format.
        /// </summary>
        public CompressedDataPacket()
        {
            bBody = new byte[0];
            bHeader = new byte[0];
            pfFormat = PacketFormats.New;
            ctContent = ContentTypes.Compressed;
            this.bIsUpdated = true;
        }

        /// <summary>
        /// Returns a string representation of the packet. This is
        /// a human readable formated representation that has nothing
        /// to do with OpenPGP or RFC2440
        /// </summary>
        /// <returns>String representation of the packet.</returns>
        /// <remarks>No remarks</remarks>
        public override string ToString()
        {
            string strReturn = "Compressed Data Packet:\r\n";
            strReturn += "Algorithm: " + caAlgorithm.ToString() + "\r\n";
            strReturn += "Compressed Data: ";
            for (int i = 0; i < this.bCompressedData.Length; i++)
            {
                string strByte = bCompressedData[i].ToString("x");
                if (strByte.Length < 2)
                {
                    strByte = "0" + strByte;
                }
                strReturn += ":" + strByte;
            }
            strReturn += "\r\n----\r\n\r\n";
            return strReturn;
        }

        /// <summary>
        /// Parses the packet given as byte array into the current
        /// class and returns this with the populated parameters.
        /// </summary>
        /// <param name="bData">A byte array containing an OpenPGP
        /// representation of the packet.</param>
        /// <returns>Returns an CompressedDataPacket that containes
        /// the parsed properties.</returns>
        /// <remarks>No remarks</remarks>
        public override Packet ParsePacket(byte[] bData)
        {
            caAlgorithm = (CompressionAlgorithms)bData[0];
            bCompressedData = new byte[bData.Length - 1];
            Array.Copy(bData, 1, bCompressedData, 0, bData.Length - 1);
            this.bIsUpdated = false;
            return this;
        }

        /// <summary>
        /// <para>Generates the content of the compressed data 
        /// packet and stores the result in the body property 
        /// of the class.</para>
        /// <para>This method SHOULD never be called directly, as it
        /// is called by the method <see cref="Generate">
        /// Generate()</see>.</para>
        /// </summary>
        /// <remarks>No remarks</remarks>
        protected override void CraftContent()
        {
            byte[] bData = new byte[bCompressedData.Length + 1];
            bData[0] = (byte)caAlgorithm;
            Array.Copy(bCompressedData, 0, bData, 1, bCompressedData.Length);

            this.bBody = bData;
        }

        public Packet[] Uncompress()
        {
            System.IO.Stream msStream = new System.IO.MemoryStream(bCompressedData);
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            //-----------------------------------------------------------------------------------------------//

            long copied = 1;

            byte[] bOutput = new byte[0];
            if (caAlgorithm == CompressionAlgorithms.ZIP)
            {
                try
                {
                    //using (var cmpStream = new ZlibStream(msStream, SharpCompress.Compressor.CompressionMode.Decompress))
                    using (var cmpStream = new SharpCompress.Compressor.Deflate.DeflateStream(msStream, SharpCompress.Compressor.CompressionMode.Decompress))
                    {
                        byte[] bWorkout = new byte[4096];
                        int xread = 1;
                        //cmpStream.BufferSize = 1024;
                        xread = cmpStream.Read(bWorkout, 0, bWorkout.Length);   ////
                        while (xread > 0)
                        {
                            byte[] bOldOutput = new byte[bOutput.Length];
                            bOutput.CopyTo(bOldOutput, 0);
                            bOutput = new byte[bOutput.Length + xread];
                            bOldOutput.CopyTo(bOutput, 0);
                            Array.Copy(bWorkout, 0, bOutput, bOldOutput.Length, xread);

                            xread = cmpStream.Read(bWorkout, 0, bWorkout.Length);  ////

                        }
                        copied = cmpStream.TotalOut;
                    }
                }
                catch (Exception e)
                {
                    string msg = messageLoader.GetString("ErrorUncompressing") + e.Message;
                    throw new Exception(msg);
                    //throw new System.Exception("Error uncompressing the message: " + e.Message);
                }
            }
            else if (caAlgorithm == CompressionAlgorithms.ZLIB)
            {
                try
                {
                    using (var cmpStream = new ZlibStream(msStream, SharpCompress.Compressor.CompressionMode.Decompress))
                    //using (var cmpStream = new SharpCompress.Compressor.Deflate.DeflateStream(msStream, SharpCompress.Compressor.CompressionMode.Decompress))
                    {
                        byte[] bWorkout = new byte[4096];
                        int xread = 1;
                        //cmpStream.BufferSize = 1024;
                        xread = cmpStream.Read(bWorkout, 0, bWorkout.Length);   ////
                        while (xread > 0)
                        {
                            byte[] bOldOutput = new byte[bOutput.Length];
                            bOutput.CopyTo(bOldOutput, 0);
                            bOutput = new byte[bOutput.Length + xread];
                            bOldOutput.CopyTo(bOutput, 0);
                            Array.Copy(bWorkout, 0, bOutput, bOldOutput.Length, xread);

                            xread = cmpStream.Read(bWorkout, 0, bWorkout.Length);  ////

                        }
                        copied = cmpStream.TotalOut;
                    }
                }
                catch (Exception e)
                {
                    string msg = messageLoader.GetString("ErrorUncompressing") + e.Message;
                    throw new Exception(msg);
                    //throw new System.Exception("Error uncompressing the message: " + e.Message);
                }
            }
            else if (caAlgorithm == CompressionAlgorithms.Uncompressed)
            {

            }
            else
            {                                
                string msg = messageLoader.GetString("ErrorUncompressingUnknownAlgo") ;
                throw new System.Exception(msg);
            }


            //-----------------------------------------------------------------------------------------------//


            //InflaterInputStream iisUnpack = new InflaterInputStream(msStream, new Inflater(true));
            /*
			byte[] bOutput = new byte[0];
			try {
				byte[] bWorkout = new byte[4096];
				int iSize = iisUnpack.Read(bWorkout, 0, bWorkout.Length);
				while (iSize > 0) {
					byte[] bOldOutput = new byte[bOutput.Length];
					bOutput.CopyTo(bOldOutput, 0);
					bOutput = new byte[bOutput.Length + iSize];
					bOldOutput.CopyTo(bOutput, 0);
					Array.Copy(bWorkout, 0, bOutput, bOldOutput.Length, iSize);
					iSize = iisUnpack.Read(bWorkout, 0, bWorkout.Length);
				}
				iisUnpack.Close();
			} catch (Exception e) {
				// TODO - Throw the exception in the main application
				// System.Windows.Forms.MessageBox.Show(e.ToString());
				throw new System.Exception("Error uncompressing the message: " + e.Message);
			}
			*/

            return Packet.ParsePackets(bOutput);

        }


    }

}
