﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** CompressedMessage.cs:                                        ***//
//*** 	Class for handling compressed messages.                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    02.03.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp.Messages.               ***//
//***                Changed to use SharpComress routines that are ***//
//***                designed to work with WinRT.                  ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher;
using SharpCompress;
using SharpCompress.Compressor.Deflate;
using System.IO;

namespace Hermes.Crypto.OpenPGP.Messages {
	
	/// <summary>
	/// Compressedmessage is an openpgp message. the content
	/// of a compressed message is always compressed. This class
	/// provides means to easily handle compressed messages, and to
	/// compress and decompress data.
	/// </summary>
	/// <remarks>
	/// Compressedmessage is an openpgp message. the content
	/// of a compressed message is always compressed. This class
	/// provides means to easily handle compressed messages, and to
	/// compress and decompress data.
	/// </remarks>
	public class CompressedMessage : Message {
		
		private CompressionAlgorithms caAlgorithm;
		private byte[] bCompressedData;
		
		/// <summary>
		/// Defines the compressionalgorithm used to
		/// compress the data. Currently only the
		/// algorithms plaintext and zip are supported.
		/// </summary>
		/// <value>The compressionalgorithm used to
		/// compress the data. Currently only the
		/// algorithms plaintext and zip are supported.
		/// </value>
		/// <remarks>No remarks</remarks>
		public CompressionAlgorithms Algorithm {
			get {
				return caAlgorithm;
			}
			set {
				// TODO - Add code to allow Algorithm Change
				caAlgorithm = value;
			}
		}
		
		/// <summary>
		/// Returns the data compressed with the given algorithm.
		/// This property is readonly. To get the data into this
		/// property, you have to call the function Compress.
		/// </summary>
		/// <value>The data compressed with the given algorithm.
		/// This property is readonly. To get the data into this
		/// property, you have to call the function Compress.</value>
		/// <remarks>No remarks</remarks>
		public byte[] CompressedData {
			get {
				return bCompressedData;
			}
		}
		
		/// <summary>
		/// Creates a new CompressedMessage. Compression algorithm
		/// defaults to ZIP.
		/// </summary>
		/// <remarks>No remarks</remarks>
		public CompressedMessage() : this(CompressionAlgorithms.ZIP) {}
		
		/// <summary>
		/// Creates a new CompressedMessage
		/// </summary>
		/// <param name="algo">The compressionalgorithm that shall be
		/// used to compress the message.</param>
		/// <remarks>No remarks</remarks>
		public CompressedMessage(CompressionAlgorithms algo) {
			bCompressedData = new byte[0];
			caAlgorithm = algo;
			pPackets = new Packet[0];
		}
		
		/// <summary>
		/// Compresses the given message into this compressed
		/// message.
		/// </summary>
		/// <param name="message">The message that is to be
		/// compressed.</param>
		/// <remarks>No remarks</remarks>
		public void Compress(Message message) {
			Compress(message.GetEncoded());
		}
		
		/// <summary>
		/// Compresses the given byte array with the currently set
		/// algorithm.
		/// </summary>
		/// <param name="bData">A binary array containing the
		/// data that is to be compressed.</param>
		/// <remarks>No remarks</remarks>
		public void Compress(byte[] bData) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (caAlgorithm == CompressionAlgorithms.ZIP) {
				MemoryStream msStream = new MemoryStream();
				//DeflaterOutputStream dosCompress = new DeflaterOutputStream(msStream, new Deflater(Deflater.DEFAULT_COMPRESSION, true));
				try {
                    //using (var cmpStream = new ZlibStream(msStream, SharpCompress.Compressor.CompressionMode.Compress, CompressionLevel.Level2, false))
                    using (var cmpStream = new DeflateStream(msStream, SharpCompress.Compressor.CompressionMode.Compress, CompressionLevel.Level2, false))
                    {
                        //byte[] outstream = null;
                        int len = (int)bData.Length;
                        //cmpStream.BufferSize = 8192;
                        //cmpStream.FlushMode = FlushType.Full;
                        cmpStream.Write(bData, 0, len);
                        cmpStream.FlushMode = FlushType.Finish;
                        cmpStream.Flush();
                        cmpStream.Dispose();
                    }

					//dosCompress.Write(bData, 0, bData.Length);
					//dosCompress.Close();
					bCompressedData = (byte[])msStream.ToArray();
				} catch (Exception e) {
                    throw new Exception(messageLoader.GetString("CmpMsgErrorCompressing") + e.Message);
                    //throw new System.Exception("Something went wrong compressing the message: " + e.Message);
				}
			} else if (caAlgorithm == CompressionAlgorithms.Uncompressed) {
				bCompressedData = bData;
			}
            else if (caAlgorithm == CompressionAlgorithms.ZLIB)
            {
                try
                {
                    MemoryStream msStream = new MemoryStream();
                    using (var cmpStream = new ZlibStream(msStream, SharpCompress.Compressor.CompressionMode.Compress, CompressionLevel.Level2, false))
                    {
                        //byte[] outstream = null;
                        int len = (int)bData.Length;
                        //cmpStream.BufferSize = 8192;
                        //cmpStream.FlushMode = FlushType.Full;
                        cmpStream.Write(bData, 0, len);
                        cmpStream.FlushMode = FlushType.Finish;
                        cmpStream.Flush();
                        cmpStream.Dispose();
                    }

                    //dosCompress.Write(bData, 0, bData.Length);
                    //dosCompress.Close();
                    bCompressedData = (byte[])msStream.ToArray();
                }
                catch (Exception e)
                {
                    throw new Exception(messageLoader.GetString("CmpMsgErrorCompressing") + e.Message);
                    //throw new System.Exception("Something went wrong compressing the message: " + e.Message);
                }
            }
		}
		
		/// <summary>
		/// Uncompresses the current compressed message and returns the message that
		/// was inside of the compressed message. Usually this should be a literal
		/// data message.
		/// </summary>
		/// <returns>Returns the message that was inside of the compressed message. 
		/// Usually this should be a literal data message.</returns>
		/// <exception cref="System.Exception">Throws an exception if the content
		/// of the compressed message is not another valid message.</exception>
		/// <remarks>No remarks</remarks>
		public Message Uncompress() {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (!(pPackets[0] is CompressedDataPacket))
				throw new System.Exception("You should never see this message. If you do, something in CompressedMessage went terribly wrong!");
			
			CompressedDataPacket cdpPacket = (CompressedDataPacket)pPackets[0];
			Packet[] pContent = cdpPacket.Uncompress();
			
			// A compressed data packet can contain:
			
			// - a literal message
			LiteralMessage lmMessage = new LiteralMessage();
			try {
				int iPos = lmMessage.ParseMessage(pContent);
				return lmMessage;
			} catch (Exception) {}
			
			// - a signed message
			SignedMessage smMessage = new SignedMessage();
			try {
				int iPos = smMessage.ParseMessage(pContent);
				return smMessage;
			} catch (Exception) {}
			
			// TODO: Try to think of other packets that might
			// occur in a compressed data packet
            throw new Exception(messageLoader.GetString("CmpMsgContentNotValidMsg"));
            //throw new Exception("The content of the compressed message does not appear to be a valid OpenPGP message!");
		}
		
		/// <summary>
		/// Parses a compressed message out of the given array of packets.
		/// In this special case, the first packet in packets MUST be
		/// a compressed data packet.
		/// </summary>
		/// <returns>Returns the number of packets used by the function.
		/// If everything works fine, it will always return 1.</returns>
		/// <param name="packets">Array of packets. The first packet in
		/// the array MUST be a compressed data packet. Otherwise an exception
		/// is thrown.</param>
		/// <remarks>No remarks</remarks>
		public override int ParseMessage(Packet[] packets) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            if (packets[0] is CompressedDataPacket)
            {
                this.pPackets = new Packet[1];
                pPackets[0] = packets[0];
                CompressedDataPacket cdpPacket = (CompressedDataPacket)packets[0];
                bCompressedData = cdpPacket.CompressedData;
                caAlgorithm = cdpPacket.Algorithm;
            }
            else
                throw new Exception(messageLoader.GetString("CmpMsgNoLitDataPkt"));
                //throw new System.ArgumentException("Expected a literal data packet as first packet in the array, but did not find it. Looks like something went terribly wrong.");
			
			return 1;
		}
		
		/// <summary>
		/// Gets the OpenPGP encoded representation of the compressed data
		/// message.
		/// </summary>
		/// <returns>Returns a byte array that contains the binary
		/// representation of the compressed message.</returns>
		/// <remarks>No remarks</remarks>
		public override byte[] GetEncoded() {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            if (pPackets.Length > 1)
                throw new Exception(messageLoader.GetString("CmpMsgInvalidPkt"));
                //throw new System.Exception("A compressed data message can contain only one packet. Something /really/ strange happened!");
			
			if (pPackets.Length == 0) {
				pPackets = new Packet[1];
				CompressedDataPacket cdpPacket = new CompressedDataPacket();
				cdpPacket.Algorithm = caAlgorithm;
				cdpPacket.CompressedData = bCompressedData;
				pPackets[0] = cdpPacket;
			}
			
			return pPackets[0].Generate();
		}
		
	}
}
