﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** ESKSequence.cs:                                              ***//
//*** 	Class for handling session keys and sequences of session   ***//
//***   keys.                                                      ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    02.03.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp.Messages.               ***//
//***                Changed so that it works in WinRT.            ***//
//***                Changed so that it handles MDC processing     ***//
//***                Changed for myPrimaryKeylKeyID processing     ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher;
using System.Collections;
using System.Collections.Generic;

namespace Hermes.Crypto.OpenPGP.Messages {
	
	/// <summary>
	/// ESKSequence is an OpenPGP message that contains one 
	/// or more session key packets (can be either public-
	/// key encryption, or summetrical encryption session keys.
	/// </summary>
	/// <remarks>
	/// ESKSequence is an OpenPGP message that contains one 
	/// or more session key packets (can be either public-
	/// key encryption, or summetrical encryption session keys.
	/// </remarks>
	public class ESKSequence : Message {
		
		private List<SymSessionKeyPacket> alSymKeys;
		private List<AsymSessionKeyPacket> alAsymKeys;
		
		// we need this arraylist so we can keep track of the
		// original order of session key packets. this is
		// neccessary so signatures that might have been made
		// over the packet can be successfully verified
		private List<Packet> alAllKeys;
		private bool bUpdated = false;
		
		/// <summary>
		/// Readonly. Returns an arraylist containing Symmetrically encrypted
		/// session key packets.
		/// </summary>
		/// <remarks>No remarks</remarks>
		/// <value>An arraylist containing Symmetrically encrypted
		/// session key packets.</value>
        public List<SymSessionKeyPacket> SymKeys
        {
			get {
				return alSymKeys;
			}
		}
		
		/// <summary>
		/// Readonly. Returns an arraylist containing Asymmetrically 
		/// encrypted session key packets.
		/// </summary>
		/// <remarks>No remarks</remarks>
		/// <value>An arraylist containing Asymmetrically encrypted
		/// session key packets.</value>
        public List<AsymSessionKeyPacket> AsymKeys
        {
			get {
				return alAsymKeys;
			}
		}
		
		/// <summary>
		/// Creates a new ESKSequence without special preferences.
		/// </summary>
		/// <remarks>No remarks</remarks>
		public ESKSequence() {
			pPackets = new Packet[0];
            alSymKeys = new List<SymSessionKeyPacket>();
            alAsymKeys = new List<AsymSessionKeyPacket>();
            alAllKeys = new List<Packet>();
		}
		
		/// <summary>
		/// Parses an ESK Sequence out of the given array of packets.
		/// In this special case, the first packet in packets MUST be
		/// either a Symmetrically Encrypted Sessionkey Packet or an 
		/// Asymmetrically Encrypted Sessionkey Packet.
		/// </summary>
		/// <returns>Returns the number of packets used by the 
		/// function.</returns>
		/// <param name="packets">Array of packets. The first packet in
		/// the array MUST be either a Symmetrically Encrypted 
		/// Sessionkey Packet or an Asymmetrically Encrypted Sessionkey 
		/// Packet. Otherwise an exception is thrown.</param>
		/// <exception cref="System.Exception">Throws an ordinary
		/// Exception of the given sequence of packets is not an
		/// ESK Sequence.</exception>
		/// <remarks>No remarks</remarks>
		public override int ParseMessage(Packet[] packet) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            Packet wkPacket = new Packet();

			int iSessionKeyCount = 0;
			// First packets must be either Symmetric Sessionkey Packets
			// or Public Key Encrypted Sessionkey Packets and we can have
			// quite a number of them
			while ((packet[iSessionKeyCount] is SymSessionKeyPacket) ||
			       (packet[iSessionKeyCount] is AsymSessionKeyPacket)) {
                       wkPacket = packet[iSessionKeyCount];
				if (packet[iSessionKeyCount] is SymSessionKeyPacket)
                    alSymKeys.Add((SymSessionKeyPacket)packet[iSessionKeyCount]);
				
				if (packet[iSessionKeyCount] is AsymSessionKeyPacket)
					alAsymKeys.Add((AsymSessionKeyPacket)packet[iSessionKeyCount]);
				
				alAllKeys.Add((Packet)packet[iSessionKeyCount]);
				
				iSessionKeyCount++;
			}
			
			// we still have no valid session key
            if (iSessionKeyCount == 0)
                throw new Exception(messageLoader.GetString("ESKNoESKSequence")); 
				//throw new Exception("This is no ESK sequence!");
			
			return iSessionKeyCount;
			
		}
		
		/// <summary>
		/// Adds an symmetrically encrypted session key to the ESK
		/// Sequence.
		/// </summary>
		/// <param name="sskpKey">A symmetrical Session key packet that
		/// is to be added the the ESKSequence.</param>
		/// <remarks>No remarks</remarks>
		public void AddSymSessionKey(SymSessionKeyPacket sskpKey) {
			bUpdated = true;
			alSymKeys.Add(sskpKey);
		}
		
		/// <summary>
		/// Adds an asymmetrically encrypted session key to the ESK
		/// Sequence.
		/// </summary>
		/// <param name="askpKey">An asymmetrical session key packet 
		/// that is to be added to the ESKSequence.</param>
		/// <remarks>No remarks</remarks>
		public void AddAsymSessionKey(AsymSessionKeyPacket askpKey) {
			bUpdated = true;
			alAsymKeys.Add(askpKey);
		}
		
		/// <summary>
		/// Gets the OpenPGP encoded representation of the ESK Sequence.
		/// </summary>
		/// <returns>Returns a byte array that represents the encoded
		/// ESKSequence.</returns>
		/// <remarks>No remarks</remarks>
		public override byte[] GetEncoded() {
			byte[] bOutput = new byte[0];
			
			if (!bUpdated) {
				//nothing was updated, we can reconstruct the message
				//in the exact same order
				IEnumerator ieKeys = alAllKeys.GetEnumerator();
				while (ieKeys.MoveNext()) {
					Packet pKey = (Packet)ieKeys.Current;
					byte[] bKey = pKey.Generate();
					byte[] bOldOutput = new byte[bOutput.Length];
					bOutput.CopyTo(bOldOutput, 0);
					bOutput = new byte[bOldOutput.Length + bKey.Length];
					
					bOldOutput.CopyTo(bOutput, 0);
					bKey.CopyTo(bOutput, bOldOutput.Length);
				}
			} else {
				// At first we will produce the Symmetrically encrypted
				// session key packets
				IEnumerator ieSymKeys = alSymKeys.GetEnumerator();
				while (ieSymKeys.MoveNext()) {
					SymSessionKeyPacket sskpKey = (SymSessionKeyPacket)ieSymKeys.Current;
					byte[] bKey = sskpKey.Generate();
					byte[] bOldOutput = new byte[bOutput.Length];
					bOutput.CopyTo(bOldOutput, 0);
					bOutput = new byte[bOldOutput.Length + bKey.Length];
					
					bOldOutput.CopyTo(bOutput, 0);
					bKey.CopyTo(bOutput, bOldOutput.Length);
				}
				
				// Now come the Public Key encrypted session key packets
				IEnumerator ieAsymKeys = alAsymKeys.GetEnumerator();
				while (ieAsymKeys.MoveNext()) {
					AsymSessionKeyPacket askpKey = (AsymSessionKeyPacket)ieAsymKeys.Current;
					byte[] bKey = askpKey.Generate();
					byte[] bOldOutput = new byte[bOutput.Length];
					bOutput.CopyTo(bOldOutput, 0);
					bOutput = new byte[bOldOutput.Length + bKey.Length];
					
					bOldOutput.CopyTo(bOutput, 0);
					bKey.CopyTo(bOutput, bOldOutput.Length);
				}
				
			}
				
			
			return bOutput;
		}
		
	}
	
}
		
