//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** EncryptedMessage.cs:                                         ***//
//*** 	Class for handling encrypted messages.                     ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    02.03.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp.Messages.               ***//
//***                Changed so that it works in WinRT.            ***//
//***                Changed so that it handles MDC processing     ***//
//***                Changed for myPrimaryKeylKeyID processing     ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
//
using System;
using Hermes.Crypto.Cipher;
using System.Collections;


namespace Hermes.Crypto.OpenPGP.Messages {
	
	/// <summary>
	/// The class encrypted message is a representation of the
	/// openpgp message "Encrypted Message". It contains an
	/// ESK Sequence as well as an symmetrically encrypted
	/// data packet.
	/// This class provides means to handle such messages.
	/// </summary>
	/// <remarks>
	/// The class encrypted message is a representation of the
	/// openpgp message "Encrypted Message". It contains an
	/// ESK Sequence as well as an symmetrically encrypted
	/// data packet.
	/// This class provides means to handle such messages.
	/// </remarks>
	public class EncryptedMessage : Message {
		private ESKSequence esKeys;
		private SymmetricallyEncryptedDataPacket sepData;
        private SymEncryptedIntegrityProtectedDataPacket seippData;
        private ModificationDetectionPacket mdcPackage;
        private byte[] mdcHash;
        private byte[] mdcRandom;
        private byte[] mdcData;

		/// <summary>
		/// Readonly. Returns true if the message was symmetrically
		/// encrypted. A message is symmentrically encrypted, if the
		/// according ESK Sequence does not contains asymmetrically
		/// encrypted session keys.
		/// </summary>
		/// <value>True if the message was symmetrically
		/// encrypted. A message is symmentrically encrypted, if the
		/// according ESK Sequence does not contains asymmetrically
		/// encrypted session keys.</value>
		/// <remarks>No remarks</remarks>
		public bool SymmetricallyEncrypted {
			get {
				if (esKeys.AsymKeys.Count == 0)
					return true;
				return false;
			}
		}
		
		/// <summary>
		/// Creates a new encrypted message.
		/// </summary>
		/// <remarks>No remarks</remarks>
		public EncryptedMessage() {
			esKeys = new ESKSequence();
		}
		
		/// <summary>
		/// Parses an Encrypted Message out of the given array of packets.
		/// In this special case, the first packets MUST be
		/// either Symmetrically Encrypted Sessionkey Packets or  
		/// Asymmetrically Encrypted Sessionkey Packets, or a mixture.
		/// After one more more sessionkey packet, a symmetrically encrypted
		/// data packet must follow.
		/// </summary>
		/// <returns>Returns the number of packets used by the 
		/// function.</returns>
		/// <param name="packets">Array of packets. The first packets MUST be
		/// either Symmetrically Encrypted Sessionkey Packets or  
		/// Asymmetrically Encrypted Sessionkey Packets, or a mixture.
		/// After one more more sessionkey packet, a symmetrically encrypted
		/// data packet must follow. Otherwise an exception is thrown.</param>
		/// <exception cref="System.Exception">Throws an ordinary
		/// Exception of the given sequence of packets is not an
		/// Encrypted Message.</exception>
		/// <remarks>No remarks</remarks>
		public override int ParseMessage(Packet[] packets) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			int iPos = esKeys.ParseMessage(packets);

            if (packets[iPos] is SymmetricallyEncryptedDataPacket)
                sepData = (SymmetricallyEncryptedDataPacket)packets[iPos];
            else if (packets[iPos] is SymEncryptedIntegrityProtectedDataPacket)
                seippData = (SymEncryptedIntegrityProtectedDataPacket)packets[iPos];
            else
                throw new Exception(messageLoader.GetString("EncMsgNoSymEncDataPkt"));
                //throw new System.ArgumentException("Expected a symmetrically encrypted data packet, but did not find one!");


			//if (!(packets[iPos] is SymmetricallyEncryptedDataPacket))
			//	throw new System.ArgumentException("Expected a symmetrically encrypted data packet, but did not find one!");
			
			//sepData = (SymmetricallyEncryptedDataPacket)packets[iPos];
			
			return ++iPos;
			
		}
		
		/// <summary>
		/// Gets the OpenPGP encoded representation of the encrypted
		/// message.
		/// </summary>
		/// <returns>Returns a byte array that contains the binary
		/// representation of the encrypted message.</returns>
		/// <remarks>No remarks</remarks>
		public override byte[] GetEncoded() {
			byte[] bOutput1 = esKeys.GetEncoded();
            byte[] bOutput2 = null;
            if (sepData != null)
            {
                bOutput2 = sepData.Generate();
            }
            if (seippData != null)
            {
                bOutput2 = seippData.Generate();
            }
			
			byte[] bOutput = new byte[bOutput1.Length + bOutput2.Length];
			bOutput1.CopyTo(bOutput, 0);
			bOutput2.CopyTo(bOutput, bOutput1.Length);
			
			return bOutput;
		}
		
		/// <summary>
		/// Finds a secret key out of the given secret keyring that is able
		/// to decrypt the current encrypted message and returns its KeyID.
		/// If such a key is not found, 0 is returned.
		/// </summary>
		/// <param name="skrRing">Secret keyring containing all secret keys
		/// known to the system.</param>
		/// <returns>Returns the KeyID of the key that is able to decrypt the
		/// encrypted message.</returns>
		/// <remarks>No remarks</remarks>
		public ulong GetFittingKeyID(SecretKeyRing skrRing) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			bool bFound = false;
			
			IEnumerator ieSessionkeys = esKeys.AsymKeys.GetEnumerator();
			while (ieSessionkeys.MoveNext()) {
				if (!(ieSessionkeys.Current is AsymSessionKeyPacket))
					throw new Exception("Strange Error!");
				
				AsymSessionKeyPacket askpKey = (AsymSessionKeyPacket)ieSessionkeys.Current;
				ulong lKeyID = askpKey.KeyID;
				
				TransportableSecretKey tskKey = skrRing.Find(lKeyID);
				if (tskKey != null) {
					return lKeyID;
				}
			}

            if (!bFound)
            {
                string message = messageLoader.GetString("EncMsgNoSecKey");
                throw new Exception(message);
                //throw new Exception("No fitting secret key was found to decrypt the message!");
            }
			
			return 0;
		}
		
		/// <summary>
		/// Decrypts the current encrypted message using the secret keys
		/// in skrKeyRing and the given passphrase.
		/// </summary>
		/// <param name="skrKeyRing">The secret keyring containing all the
		/// secret keys know to the sytem.</param>
		/// <param name="strPassphrase">The passphrase that was used to
		/// encrypt the secret key material in the key that decrypts
		/// the message.</param>
        /// <param name="myPrimaryKeylKeyID">The ulong key id value of the
        /// user's primary, or default, key. This key is always include in 
        /// the list of keys used to encrypt the message, so the user can
        /// decrypt it and read it later.</param>
		/// <returns>Returns the message that was encrypted. Usually this is
		/// an compressed or literal message.</returns>
		/// <remarks>No remarks</remarks>
        public Message Decrypt(SecretKeyRing skrKeyRing, string strPassphrase, ulong myPrimaryKeylKeyID = 0)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			TransportableSecretKey tskSecretKey = new TransportableSecretKey();
			AsymSessionKeyPacket askpSessionKey = new AsymSessionKeyPacket();
			bool bFound = false;
			
			// let's see, if we can find a fitting Sessionkey packet
			IEnumerator ieSessionkeys = esKeys.AsymKeys.GetEnumerator();
			while (ieSessionkeys.MoveNext()) {
				if (!(ieSessionkeys.Current is AsymSessionKeyPacket))
					throw new Exception("Strange Error!");
				
				AsymSessionKeyPacket askpKey = (AsymSessionKeyPacket)ieSessionkeys.Current;
				ulong lKeyID = askpKey.KeyID;
				
				TransportableSecretKey tskKey = skrKeyRing.Find(lKeyID);
				if (tskKey != null) {
					bFound = true;
					tskSecretKey = tskKey;
					askpSessionKey = askpKey;
                    // If the user has define a "Primary Key" and it is found, use it
                    if (lKeyID == myPrimaryKeylKeyID)
                        break;
                    if (tskSecretKey.PrimaryKey.PublicKey.KeyID == myPrimaryKeylKeyID)
                        break;
				}
			}

            if (!bFound)
            {
                string message = messageLoader.GetString("EncMsgNoSecKey");
                throw new Exception(message);
                //throw new Exception("No fitting secret key was found to decrypt the message!");
            }
			
			askpSessionKey.DecryptSessionKey(tskSecretKey, strPassphrase);
			byte[] bKey = askpSessionKey.SessionKey;
			
			Packet[] pContent = new Packet[0];
			try {
                if (sepData != null)
                {
                    SymmetricAlgorithm saAlgo = CipherHelper.CreateSymAlgorithm(askpSessionKey.SymmetricAlgorithm);
                    pContent = sepData.Decrypt(bKey, saAlgo);
                }
                if (seippData != null)
                {
                    //byte[] wkB = new byte[seippData.Body.Length - 1];
                    //Array.Copy(seippData.Body, 1, wkB, 0, seippData.Body.Length - 1);
                    //mdcRandom = wkB;



                    SymmetricAlgorithm saAlgo = CipherHelper.CreateSymAlgorithm(askpSessionKey.SymmetricAlgorithm);
                    pContent = seippData.Decrypt(bKey, saAlgo);

                    // Here seems to be a good place to check the mdc package   JRS!!!!!!!!!!!
                    // The mdc package appears to be 30 bytes before the end of the message.
                    // However, I thought the total length of the mdc package was 22 bytes, so I don't
                    // know what the other 8 bytes are.  I think I'll just make a copy of the mdc and 
                    // not remove it from the integrity protected package. The uncompress routine should
                    // take remove it for us later.

                    byte[] wkData= new byte[pContent[0].Header.Length + pContent[0].Body.Length];

                    Array.Copy(pContent[0].Header, 0, wkData, 0, pContent[0].Header.Length);
                    Array.Copy(pContent[0].Body, 0, wkData, pContent[0].Header.Length, pContent[0].Body.Length);


                    // If we have a Packet with an absolute data lenght in the header, then the parse routine
                    // will find the MDC packet and put it in the pContent array.  If the Packet has an infinite
                    // data length then the parse routine will return the MDC packet in the compressed packet and
                    // we must find it ourselves.
                    // If the number of packets in the pContent array is greater the 1 then we probably have 
                    // a compressed packet with an absolute length, and the second packet returned is probably
                    // the MDC packet.  The MDC packet header is 211, 20

                    if (pContent.Length > 1)
                    {
                        if (pContent[1].Header[0] == 211 && pContent[1].Header[1] == 20)
                        {
                            mdcData = wkData;
                            mdcRandom = seippData.RandomData;
                            mdcHash = pContent[1].Body;

                        } 
                    }
                    else
                    {
                        bool notFound = true;
                        int x;

                        try
                        {
                            for (x = wkData.Length - 1; notFound; x--)
                            {
                                if (wkData[x] == 211 && wkData[x + 1] == 20)
                                    notFound = false;
                            }

                            x++;       // reposition x to account for last x-- in previous loop
                            mdcData = new byte[x];
                            Array.Copy(wkData, 0, mdcData, 0, x);  // copy compressed text 
                            x += 2;    // reposition x to skip header
                            byte[] mdcTemp = new byte[20];
                            Array.Copy(wkData, x, mdcTemp, 0, 20);
                            mdcRandom = seippData.RandomData;
                            mdcHash = mdcTemp;

                        }
                        catch
                        {
                            mdcHash = new byte[1];
                            mdcHash[0] = 0xFF;
                            mdcData = new byte[1];
                            mdcData[0] = 0xFF;
                        }
                    }
            
                }
			} catch (Exception e) {
                throw new Exception(messageLoader.GetString("EncMsgDecryptFailed") + e.Message);
                //throw new System.Exception("Decryption of the Message failed: " + e.Message);
			}
			
			// now we need to look what kind of message was hidden in the
			// encrypted data
			
            

			// it can be either a literal message
			LiteralMessage lmLiteral = new LiteralMessage();
			try {
				int iPos = lmLiteral.ParseMessage(pContent);


				return lmLiteral;
			} catch (Exception) {}
			
			// or an compressed Message
			CompressedMessage cmCompressed = new CompressedMessage();
			try {
				int iPos = cmCompressed.ParseMessage(pContent);

                
				return cmCompressed;
			} catch (Exception) {}
            throw new Exception(messageLoader.GetString("EncMsgNotValidMsg"));
            //throw new System.ArgumentException("Encrypted package content is not a valid message!");
		}
		
		/// <summary>
		/// Decrypts the encrypted message if it is a symmetrically encrypted 
		/// message with the passphrase given as argument.
		/// </summary>
		/// <param name="strPassphrase">The passphrase that was used to encrypt
		/// the message</param>
		/// <returns>Returns the message that was encrypted. Usually this is
		/// an compressed or literal message.</returns>
		/// <remarks>No remarks</remarks>
		public Message Decrypt(string strPassphrase) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			if (esKeys.SymKeys.Count == 0)
                throw new Exception(messageLoader.GetString("EncMsgNotSymEnc"));
                //throw new Exception("This message is not symmetrically encrypted. Please provide a keyring rather than a passphrase!");
			
			Packet[] pContent = new Packet[0];
			Packet[] pReturn = new Packet[0];
			
			IEnumerator ieKeys = esKeys.SymKeys.GetEnumerator();
			while (ieKeys.MoveNext()) {
				SymSessionKeyPacket skpKey = (SymSessionKeyPacket)ieKeys.Current;
				byte[] key = skpKey.S2KSpecifier.GetKey(strPassphrase, CipherHelper.CipherKeySize(skpKey.Algorithm));

				try {
					SymmetricAlgorithm saAlgo = CipherHelper.CreateSymAlgorithm(skpKey.Algorithm);
					pContent = sepData.Decrypt(key, saAlgo);
				} 
                //catch (System.Security.Cryptography.CryptographicException) {}
                catch (System.Exception) { }
				if (pContent.Length > 0) {
					pReturn = pContent;
				}
			}

            if (pReturn.Length == 0)
                //throw new System.Security.Cryptography.CryptographicException("Wrong passphrase!");
                throw new Exception(messageLoader.GetString("EncMsgWrongPassPhrase"));
                //throw new System.Exception("EncryptedMessage: Wrong passphrase!");
			
			// now we need to look what kind of message was hidden in the
			// encrypted data
			
			// it can be either a literal message
			LiteralMessage lmLiteral = new LiteralMessage();
			try {
				int iPos = lmLiteral.ParseMessage(pReturn);
				return lmLiteral;
			} catch (Exception) {}
			
			// or an compressed Message
			CompressedMessage cmCompressed = new CompressedMessage();
			try {
				int iPos = cmCompressed.ParseMessage(pReturn);
				return cmCompressed;
			} catch (Exception) {}
            throw new Exception(messageLoader.GetString("EncMsgNotValidMsg"));
            //throw new System.ArgumentException("Encrypted package content is not a valid message!");
		}


        public byte[] MDCRandom
        {
            get
            {
                return mdcRandom;
            }
            set
            {
                mdcRandom = value;
            }
        }

        public byte[] MDCHash
        {
            get
            {
                return mdcHash;
            }
            set
            {
                mdcHash = value;
            }
        }

        public byte[] MDCData
        {
            get
            {
                return mdcData;
            }
            set
            {
                mdcData = value;
            }
        }
	}
}
