﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.                                       ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** ModificationDetectionPacket.cs:                              ***//
//*** 	Class for handling modified detection packets.             ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    31.10.2013  James R. Stembridge - Created this module.    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher;

namespace Hermes.Crypto.OpenPGP
{

    public class ModificationDetectionPacket : Packet
    {



        /// <summary>
        /// Creates a new SModificationDetectionPacket with 
        /// the parameters in pSource
        /// </summary>
        /// <param name="pSource">Packet from which the
        /// parameters are derived</param>
        public ModificationDetectionPacket(Packet pSource)
        {
            lLength = pSource.Length;
            bBody = pSource.Body;
            ctContent = pSource.Content;
            pfFormat = pSource.Format;
            bHeader = pSource.Header;
            this.bIsUpdated = false;
        }

        /// <summary>
        /// Creates a new ModificationDetectionPacket. 
        /// Format defaults to new packet format.
        /// </summary>
        public ModificationDetectionPacket()
        {
            bBody = new byte[20];
            bHeader = new byte[2];
            pfFormat = PacketFormats.New;
            ctContent = ContentTypes.SymEncryptIntegrity;
            this.bIsUpdated = true;
        }

        /// <summary>
        /// Creates a new ModificationDetectionPacket
        /// Load binary data into packet
        /// </summary>
        /// <param name="binaryData"></param>
        public ModificationDetectionPacket(byte[] binaryData)
        {
            bBody = new byte[20];
            bHeader = new byte[2];
            pfFormat = PacketFormats.New;
            ctContent = ContentTypes.SymEncryptIntegrity;
            this.bIsUpdated = true;
            Array.Copy(binaryData, 0, bHeader, 0, 2);
            Array.Copy(binaryData, 2, bBody, 0, 20);

        }

        /// <summary>
        /// Returns a string representation of the packet. This is
        /// a human readable formated representation that has nothing
        /// to do with OpenPGP or RFC2440
        /// </summary>
        /// <returns>String representation of the packet.</returns>
        /// <remarks>No remarks</remarks>
        public override string ToString()
        {
            string strReturn = "Modification Detection Packet:\r\n";
            for (int i = 0; i < this.bBody.Length; i++)
            {
                string strByte = bBody[i].ToString("x");
                if (strByte.Length < 2)
                {
                    strByte = "0" + strByte;
                }
                strReturn += ":" + strByte;
            }
            strReturn += "\r\n----\r\n\r\n";
            return strReturn;
        }

        /// <summary>
        /// Parses the packet given as byte array into the current
        /// class and returns this with the populated parameters.
        /// </summary>
        /// <param name="bData">A byte array containing an OpenPGP
        /// representation of the packet.</param>
        /// <returns>Returns an ModificationDetectionPacket
        ///  that containes the parsed properties.</returns>
        /// <remarks>No remarks</remarks>
        public override Packet ParsePacket(byte[] bData)
        {
            //nothing to be done. the body is the encrypted data.
            this.bIsUpdated = false;
            return this;
        }

        // we don't need no craftcontent!

        /// <summary>
        /// Sets or gets the Header
        /// </summary>
        /// <param name="bHeader">A byte array containing an OpenPGP
        /// representation of the packet header.</param>
        /// <returns>Returns an ModificationDetectionPacket header
        ///  that containes the parsed properties.</returns>
        /// <remarks>No remarks</remarks>
        public byte[] Header
        {
            get
            {
                return bHeader;
            }
            set
            {
                this.bIsUpdated = true;
                bHeader = value;
            }
        }

        /// <summary>
        /// Sets or gets the Body
        /// </summary>
        /// <param name="bBody">A byte array containing an OpenPGP
        /// representation of the packet body.</param>
        /// <returns>Returns an ModificationDetectionPacket body
        ///  that containes the parsed properties.</returns>
        /// <remarks>No remarks</remarks>
        public byte[] Body
        {
            get
            {
                return bBody;
            }
            set
            {
                this.bIsUpdated = true;
                bBody = value;
            }
        }


    }

}
