//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** String2KeySpecifier.cs:                                      ***//
//*** 	Class for handling string to key specifiers.               ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    18.03.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    14.06.2003: Changed Namespace from                        ***//
//***                SharpPrivacy.OpenPGP to                       ***//
//***                SharpPrivacy.SharpPrivacyLib.OpenPGP          ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp                         ***//
//***                Changed so that it works in WinRT.            ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
//using System.Security.Cryptography;
using Windows.Storage.Streams;
using Windows.Security.Cryptography;
using Windows.Security.Cryptography.Core;
using System.Runtime.InteropServices.WindowsRuntime;

using Hermes.Crypto.Cipher;

namespace Hermes.Crypto.OpenPGP
{

    public class String2KeySpecifier : object
    {
        private String2KeySpecifierTypes s2kType;
        private HashAlgorithms haHashAlgorithm;
        private ulong lSalt;
        private byte bCount;

        public String2KeySpecifier(byte[] bSpecifier)
        {
            ParseSpecifier(bSpecifier);
        }

        public String2KeySpecifier()
        {

        }

        public HashAlgorithms HashAlgorithm
        {
            get
            {
                return haHashAlgorithm;
            }
            set
            {
                haHashAlgorithm = value;
            }
        }

        public String2KeySpecifierTypes Type
        {
            get
            {
                return s2kType;
            }
            set
            {
                s2kType = value;
            }
        }

        public ulong Salt
        {
            get
            {
                return lSalt;
            }
            set
            {
                lSalt = value;
            }
        }

        public byte Count
        {
            get
            {
                return bCount;
            }
            set
            {
                bCount = value;
            }
        }

        public byte[] CraftContent()
        {
            byte[] bData = new byte[0];
            if (this.Type == String2KeySpecifierTypes.SimpleS2K)
            {
                bData = new byte[2];
                bData[0] = 0x00;
                bData[1] = (byte)this.HashAlgorithm;
            }
            else if (this.Type == String2KeySpecifierTypes.SimpleS2K)
            {
                bData = new byte[10];
                bData[0] = (byte)this.Type;
                bData[1] = (byte)this.HashAlgorithm;

                bData[2] = (byte)((lSalt >> 56) & 0xFF);
                bData[3] = (byte)((lSalt >> 48) & 0xFF);
                bData[4] = (byte)((lSalt >> 40) & 0xFF);
                bData[5] = (byte)((lSalt >> 32) & 0xFF);
                bData[6] = (byte)((lSalt >> 24) & 0xFF);
                bData[7] = (byte)((lSalt >> 16) & 0xFF);
                bData[8] = (byte)((lSalt >> 8) & 0xFF);
                bData[9] = (byte)(lSalt & 0xFF);
            }
            else
            {
                bData = new byte[11];
                bData[0] = (byte)this.Type;
                bData[1] = (byte)this.HashAlgorithm;

                bData[2] = (byte)((lSalt >> 56) & 0xFF);
                bData[3] = (byte)((lSalt >> 48) & 0xFF);
                bData[4] = (byte)((lSalt >> 40) & 0xFF);
                bData[5] = (byte)((lSalt >> 32) & 0xFF);
                bData[6] = (byte)((lSalt >> 24) & 0xFF);
                bData[7] = (byte)((lSalt >> 16) & 0xFF);
                bData[8] = (byte)((lSalt >> 8) & 0xFF);
                bData[9] = (byte)(lSalt & 0xFF);
                bData[10] = this.bCount;
            }

            return bData;
        }

        /// <summary>
        /// Returns a string representation of the String2Key specifier.
        /// This is a human readable formated representation that has 
        /// nothing to do with OpenPGP or RFC2440
        /// </summary>
        /// <returns>String representation of the specifier.</returns>
        /// <remarks>No remarks</remarks>
        public override string ToString()
        {
            string strReturn = "";

            if (this.Type == String2KeySpecifierTypes.IteraterSaltedS2K)
            {
                strReturn += "Iterated and Salted String2Key Specifier:\r\n";
            }
            else if (this.Type == String2KeySpecifierTypes.SaltedS2K)
            {
                strReturn += "Salted String2Key Specifier:\r\n";
            }
            else if (this.Type == String2KeySpecifierTypes.SimpleS2K)
            {
                strReturn += "Simple String2Key Specifier:\r\n";
            }

            strReturn += "Salt: " + this.Salt.ToString("x") + "\r\n";
            strReturn += "Iterationcount: " + this.Count.ToString() + "\r\n\r\n";

            return strReturn;
        }

        public int Length(byte bType)
        {
            String2KeySpecifierTypes s2kstType = (String2KeySpecifierTypes)bType;
            switch (s2kstType)
            {
                case String2KeySpecifierTypes.SimpleS2K:
                    return 2;
                case String2KeySpecifierTypes.SaltedS2K:
                    return 10;
                case String2KeySpecifierTypes.IteraterSaltedS2K:
                    return 11;
            }

            return 0;
        }

        public void ParseSpecifier(byte[] bSpecifier)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            if (bSpecifier.Length < 2)
            {
                throw new Exception(messageLoader.GetString("Str2KeyNotValidS2KSpec"));
                //throw(new System.ArgumentException("Not a valid String2Key specifier!"));
            }
            this.Type = (String2KeySpecifierTypes)bSpecifier[0];
            this.HashAlgorithm = (HashAlgorithms)bSpecifier[1];

            if (this.Type == String2KeySpecifierTypes.SaltedS2K ||
                this.Type == String2KeySpecifierTypes.IteraterSaltedS2K)
            {
                if (bSpecifier.Length < 10)
                {
                    throw new Exception(messageLoader.GetString("Str2KeyNotValidS2KSpec"));
                    //throw(new System.ArgumentException("Not a valid String2Key specifier!"));
                }
                lSalt = ((ulong)bSpecifier[2]) << 56;
                lSalt ^= ((ulong)bSpecifier[3]) << 48;
                lSalt ^= ((ulong)bSpecifier[4]) << 40;
                lSalt ^= ((ulong)bSpecifier[5]) << 32;
                lSalt ^= ((ulong)bSpecifier[6]) << 24;
                lSalt ^= ((ulong)bSpecifier[7]) << 16;
                lSalt ^= ((ulong)bSpecifier[8]) << 8;
                lSalt ^= ((ulong)bSpecifier[9]);
            }

            if (this.Type == String2KeySpecifierTypes.IteraterSaltedS2K)
            {
                if (bSpecifier.Length < 11)
                {
                    throw new Exception(messageLoader.GetString("Str2KeyNotValidS2KSpec"));
                    //throw(new System.ArgumentException("Not a valid String2Key specifier!"));
                }
                bCount = bSpecifier[10];
            }
        }

        public byte[] GetKey(string strPassphrase, int nKeySize)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            //System.Security.Cryptography.HashAlgorithm haHash;
            string hashAlg = "MD5";
            switch (this.HashAlgorithm)
            {        // Replace with new Windows 8 Hash Algorithm stuff
                case HashAlgorithms.MD5:
                    //haHash = MD5.Create();
                    hashAlg = HashAlgorithmNames.Md5;
                    break;
                case HashAlgorithms.SHA1:
                    //haHash = SHA1.Create();
                    hashAlg = HashAlgorithmNames.Sha1;
                    break;
                default:
                    throw new Exception(messageLoader.GetString("Str2KeyBadHashAlgo"));
                //throw(new Exception("Currently only MD5 and SHA1 are implemented as Hash algorithms!"));
            }

            // Create a HashAlgorithmProvider object.
            HashAlgorithmProvider objAlgProv = HashAlgorithmProvider.OpenAlgorithm(hashAlg);

            byte[] bPassphrase = System.Text.Encoding.UTF8.GetBytes(strPassphrase);  //Convert pass phrase

            int iPosition = 0;
            int iCount = 0;
            byte[] bReturn = new byte[(nKeySize + 7) / 8];
            while (iPosition < bReturn.Length)
            {
                int iOffset = 0;
                byte[] bHashContext = new byte[0];
                if (this.Type == String2KeySpecifierTypes.SimpleS2K)
                {
                    bHashContext = new byte[bPassphrase.Length + iCount];
                }
                else
                {
                    iOffset = 8;
                    bHashContext = new byte[bPassphrase.Length + iCount + iOffset];
                }


                for (int i = 0; i < iCount; i++)
                {
                    bHashContext[i] = 0;
                }

                if (this.Type == String2KeySpecifierTypes.SaltedS2K ||
                    this.Type == String2KeySpecifierTypes.IteraterSaltedS2K)
                {
                    bHashContext[iCount] = (byte)((lSalt >> 56) & 0xFF);
                    bHashContext[iCount + 1] = (byte)((lSalt >> 48) & 0xFF);
                    bHashContext[iCount + 2] = (byte)((lSalt >> 40) & 0xFF);
                    bHashContext[iCount + 3] = (byte)((lSalt >> 32) & 0xFF);
                    bHashContext[iCount + 4] = (byte)((lSalt >> 24) & 0xFF);
                    bHashContext[iCount + 5] = (byte)((lSalt >> 16) & 0xFF);
                    bHashContext[iCount + 6] = (byte)((lSalt >> 8) & 0xFF);
                    bHashContext[iCount + 7] = (byte)(lSalt & 0xFF);
                }

                Array.Copy(bPassphrase, 0, bHashContext, iCount + iOffset, bPassphrase.Length);

                if (this.Type == String2KeySpecifierTypes.IteraterSaltedS2K)
                {
                    uint lIterationCount = (uint)((16 + (bCount & 15))) << ((bCount >> 4) + 6);
                    if (lIterationCount < bPassphrase.Length + 8)
                    {
                        lIterationCount = (uint)bHashContext.Length;
                    }
                    byte[] bOldHashContext = new byte[bHashContext.Length];
                    Array.Copy(bHashContext, bOldHashContext, bHashContext.Length);
                    bHashContext = new byte[lIterationCount + iCount];
                    Array.Copy(bOldHashContext, bHashContext, bOldHashContext.Length);

                    byte[] bOneIteration = new byte[bOldHashContext.Length - iCount];
                    Array.Copy(bOldHashContext, iCount, bOneIteration, 0, bOldHashContext.Length - iCount);

                    int iIterationPos = bOldHashContext.Length;
                    while (iIterationPos < (bHashContext.Length))
                    {
                        if (bHashContext.Length < (iIterationPos + bOneIteration.Length))
                        {
                            Array.Copy(bOneIteration, 0, bHashContext, iIterationPos, bHashContext.Length - iIterationPos);
                        }
                        else
                        {
                            Array.Copy(bOneIteration, 0, bHashContext, iIterationPos, bOneIteration.Length);
                        }
                        iIterationPos += bOneIteration.Length;
                    }
                }
                //byte[] bHash = haHash.ComputeHash(bHashContext); // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                //String strEncodedHash = this.ComputeHash(hashAlg, bHashContext);
                //byte[] bHash = this.ComputeHash(hashAlg, bHashContext);

                IBuffer buffMsg = bHashContext.AsBuffer();
                // Hash the message.
                IBuffer buffHash = objAlgProv.HashData(buffMsg);

                // Verify that the hash length equals the length specified for the algorithm.
                if (buffHash.Length != objAlgProv.HashLength)
                {
                    throw new Exception(messageLoader.GetString("Str2KeyHashError"));
                    //throw new Exception("There was an error creating the hash");
                }

                // Convert the IBuffer (hash) to a Byte Array
                Byte[] bHash = buffHash.ToArray();


                if (bHash.Length > (bReturn.Length - iPosition))
                {
                    Array.Copy(bHash, 0, bReturn, iPosition, bReturn.Length - iPosition);
                }
                else
                {
                    Array.Copy(bHash, 0, bReturn, iPosition, bHash.Length);
                }
                iPosition += bHash.Length;
                iCount++;
            }

            return bReturn;
        }

        //public String ComputeHash(String strAlgName, String strMsg)
        public Byte[] ComputeHash(String strAlgName, Byte[] strMsg)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            // Convert the message string to binary data.
            //IBuffer buffUtf8Msg = CryptographicBuffer.ConvertStringToBinary(strMsg, BinaryStringEncoding.Utf8);
            IBuffer buffUtf8Msg = strMsg.AsBuffer();

            // Create a HashAlgorithmProvider object.
            HashAlgorithmProvider objAlgProv = HashAlgorithmProvider.OpenAlgorithm(strAlgName);

            // Demonstrate how to retrieve the name of the hashing algorithm.
            String strAlgNameUsed = objAlgProv.AlgorithmName;

            // Hash the message.
            IBuffer buffHash = objAlgProv.HashData(buffUtf8Msg);

            // Verify that the hash length equals the length specified for the algorithm.
            if (buffHash.Length != objAlgProv.HashLength)
            {
                throw new Exception(messageLoader.GetString("Str2KeyHashError"));
                //throw new Exception("There was an error creating the hash");
            }

            // Convert the hash to a string (for display).
            //String strHashBase64 = CryptographicBuffer.EncodeToBase64String(buffHash);

            // Return the encoded string
            //return strHashBase64;

            // Convert the IBuffer (hash) to a Byte Array
            Byte[] returnBuffer = buffHash.ToArray();

            // return the hash
            return returnBuffer;
        }


    }

}
