﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.                                       ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** SymEncryptedIntegrityProtectedDataPacket.cs:                 ***//
//*** 	Class for handling symmetrically encrypted data packets.   ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    31.10.2013  James R. Stembridge - Created this module.    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher;

namespace Hermes.Crypto.OpenPGP {

	public class SymEncryptedIntegrityProtectedDataPacket : Packet {
		
		/* This file is intentionally empty.
		 * The only proptery that a symmetrically
		 * encrypted data packet contains is the
		 * encrypted data. It is stored in the
		 * Body property of this class
		 */


        private byte[] randomData;

		/// <summary>
		/// Creates a new SymEncryptedIntegrityProtectedDataPacket with 
		/// the parameters in pSource
		/// </summary>
		/// <param name="pSource">Packet from which the
		/// parameters are derived</param>
		public SymEncryptedIntegrityProtectedDataPacket(Packet pSource) {
			lLength = pSource.Length;
			bBody = pSource.Body;
			ctContent = pSource.Content;
			pfFormat = pSource.Format;
			bHeader = pSource.Header;
			this.bIsUpdated = false;
		}
		
		/// <summary>
		/// Creates a new SymEncryptedIntegrityProtectedDataPacket. 
		/// Format defaults to new packet format.
		/// </summary>
        public SymEncryptedIntegrityProtectedDataPacket()
        {
            bBody = new byte[0];
            bHeader = new byte[0];
            pfFormat = PacketFormats.New;
            ctContent = ContentTypes.SymEncryptIntegrity;
            this.bIsUpdated = true;
        }


        public byte[] RandomData
        {
            get
            {
                return randomData;
            }
            
        }


        /// <summary>
		/// Returns a string representation of the packet. This is
		/// a human readable formated representation that has nothing
		/// to do with OpenPGP or RFC2440
		/// </summary>
		/// <returns>String representation of the packet.</returns>
		/// <remarks>No remarks</remarks>
		public override string ToString() {
			string strReturn = "Symmetrically Encrypted Integrity Protected Data Packet:\r\n";
			for (int i=0; i<this.bBody.Length; i++) {
				string strByte = bBody[i].ToString("x");
				if (strByte.Length < 2) {
					strByte = "0" + strByte;
				}
				strReturn += ":" + strByte;
			}
			strReturn += "\r\n----\r\n\r\n";
			return strReturn;
		}
		
		/// <summary>
		/// Parses the packet given as byte array into the current
		/// class and returns this with the populated parameters.
		/// </summary>
		/// <param name="bData">A byte array containing an OpenPGP
		/// representation of the packet.</param>
		/// <returns>Returns an SymmetricallyEncryptedDataPacket
		///  that containes the parsed properties.</returns>
		/// <remarks>No remarks</remarks>
		public override Packet ParsePacket(byte[] bData) {
			//nothing to be done. the body is the encrypted data.
			this.bIsUpdated = false;
			return this;
		}
		
		// we don't to craftcontent!
		
		public Packet[] Decrypt(byte[] bKey, SymmetricAlgorithm saAlgo) {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
			saAlgo.Mode = CipherMode.OpenPGP_CFB;
            //saAlgo.Mode = CipherMode.CFB;
			saAlgo.Key = bKey;
            saAlgo.IntegrityProtected = true;
			ICryptoTransform ictDecrypt = saAlgo.CreateDecryptor();
			int iBS = saAlgo.BlockSize >> 3;

            // A Summetrically Encrypted Integrity Protected data packet body has a first octet
            // that is the version number of the packet format, whereas an old format Symmetrically 
            // Encrypted data packet does not.  Therefore, we need to remove the first octet
            // (the version number octet) from the packet before we call the transform routine.
            byte[] newBody = new byte[bBody.Length - 1];
            Array.Copy(bBody, 1, newBody, 0, bBody.Length-1);

			//int iLength = bBody.Length - iBS - 2;
            int iLength = newBody.Length - iBS - 2;

			byte[] bOutput = new byte[bBody.Length-1];
			ictDecrypt.TransformBlock(newBody, 0, newBody.Length, ref bOutput, 0);

            int x;
            for (x = bOutput.Length ; bOutput[x-1] == 0; x--)
            {
            }

            int BlockSizeByte = (saAlgo.BlockSize >> 3);
            x -= BlockSizeByte + 2; // decrease buffer length by the length of the random data (IV)

			byte[] bFinal = new byte[x];
			//Array.Copy(bOutput, 18, bFinal, 0, iLength - 22);
            Array.Copy(bOutput, BlockSizeByte + 2, bFinal, 0, x);
            randomData = new byte[BlockSizeByte + 2];
            Array.Copy(bOutput, randomData, BlockSizeByte + 2);
            if ((randomData[BlockSizeByte - 2] != randomData[BlockSizeByte]) || (randomData[BlockSizeByte - 1] != randomData[BlockSizeByte + 1]))
                throw new Exception(messageLoader.GetString("SymEncIntegProtWrongKey"));
                //throw new System.Exception("SymEncryptedIntegrityProtected: Wrong Key!!");
			Packet[] pReturn = Packet.ParsePackets(bFinal);
			return pReturn;
		}
		
	}
	
}
