﻿//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright © 2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** SymmetricallyEncryptedDataPacket.cs:                         ***//
//*** 	Class for handling symmetrically encrypted data packets.   ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    18.03.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    14.06.2003: Changed Namespace from                        ***//
//***                SharpPrivacy.OpenPGP to                       ***//
//***                SharpPrivacy.SharpPrivacyLib.OpenPGP          ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp                         ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher;

namespace Hermes.Crypto.OpenPGP
{

    public class SymmetricallyEncryptedDataPacket : Packet
    {

        /* This file is intentionally empty.
         * The only proptery that a symmetrically
         * encrypted data packet contains is the
         * encrypted data. It is stored in the
         * Body property of this class
         */

        /// <summary>
        /// Creates a new SymmetricallyEncryptedDataPacket with 
        /// the parameters in pSource
        /// </summary>
        /// <param name="pSource">Packet from which the
        /// parameters are derived</param>
        public SymmetricallyEncryptedDataPacket(Packet pSource)
        {
            lLength = pSource.Length;
            bBody = pSource.Body;
            ctContent = pSource.Content;
            pfFormat = pSource.Format;
            bHeader = pSource.Header;
            this.bIsUpdated = false;
        }

        /// <summary>
        /// Creates a new symmetrically encrypted data packet. 
        /// Format defaults to new packet format.
        /// </summary>
        public SymmetricallyEncryptedDataPacket()
        {
            bBody = new byte[0];
            bHeader = new byte[0];
            pfFormat = PacketFormats.New;
            ctContent = ContentTypes.SymEncrypted;
            this.bIsUpdated = true;
        }

        /// <summary>
        /// Returns a string representation of the packet. This is
        /// a human readable formated representation that has nothing
        /// to do with OpenPGP or RFC2440
        /// </summary>
        /// <returns>String representation of the packet.</returns>
        /// <remarks>No remarks</remarks>
        public override string ToString()
        {
            string strReturn = "Symmetrically Encrypted Data Packet:\r\n";
            for (int i = 0; i < this.bBody.Length; i++)
            {
                string strByte = bBody[i].ToString("x");
                if (strByte.Length < 2)
                {
                    strByte = "0" + strByte;
                }
                strReturn += ":" + strByte;
            }
            strReturn += "\r\n----\r\n\r\n";
            return strReturn;
        }

        /// <summary>
        /// Parses the packet given as byte array into the current
        /// class and returns this with the populated parameters.
        /// </summary>
        /// <param name="bData">A byte array containing an OpenPGP
        /// representation of the packet.</param>
        /// <returns>Returns an SymmetricallyEncryptedDataPacket
        ///  that containes the parsed properties.</returns>
        /// <remarks>No remarks</remarks>
        public override Packet ParsePacket(byte[] bData)
        {
            //nothing to be done. the body is the encrypted data.
            this.bIsUpdated = false;
            return this;
        }

        // we don't need no craftcontent!

        public Packet[] Decrypt(byte[] bKey, SymmetricAlgorithm saAlgo)
        {

            saAlgo.Mode = CipherMode.OpenPGP_CFB;
            saAlgo.Key = bKey;
            ICryptoTransform ictDecrypt = saAlgo.CreateDecryptor();
            int iBS = saAlgo.BlockSize >> 3;
            int iLength = bBody.Length - iBS - 2;
            byte[] bOutput = new byte[bBody.Length];
            ictDecrypt.TransformBlock(bBody, 0, bBody.Length, ref bOutput, 0);
            byte[] bFinal = new byte[iLength];
            Array.Copy(bOutput, bFinal, iLength);

            Packet[] pReturn = Packet.ParsePackets(bFinal);
            return pReturn;
        }

    }

}
