//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** PublicKeyRing.cs:                                            ***//
//*** 	Class for managing the Public Key Ring.                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    07.04.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    14.06.2003: Changed Namespace from                        ***//
//***                SharpPrivacy.OpenPGP to                       ***//
//***                SharpPrivacy.SharpPrivacyLib.OpenPGP          ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp                         ***//
//***                Changed so that it works in WinRT.            ***//
//***                Added a sort routine                          ***//
//***                Added an export routine                       ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.Cipher.Math;
using Hermes.Crypto.OpenPGP;
using Hermes.Crypto.OpenPGP.Messages;
using System.Collections;
using System.Collections.Generic;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.Storage.Pickers;
using System.Threading.Tasks;

namespace Hermes.Crypto
{

    /// <summary>
    /// Class to manage the Public KeyRing 
    /// </summary>
    public class PublicKeyRing
    {

        private List<TransportablePublicKey> listPublicKeys;
        //private bool bIsUpdated = false;
        //private string strLoadingPath;

        //public bool IsUpdated {
        //	get {
        //		return bIsUpdated;
        //	}
        //}

        /// <summary>
        /// Gets or sets the key list
        /// </summary>
        public List<TransportablePublicKey> PublicKeys
        {
            get
            {
                return listPublicKeys;
            }
            set
            {
                listPublicKeys = value;
            }
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public PublicKeyRing()
        {
            listPublicKeys = new List<TransportablePublicKey>();
        }

        /// <summary>
        /// Loads the keyring
        /// </summary>
        /// <param name="keyLocation">The keyring file location (localstore, filesystem), or a string of keys.</param>
        /// <param name="initial">Boolian indicator for initial (clean) load vs. add to (import) load</param>
        public async Task Load(string keyLocation, bool initial = true)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            string keyString = string.Empty;
            string errMsg = string.Empty;

            if (keyLocation == "localstore")
            {
                var _pkrName = "publickeyring.asc";
                var _Folder = Windows.Storage.ApplicationData.Current.RoamingFolder;
                //var _Option = Windows.Storage.CreationCollisionOption.ReplaceExisting;
                var _Option = Windows.Storage.CreationCollisionOption.FailIfExists;
                bool fileExist = true;
                try
                {
                    var testFile = await _Folder.CreateFileAsync(_pkrName, _Option);
                    fileExist = false;
                }
                catch
                {
                    fileExist = true;
                }
                if (!fileExist)
                {
                    //MessageDialog dialog = new MessageDialog("No Public Key Ring exists. One will be created.\nImport public keys using the key ring manager.", "Public Key Ring");
                    errMsg = messageLoader.GetString("NoPubKeyRingExists1") + "\n" + messageLoader.GetString("NoPubKeyRingExists2");
                }
                // acquire file
                var _ReadFile = await _Folder.GetFileAsync(_pkrName);
                // read content
                var _ReadThis = await Windows.Storage.FileIO.ReadTextAsync(_ReadFile);
                keyString = _ReadThis;
            }
            else if (keyLocation == "filesystem")
            {
                // define picker
                var _Picker = new FileOpenPicker
                {
                    ViewMode = PickerViewMode.List,
                    SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                };
                _Picker.FileTypeFilter.Add("*");
                // let user pick file
                var _File = await _Picker.PickSingleFileAsync();
                if (_File == null)
                {
                    errMsg = messageLoader.GetString("NoFileSelected");
                    throw new Exception(errMsg);
                    //await new Windows.UI.Popups.MessageDialog("No file").ShowAsync();
                }
                // read content
                var _Content = await Windows.Storage.FileIO.ReadTextAsync(_File);
                //await new Windows.UI.Popups.MessageDialog(_Content).ShowAsync();
                keyString = _Content;
            }
            else
            {
                keyString = keyLocation;
            }

            if (initial)
                this.PublicKeys = new List<TransportablePublicKey>();

            ArmorTypes atType = new ArmorTypes();
            string strKey = Armor.RemoveArmor(keyString, ref atType, ref keyString);
            while (strKey.Length > 0)
            {
                TransportablePublicKey[] transPubKeys = TransportablePublicKey.SplitKeys(strKey);
                foreach (TransportablePublicKey transPubKey in transPubKeys)
                {
                    this.AddKey(transPubKey);
                }
                strKey = Armor.RemoveArmor(keyString, ref atType, ref keyString);
            }
            //this.bIsUpdated = false;

            Sort();

            if (errMsg != string.Empty)
                throw new Exception(errMsg);

            return;
        }

        /// <summary>
        /// Sorts the public keyring by primary userid
        /// </summary>
        public void Sort()
        {
            listPublicKeys.Sort(delegate(TransportablePublicKey pk1, TransportablePublicKey pk2)
            {
                return pk1.PrimaryUserID.CompareTo(pk2.PrimaryUserID);
            });
        }

        /// <summary>
        /// Saves the keyring to a specific location
        /// </summary>
        /// <param name="keyLocation">The keyring file location (localstore, backup), 
        /// backup uses file system.</param>
        public async Task Save(string keyLocation)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            // Build the output buffer
            var _outPut = "";
            IEnumerator ieKeys = this.PublicKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (ieKeys.Current is TransportablePublicKey)
                {
                    try
                    {
                        TransportablePublicKey transPubKey = (TransportablePublicKey)ieKeys.Current;
                        byte[] bKey = transPubKey.Generate();
                        string strKey = Armor.WrapPublicKey(bKey);
                        _outPut += strKey;
                    }
                    catch (Exception e)
                    {
                        string message = messageLoader.GetString("ErrorSavingPubKey") + e.Message;
                        throw new Exception(message);
                        //MessageBox.Show("Error while trying to save a public key: " + e.Message, "Error...", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }
                }
            }

            if (keyLocation == "localstore")
            {
                var _pkrName = "publickeyring.asc";
                var _Folder = Windows.Storage.ApplicationData.Current.RoamingFolder;
                var _Option = Windows.Storage.CreationCollisionOption.ReplaceExisting;

                var _pkrFile = await _Folder.CreateFileAsync(_pkrName, _Option);
                await Windows.Storage.FileIO.WriteTextAsync(_pkrFile, _outPut);
                //this.bIsUpdated = false;
            }
            else if (keyLocation == "backup")
            {
                string strKey;
                string myUserName = string.Empty;
                string myKeyID = string.Empty;
                string myUserID = string.Empty;
                string myTime = DateTime.Now.ToString();
                string fileNameToUse = string.Empty;

                fileNameToUse = "PKR_Backup_" + myTime;
                fileNameToUse = fileNameToUse.Replace(":", "_");
                fileNameToUse = fileNameToUse.Replace(" ", "_");
                fileNameToUse = fileNameToUse.Replace("/", "_");
                fileNameToUse = fileNameToUse.Replace(".", "_");
                fileNameToUse += ".asc";
                var savePicker = new FileSavePicker
                {
                    SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                    SuggestedFileName = fileNameToUse
                };

                savePicker.FileTypeChoices.Add("Armored", new List<string>() { ".asc" });
                savePicker.FileTypeChoices.Add("Text", new List<string>() { ".txt" });

                StorageFile file2 = await savePicker.PickSaveFileAsync();

                if (file2 != null)
                    try
                    {
                        await FileIO.WriteTextAsync(file2, _outPut);
                    }
                    catch (Exception e)
                    {
                        string message = messageLoader.GetString("ErrorSavingPubKey") + e.Message;
                        throw new Exception(message);
                        //throw new Exception("Error while trying to save the public key ring: " + e.Message); ;
                    }
                else
                {
                    string message = messageLoader.GetString("ErrorSavingPubKeyOutputFile");
                    throw new Exception(message);
                    //throw new Exception("Error while trying to save the public key ring: unable to allocate output file");
                }
            }
            else
            {
                string message = messageLoader.GetString("ErrorSavingPubKeyInvalidKeyLoc");
                throw new Exception(message);
            }
        }

        /// <summary>
        /// Export a key to a location
        /// </summary>
        /// <param name="KeyID">key to export</param>
        public async Task Export(ulong KeyID)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            string strKey;
            string myUserName = string.Empty;
            string myKeyID = string.Empty;
            string myUserID = string.Empty;
            string myTime = DateTime.Now.ToString();
            try
            {
                TransportablePublicKey transPubKey = this.Find(KeyID, false);
                byte[] bKey = transPubKey.Generate();
                strKey = Armor.WrapPublicKey(bKey);
                //string strWork = transPubKey.PrimaryUserID;
                //myUserID = strWork;
                //string[] idWork1 = strWork.Split('<');
                //if (idWork1.Length > 1)
                //{
                //    myUserName = idWork1[0].Trim();
                //}
                //else
                //{
                //    myUserName = strWork.Trim();
                //}
                myUserName = transPubKey.PrimaryUserID.Trim();
                myKeyID = transPubKey.PrimaryKey.KeyID.ToString("X");
            }
            catch (Exception e)
            {
                string message = messageLoader.GetString("ErrorExportingPubKey") + e.Message;
                throw new Exception(message);
                //throw new Exception("Error while trying to export a public key: " + e.Message);
            }

            string fileNameToUse = string.Empty;

            fileNameToUse = myUserName.Trim() + "_" + myKeyID.Trim() + "_" + myTime;
            fileNameToUse = fileNameToUse.Replace(":", "_");
            fileNameToUse = fileNameToUse.Replace(" ", "_");
            fileNameToUse = fileNameToUse.Replace("/", "_");
            fileNameToUse = fileNameToUse.Replace(".", "_");
            fileNameToUse = fileNameToUse.Replace("<", "_");
            fileNameToUse = fileNameToUse.Replace(">", "_");

            fileNameToUse += "_pkey.asc";
            fileNameToUse = fileNameToUse.Replace("__", "_");
            var savePicker = new FileSavePicker
            {
                SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                SuggestedFileName = fileNameToUse
            };

            savePicker.FileTypeChoices.Add("Armored", new List<string>() { ".asc" });
            savePicker.FileTypeChoices.Add("Text", new List<string>() { ".txt" });

            StorageFile file = await savePicker.PickSaveFileAsync();

            if (file != null)
                try
                {
                    await FileIO.WriteTextAsync(file, strKey);
                }
                catch (Exception e)
                {
                    string message = messageLoader.GetString("ErrorExportingPubKey") + e.Message;
                    throw new Exception(message);
                    //throw new Exception("Error while trying to export a public key: " + e.Message); ;
                }
            else
            {
                string message = messageLoader.GetString("ErrorExportingPubKeyOutputFile");
                throw new Exception(message);
                //throw new Exception("Error while trying to export a public key: unable to allocate output file"); ;
            }
        }

        /// <summary>
        /// Add a key to the keyring
        /// </summary>
        /// <param name="tspk">the key to be added</param>
        public void AddPublicKey(TransportablePublicKey tspk)
        {
            if (tspk != null)
            {
                if (this.Find(tspk.PrimaryKey.KeyID, false) == null)
                {
                    this.AddKey(tspk);
                }
            }
        }

        /// <summary>
        /// Private method to add a key doing checks
        /// </summary>
        /// <param name="transPubKey">the key to be saved</param>
        private void AddKey(TransportablePublicKey transPubKey)
        {
            //bIsUpdated = true;
            TransportablePublicKey local = null;
            if ((local = this.Find(transPubKey.PrimaryKey.KeyID, false)) == null)
            {
                listPublicKeys.Add(transPubKey);
            }
            else
            {
                TransportablePublicKey joinKey = new TransportablePublicKey();
                joinKey.PrimaryKey = local.PrimaryKey;
                //Revocations
                List<SignaturePacket> toBeAdded = new List<SignaturePacket>();
                toBeAdded.AddRange(local.RevocationSignatures);
                toBeAdded.AddRange(transPubKey.RevocationSignatures);
                foreach (SignaturePacket localpacket in local.RevocationSignatures)
                {
                    foreach (SignaturePacket packet in transPubKey.RevocationSignatures)
                    {
                        if (localpacket.Body == packet.Body && localpacket.Header == packet.Header)
                        {
                            toBeAdded.Remove(packet);
                            continue;
                        }
                    }
                }
                joinKey.RevocationSignatures = toBeAdded;

                //Revokers
                toBeAdded = new List<SignaturePacket>();
                toBeAdded.AddRange(local.RevocationKeys);
                toBeAdded.AddRange(transPubKey.RevocationKeys);
                foreach (SignaturePacket localpacket in local.RevocationKeys)
                {
                    foreach (SignaturePacket packet in transPubKey.RevocationKeys)
                    {
                        if (localpacket.Body == packet.Body && localpacket.Header == packet.Header)
                        {
                            toBeAdded.Remove(packet);
                            continue;
                        }
                    }
                }
                joinKey.RevocationKeys = toBeAdded;

                //CERTIFICATES
                List<CertifiedUserID> certPkgToBeAdded = new List<CertifiedUserID>();
                certPkgToBeAdded.AddRange(local.Certifications);
                certPkgToBeAdded.AddRange(transPubKey.Certifications);
                foreach (CertifiedUserID localpacket in local.Certifications)
                {
                    foreach (CertifiedUserID packet in transPubKey.Certifications)
                    {
                        if (localpacket.UserID == packet.UserID)
                        {
                            //ArrayList certificatesToBeAdded = new ArrayList();
                            List<SignaturePacket> certificatesToBeAdded = new List<SignaturePacket>();
                            certificatesToBeAdded.AddRange(localpacket.Certificates);
                            certificatesToBeAdded.AddRange(packet.Certificates);
                            foreach (SignaturePacket signatureLocal in localpacket.Certificates)
                            {
                                foreach (SignaturePacket signature in localpacket.Certificates)
                                {
                                    if (signatureLocal.Header == signature.Header && signatureLocal.Body == signature.Body)
                                    {
                                        certificatesToBeAdded.Remove(signature);
                                        continue;
                                    }
                                }
                            }
                            localpacket.Certificates = certificatesToBeAdded;
                            certPkgToBeAdded.Remove(packet);
                            continue;
                        }
                    }
                }
                joinKey.Certifications = certPkgToBeAdded;

                //SUBKEYS
                List<CertifiedPublicSubkey> subKeyToBeAdded = new List<CertifiedPublicSubkey>();
                subKeyToBeAdded.AddRange(local.SubKeys);
                subKeyToBeAdded.AddRange(transPubKey.SubKeys);
                foreach (CertifiedPublicSubkey localpacket in local.SubKeys)
                {
                    foreach (CertifiedPublicSubkey packet in transPubKey.SubKeys)
                    {
                        if (localpacket.Subkey.KeyID == packet.Subkey.KeyID)
                        {
                            subKeyToBeAdded.Remove(packet);
                            if (localpacket.KeyBindingSignature == null)
                                localpacket.KeyBindingSignature = packet.KeyBindingSignature;
                            if (localpacket.RevocationSignature == null)
                                localpacket.RevocationSignature = packet.RevocationSignature;
                            continue;
                        }
                    }
                }
                joinKey.SubKeys = subKeyToBeAdded;

                listPublicKeys.Remove(local);
                listPublicKeys.Add(joinKey);
            }
        }

        /// <summary>
        /// Removes a key from the ring
        /// </summary>
        /// <param name="lKeyID">the key to remove</param>
        public void Delete(ulong lKeyID)
        {
            //bIsUpdated = true;
            listPublicKeys.Remove(Find(lKeyID, false));
        }

        /// <summary>
        /// Removes a key from the ring
        /// </summary>
        /// <param name="transPubKey">the key to remove</param>
        public void Delete(TransportablePublicKey transPubKey)
        {
            //bIsUpdated = true;
            listPublicKeys.Remove(transPubKey);
        }

        /// <summary>
        /// Finds a Key given a keyid. Performs a remote LDAP search if specified.
        /// </summary>
        /// <param name="lKeyID">Key to find</param>
        /// <param name="remote">LDAP search</param>
        /// <returns>a key</returns>
        public TransportablePublicKey Find(ulong lKeyID, bool remote)
        {
            IEnumerator ieKeys = listPublicKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                TransportablePublicKey transPubKey = (TransportablePublicKey)ieKeys.Current;
                if (transPubKey.PrimaryKey.KeyID == lKeyID)
                {
                    return transPubKey;
                }
                IEnumerator ieSubkeys = transPubKey.SubKeys.GetEnumerator();
                while (ieSubkeys.MoveNext())
                {
                    CertifiedPublicSubkey certPubSubkey = (CertifiedPublicSubkey)ieSubkeys.Current;
                    if (certPubSubkey.Subkey.KeyID == lKeyID)
                        return transPubKey;
                }
            }

            /*
                        if (remote) {
                            ldapKeyFinder.KeyFinder kf = new ldapKeyFinder.KeyFinder();
				
                            string key = kf.MyLDAPSearch(SharpPrivacyLibrary.LdapKeyServer, SharpPrivacyLibrary.LdapPort,"pgpkey","(pgpsignerid="+lKeyID.ToString("X")+")");
                            if (key != null) {
                                ArmorTypes atType = new ArmorTypes();
                                string strKey = Armor.RemoveArmor(key, ref atType, ref key);
                                if (strKey.Length > 0) {
                                    TransportablePublicKey transPubKey = new TransportablePublicKey(strKey);
                                    AddPublicKey(transPubKey);
                                    return transPubKey;
                                }
                            }
                        }
            */
            return null;
        }

        //public TransportablePublicKey[] FindLdap(ulong KeyID) {
        //	
        //}

        /// <summary>
        /// Find a list of keys which contains the result of the query done using userID as argument.
        /// </summary>
        /// <param name="userID">User ID contained in the keys to list</param>
        /// <returns>a list of keys</returns>
        public List<TransportablePublicKey> FindPublicKeysByID(string userID)
        {
            List<TransportablePublicKey> pkList = new List<TransportablePublicKey>();
            IEnumerator ieKeys = this.PublicKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (!(ieKeys.Current is TransportablePublicKey))
                {
                    continue;
                }
                TransportablePublicKey key = ((TransportablePublicKey)ieKeys.Current);
                if (key.PrimaryUserID.IndexOf(userID) >= 0)
                {
                    pkList.Add(key);
                }
            }
            return pkList;
        }
        /*
                /// <summary>
                /// Find a remote list of keys which contains the result of the query done using userID as argument.
                /// </summary>
                /// <param name="userID">User ID contained in the keys to list</param>
                /// <returns>a list of keys</returns>
                public List<TransportablePublicKey> FindRemotePublicKeysByUserID(string userID)
                {
                    List<TransportablePublicKey> pkList = new List<TransportablePublicKey>();
                    ldapKeyFinder.KeyFinder kf = new ldapKeyFinder.KeyFinder();	
                    byte result = kf.MyLDAPSearchByID(SharpPrivacyLibrary.LdapKeyServer,SharpPrivacyLibrary.LdapPort,"pgpkeyid","(pgpuserid=*"+userID+"*)");
                    if(result == 1) {
                        string[] keys = kf.getKeys();
                        pkList = new List<TransportablePublicKey>(keys);
                    }
                    return pkList;
                }
        */
        /// <summary>
        /// Finds a remote public key radix64 encoded and server stored
        /// </summary>
        /// <param name="keyID">the keyid of the key as hexadecimal code</param>
        /// <returns>a public key radix64 encoded and PGP armored</returns>
        //public string FindRemotePublicKeyByKeyID(string keyID) {
        /*
                    string key = null;
                    ldapKeyFinder.KeyFinder kf = new ldapKeyFinder.KeyFinder();	
                    key = kf.MyLDAPSearch(SharpPrivacyLibrary.LdapKeyServer,SharpPrivacyLibrary.LdapPort,"pgpkey","(pgpkeyid="+keyID+")");
                    if(key.Length > 0) {
                        return key;
                    }
         */
        //	return null;
        //}

        /// <summary>
        /// Finds a local or remote key given its fingerprint
        /// </summary>
        /// <param name="fingerprint">fingerprint</param>
        /// <returns>a public key</returns>
        public TransportablePublicKey FindPublicKey(string fingerprint)
        {
            IEnumerator ieKeys = this.PublicKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (!(ieKeys.Current is TransportablePublicKey))
                {
                    continue;
                }
                TransportablePublicKey key = ((TransportablePublicKey)ieKeys.Current);
                if (key.PrimaryKey.Fingerprint.ToString() == fingerprint)
                {
                    return key;
                }
            }
            /*			
                        ldapKeyFinder.KeyFinder kf = new ldapKeyFinder.KeyFinder();	
                        string remoteKey = kf.MyLDAPSearch(SharpPrivacyLibrary.LdapKeyServer,SharpPrivacyLibrary.LdapPort,"pgpkey","(pgpsignerid="+fingerprint+")");
                        if (remoteKey != null) {
                            ArmorTypes atType = new ArmorTypes();
                            string strKey = Armor.RemoveArmor(remoteKey, ref atType, ref remoteKey);
                            if (strKey.Length > 0) {
                                TransportablePublicKey transPubKey = new TransportablePublicKey(strKey);
                                AddPublicKey(transPubKey);
                                return transPubKey;
                            }
                        }
            */
            return null;
        }

        /// <summary>
        /// Verifies key signatures
        /// </summary>
        /// <param name="keyID">the key to be verified</param>
        /// <returns>
        /// list of invalid signatures or objects containing invalid signatures:
        /// SignaturePacket
        /// CertifiedUserID
        /// CertifiedPublicSubkey
        /// </returns>
        public List<SignaturePacket> verifySignatures(ulong keyID)
        {
            return this.verifySignatures(this.Find(keyID, true));
        }

        /// <summary>
        /// Verifies key signatures
        /// </summary>
        /// <param name="transPubKey">the key to be verified</param>
        /// <returns>
        /// list of invalid signatures or objects containing invalid signatures:
        /// SignaturePacket
        /// CertifiedUserID
        /// CertifiedPublicSubkey
        /// </returns>
        private List<SignaturePacket> verifySignatures(TransportablePublicKey transPubKey)
        {
            List<SignaturePacket> signaturesNotOK = new List<SignaturePacket>();

            //Verify Revocation Signatures
            foreach (SignaturePacket revocation in transPubKey.RevocationSignatures)
            {
                if (revocation.KeyID == transPubKey.PrimaryKey.KeyID)
                {
                    PublicKeyPacket pkpKey = transPubKey.FindKey(revocation.KeyID);
                    byte[] key = new byte[transPubKey.PrimaryKey.Length];
                    transPubKey.PrimaryKey.Header.CopyTo(key, 0);
                    transPubKey.PrimaryKey.Body.CopyTo(key, transPubKey.PrimaryKey.Header.Length);
                    if (pkpKey == null)
                    {
                        revocation.SignatureStatus = SignatureStatusTypes.Signing_Key_Not_Available;
                        signaturesNotOK.Add(revocation);
                    }
                    else
                    {
                        revocation.Verify(key, pkpKey);
                        if (revocation.SignatureStatus == SignatureStatusTypes.Invalid ||
                            revocation.SignatureStatus == SignatureStatusTypes.Not_Verified ||
                            revocation.SignatureStatus == SignatureStatusTypes.Signing_Key_Not_Available)
                        {
                            signaturesNotOK.Add(revocation);
                        }
                    }
                }
                else
                {
                    TransportablePublicKey revTransPubKey = this.Find(revocation.KeyID, true);
                    if (revTransPubKey != null)
                    {
                        foreach (SignaturePacket spPacket in transPubKey.RevocationKeys)
                        {
                            foreach (BigInteger revoker in spPacket.FindRevokerKeys())
                            {
                                if (revoker.ToString() == revTransPubKey.PrimaryKey.Fingerprint.ToString())
                                {
                                    PublicKeyPacket pkpKey = revTransPubKey.PrimaryKey;
                                    byte[] key = new byte[transPubKey.PrimaryKey.Length];
                                    transPubKey.PrimaryKey.Header.CopyTo(key, 0);
                                    transPubKey.PrimaryKey.Body.CopyTo(key, transPubKey.PrimaryKey.Header.Length);
                                    revocation.Verify(key, pkpKey);
                                    if (revocation.SignatureStatus == SignatureStatusTypes.Invalid ||
                                        revocation.SignatureStatus == SignatureStatusTypes.Not_Verified ||
                                        revocation.SignatureStatus == SignatureStatusTypes.Signing_Key_Not_Available)
                                    {
                                        signaturesNotOK.Add(revocation);
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        revocation.SignatureStatus = SignatureStatusTypes.Signing_Key_Not_Available;
                        signaturesNotOK.Add(revocation);
                    }
                }
            }

            //Verify UserID
            foreach (CertifiedUserID userId in transPubKey.Certifications)
            {
                userId.Validate(transPubKey.PrimaryKey, this);
                if (userId.CertificationValidityStatus == CertifiedUserID.ValidityStatus.Invalid ||
                    userId.CertificationValidityStatus == CertifiedUserID.ValidityStatus.NotYetValidated ||
                    userId.CertificationValidityStatus == CertifiedUserID.ValidityStatus.ValidationKeyUnavailable)
                {
                    foreach (SignaturePacket sp in userId.Certificates)
                    {
                        if (sp.SignatureStatus != SignatureStatusTypes.Valid)
                            signaturesNotOK.Add(sp);
                    }
                }
            }

            foreach (CertifiedPublicSubkey certPubSubKey in transPubKey.SubKeys)
            {
                if (certPubSubKey.KeyBindingSignature == null)
                {
                    //signaturesNotOK.Add(certPubSubKey);  // CHECK THIS OUT - JRS - DON'T FULLY UNDERSTAND!!!!!!!
                }
                else
                {
                    certPubSubKey.VerifyKeyBindingSignature(transPubKey.PrimaryKey);
                    if (certPubSubKey.KeyBindingSignature.SignatureStatus == SignatureStatusTypes.Invalid ||
                        certPubSubKey.KeyBindingSignature.SignatureStatus == SignatureStatusTypes.Not_Verified ||
                        certPubSubKey.KeyBindingSignature.SignatureStatus == SignatureStatusTypes.Signing_Key_Not_Available)
                    {
                        signaturesNotOK.Add(certPubSubKey.KeyBindingSignature);
                    }
                }

                //Verify Subkey Revocation Signature
                SignaturePacket revocation = certPubSubKey.RevocationSignature;
                if (revocation != null)
                {
                    if (revocation.KeyID == transPubKey.PrimaryKey.KeyID)
                    {
                        byte[] subkey = new byte[certPubSubKey.Subkey.Length];
                        certPubSubKey.Subkey.Header.CopyTo(subkey, 0);
                        certPubSubKey.Subkey.Body.CopyTo(subkey, certPubSubKey.Subkey.Header.Length);
                        subkey[0] = 0x99;

                        byte[] mainkey = new byte[transPubKey.PrimaryKey.Length];
                        transPubKey.PrimaryKey.Header.CopyTo(mainkey, 0);
                        transPubKey.PrimaryKey.Body.CopyTo(mainkey, transPubKey.PrimaryKey.Header.Length);

                        byte[] key = new byte[subkey.Length + mainkey.Length];
                        mainkey.CopyTo(key, 0);
                        subkey.CopyTo(key, mainkey.Length);

                        revocation.Verify(key, transPubKey.PrimaryKey);
                        if (revocation.SignatureStatus == SignatureStatusTypes.Invalid ||
                            revocation.SignatureStatus == SignatureStatusTypes.Not_Verified ||
                            revocation.SignatureStatus == SignatureStatusTypes.Signing_Key_Not_Available)
                        {
                            signaturesNotOK.Add(revocation);
                        }

                    }
                    else
                    {
                        TransportablePublicKey revTransPubKey = this.Find(revocation.KeyID, true);
                        if (revTransPubKey != null)
                        {
                            foreach (SignaturePacket spPacket in transPubKey.RevocationKeys)
                            {
                                foreach (BigInteger revoker in spPacket.FindRevokerKeys())
                                {
                                    if (revoker.ToString() == revTransPubKey.PrimaryKey.Fingerprint.ToString())
                                    {
                                        byte[] subkey = new byte[certPubSubKey.Subkey.Length];
                                        certPubSubKey.Subkey.Header.CopyTo(subkey, 0);
                                        certPubSubKey.Subkey.Body.CopyTo(subkey, certPubSubKey.Subkey.Header.Length);
                                        subkey[0] = 0x99;

                                        byte[] mainkey = new byte[revTransPubKey.PrimaryKey.Length];
                                        transPubKey.PrimaryKey.Header.CopyTo(mainkey, 0);
                                        transPubKey.PrimaryKey.Body.CopyTo(mainkey, revTransPubKey.PrimaryKey.Header.Length);

                                        byte[] key = new byte[subkey.Length + mainkey.Length];
                                        mainkey.CopyTo(key, 0);
                                        subkey.CopyTo(key, mainkey.Length);

                                        revocation.Verify(key, revTransPubKey.PrimaryKey);
                                        if (revocation.SignatureStatus == SignatureStatusTypes.Invalid ||
                                            revocation.SignatureStatus == SignatureStatusTypes.Not_Verified ||
                                            revocation.SignatureStatus == SignatureStatusTypes.Signing_Key_Not_Available)
                                        {
                                            signaturesNotOK.Add(revocation);
                                        }
                                    }
                                }
                            }
                        }
                        else
                        {
                            signaturesNotOK.Add(revocation);
                        }
                    }
                }
            }
            return signaturesNotOK;
        }

        /// <summary>
        /// Verifies a key user id certificetion revocation status
        /// </summary>
        /// <param name="keyID">the key to verify</param>
        /// <param name="userID">the user id to verify</param>
        /// <param name="certifierKeyID">the key that issued the certification</param>
        /// <returns>the revocation status of the user id</returns>
        public bool isRevoked(ulong keyID, string userID, ulong certifierKeyID)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            TransportablePublicKey transPubKey = this.Find(keyID, true);
            if (transPubKey == null)
                return false;
            bool found = false;
            CertifiedUserID toBeVerified = null;
            foreach (CertifiedUserID cui in transPubKey.Certifications)
            {
                if (cui.UserID.UserID == userID)
                {
                    found = true;
                    toBeVerified = cui;
                    break;
                }
            }
            if (!found)
            {

                throw new Exception(messageLoader.GetString("UserIdNotFoundInKeyCerts"));
            }

            toBeVerified.Validate(transPubKey.PrimaryKey, this);
            foreach (SignaturePacket sign in toBeVerified.Certificates)
            {
                if (sign.SignatureType == SignatureTypes.CertificationRevocationSignature && sign.KeyID == certifierKeyID && sign.SignatureStatus == SignatureStatusTypes.Valid && sign.isRevocable())
                {
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Verifies the revocation status of a key
        /// </summary>
        /// <param name="KeyID">the key to verify</param>
        /// <returns>the revocation status of the key</returns>
        public bool isRevoked(ulong KeyID)
        {
            TransportablePublicKey transPubKey = this.Find(KeyID, true);
            if (transPubKey == null)
                return false;
            if (transPubKey.PrimaryKey.KeyID == KeyID)
            {
                List<SignaturePacket> nvsigatures = this.verifySignatures(transPubKey);
                foreach (SignaturePacket sp in transPubKey.PrimaryUserIDCert.Certificates)
                {
                    if (sp.SignatureType == SignatureTypes.UserIDSignature ||
                        sp.SignatureType == SignatureTypes.UserIDSignature_CasualVerification ||
                        sp.SignatureType == SignatureTypes.UserIDSignature_NoVerification ||
                        sp.SignatureType == SignatureTypes.UserIDSignature_PositivVerification)
                    {
                        if (!sp.isRevocable())
                            return false;
                    }
                }
                if (transPubKey.RevocationSignatures == null || transPubKey.RevocationSignatures.Count == 0)
                {
                    return false;
                }
                else
                {
                    foreach (SignaturePacket revocation in transPubKey.RevocationSignatures)
                    {
                        if (revocation.KeyID == transPubKey.PrimaryKey.KeyID)
                        {
                            PublicKeyPacket pkpKey = transPubKey.FindKey(revocation.KeyID);
                            byte[] key = new byte[transPubKey.PrimaryKey.Length];
                            transPubKey.PrimaryKey.Header.CopyTo(key, 0);
                            transPubKey.PrimaryKey.Body.CopyTo(key, transPubKey.PrimaryKey.Header.Length);
                            revocation.Verify(key, pkpKey);
                            if (revocation.SignatureStatus == SignatureStatusTypes.Valid)
                            {
                                return true;
                            }
                            else if (revocation.SignatureStatus == SignatureStatusTypes.Invalid)
                            {
                                continue;
                            }
                            else
                            {
                                continue;
                            }
                        }
                        else
                        {
                            TransportablePublicKey revTransPubKey = this.Find(revocation.KeyID, true);
                            if (revTransPubKey == null)
                                return false;
                            foreach (SignaturePacket spPacket in transPubKey.RevocationKeys)
                            {
                                foreach (BigInteger revoker in spPacket.FindRevokerKeys())
                                {
                                    if (revoker.ToString() == revTransPubKey.PrimaryKey.Fingerprint.ToString())
                                    {
                                        PublicKeyPacket pkpKey = revTransPubKey.PrimaryKey;
                                        byte[] key = new byte[transPubKey.PrimaryKey.Length];
                                        transPubKey.PrimaryKey.Header.CopyTo(key, 0);
                                        transPubKey.PrimaryKey.Body.CopyTo(key, transPubKey.PrimaryKey.Header.Length);
                                        revocation.Verify(key, pkpKey);
                                        if (revocation.SignatureStatus == SignatureStatusTypes.Valid)
                                        {
                                            return true;
                                        }
                                        else if (revocation.SignatureStatus == SignatureStatusTypes.Invalid)
                                        {
                                            continue;
                                        }
                                        else
                                        {
                                            continue;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else
            {
                List<SignaturePacket> signaturesNotOK = this.verifySignatures(transPubKey);
                foreach (CertifiedPublicSubkey certPubSubKey in transPubKey.SubKeys)
                {
                    if (certPubSubKey.Subkey.KeyID == KeyID)
                    {
                        if (certPubSubKey.RevocationSignature != null && !signaturesNotOK.Contains(certPubSubKey.RevocationSignature) && certPubSubKey.KeyBindingSignature.isRevocable())
                        {
                            ulong issuer = certPubSubKey.RevocationSignature.KeyID;
                            if (issuer == transPubKey.PrimaryKey.KeyID)
                            {
                                return true;
                            }
                            else
                            {
                                foreach (SignaturePacket spPacket in transPubKey.RevocationKeys)
                                {
                                    foreach (BigInteger revoker in spPacket.FindRevokerKeys())
                                    {
                                        if (revoker == this.Find(issuer, true).PrimaryKey.Fingerprint)
                                        {
                                            return true;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            return false;
        }
    }

}
