//*** ************************************************************ ***//
//*** ************************************************************ ***//
//*** Hermes Crypto Library                                        ***//
//*** ************************************************************ ***//
//***                                                              ***//
//***            Copyright  2013 James R. Stembridge              ***//
//***                                                              ***//
//*** This file is part of the source code distribution of the     ***//
//*** Hermes Crypto Library.  This work is a derivative work       ***//
//*** based on of the source code distribution of SharpPrivacy.    ***//
//*** SharpPrivacy was written and copyrighted under the Gnu       ***//
//*** General Public License by Daniel Fabian in 2003. The         ***//
//*** original source code implementation of SharpPrivacy was      ***//
//*** found at http://sourceforge.net/projects/sharpprivacy/.      ***//
//*** The SharpPrivacy project seems to be dead now, but the       ***//
//*** source code was still available at the time of this writing. ***//
//***                                                              ***//
//*** The Hermes Crypto Library is an implementation of the        ***//
//*** openPGP standard, as defined by RFC4880, designed to operate ***//
//*** in the WinRT (Metro) environment. The code is designed to    ***//
//*** be compiled as a class library to create a dll that you can  ***//
//*** dynamically invoke from other programs.                      ***//
//***                                                              ***//
//*** This Hermes Crypto Library is free software; you can         ***//
//*** redistribute it and/or modify it under the terms of          ***//
//*** version 2 of the GNU General Public License as published by  ***//
//*** the Free Software Foundation.                                ***//
//***                                                              ***//
//*** This work is distributed in the hope that it will be useful, ***//
//*** but WITHOUT ANY WARRANTY; without even the implied warranty  ***//
//*** of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  ***//
//*** the GNU General Public License for more details. If you      ***//
//*** cannot comply with this FOR ANY REASON, then you do not have ***//
//*** a license to run, copy, modify, or otherwise use this work.  ***//
//***                                                              ***//
//*** You should have received a copy of the GNU General Public    ***//
//*** License along with this program; if not, write to the        ***//
//*** Free Software Foundation, Inc., 59 Temple Place, Suite 330,  ***//
//*** Boston, MA  02111-1307  USA                                  ***//
//*** or visit their web site at                                   ***//
//*** http://www.gnu.org/copyleft/gpl.html.                        ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** SecretKeyRing.cs:                                            ***//
//*** 	Class for managing the Secret Key Ring.                    ***//
//***                                                              ***//
//*** ************************************************************ ***//
//***                                                              ***//
//*** Change Log:                                                  ***//
//***    Version: 0.1.0 (initial release)                          ***//
//***    07.04.2003: Daniel Fabian - Created this file.            ***//
//***    01.06.2003: Daniel Fabian - Added header for the first    ***//
//***                beta release.                                 ***//
//***    14.06.2003: Changed Namespace from                        ***//
//***                SharpPrivacy.OpenPGP to                       ***//
//***                SharpPrivacy.SharpPrivacyLib.OpenPGP          ***//
//***    31.10.2013  James R. Stembridge - Changed Namespace to    ***//
//***                Hermes.Crypto.OpenPgp                         ***//
//***                Changed so that it works in WinRT.            ***//
//***                Added a sort routine                          ***//
//***                Added an export routine                       ***//
//***                                                              ***//
//*** ************************************************************ ***//
//*** ************************************************************ ***//
//
using System;
using Hermes.Crypto.OpenPGP;
using Hermes.Crypto.OpenPGP.Messages;
using System.Collections;
using System.Collections.Generic;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.Storage.Pickers;
using System.Threading.Tasks;

namespace Hermes.Crypto
{

    /// <summary>
    ///  Class to manage a KeyRing containing PGP-Armored radix64 encoded Keys
    /// </summary>
    public class SecretKeyRing
    {

        private List<TransportableSecretKey> listSecretKeys;

        /// <summary>
        /// Gets the key list
        /// </summary>
        public List<TransportableSecretKey> SecretKeys
        {
            get
            {
                return listSecretKeys;
            }
            set
            {
                listSecretKeys = value;
            }
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        public SecretKeyRing()
        {
            listSecretKeys = new List<TransportableSecretKey>();
        }

        /// <summary>
        /// Loads a keyring file
        /// </summary>
        /// <param name="keyLocation">The keyring file location (localstore, filesystem), or a string of keys.</param>
        /// <param name="initial">Boolian indicator for initial (clean) load vs. add to (import) load</param>
        public async Task Load(string keyLocation, bool initial = true)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            string keyString = string.Empty;
            string errMsg = string.Empty;

            if (keyLocation == "localstore")
            {

                var _skrName = "secretkeyring.asc";
                var _Folder = Windows.Storage.ApplicationData.Current.LocalFolder;
                //var _Option = Windows.Storage.CreationCollisionOption.ReplaceExisting;
                var _Option = Windows.Storage.CreationCollisionOption.FailIfExists;
                bool fileExist = true;
                //var _WriteThis = "Hello world!";
                try
                {
                    var testFile = await _Folder.CreateFileAsync(_skrName, _Option);
                    // write content
                    //await Windows.Storage.FileIO.WriteTextAsync(testFile, _WriteThis);
                    fileExist = false;
                }
                catch
                {
                    fileExist = true;
                }

                if (!fileExist)
                {

                    //MessageDialog dialog = new MessageDialog("No Private Secret Key Ring exists. One will be created.\nPlease generate or import a private secret key\nusing the key ring manager.", "Private Secret Key Ring");
                    //MessageDialog dialog = new MessageDialog(messageLoader.GetString("NoSecKeyRingExists1") + "\n" + messageLoader.GetString("NoSecKeyRingExists2") + "\n" + messageLoader.GetString("NoSecKeyRingExists3"), messageLoader.GetString("Info"));
                    //await dialog.ShowAsync();
                    errMsg = messageLoader.GetString("NoSecKeyRingExists1") + "\n" + messageLoader.GetString("NoSecKeyRingExists2") + "\n" + messageLoader.GetString("NoSecKeyRingExists3");
                    throw new Exception(errMsg);
                }
                // acquire file
                var _ReadFile = await _Folder.GetFileAsync(_skrName);
                //System.Diagnostics.Debug.Assert(_ReadFile != null, "Acquire file");

                // read content
                var _ReadThis = await Windows.Storage.FileIO.ReadTextAsync(_ReadFile);
                //System.Diagnostics.Debug.Assert(_WriteThis == _ReadThis, "Contents match");

                keyString = _ReadThis;

            }
            else if (keyLocation == "filesystem")
            {
                // define picker
                var _Picker = new FileOpenPicker
                {
                    ViewMode = PickerViewMode.List,
                    SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                };
                _Picker.FileTypeFilter.Add("*");

                // let user pick file
                var _File = await _Picker.PickSingleFileAsync();
                if (_File == null)
                {
                    errMsg = messageLoader.GetString("NoFileSelected");
                    throw new Exception(errMsg);
                    //await new Windows.UI.Popups.MessageDialog("No file").ShowAsync();
                    return;
                }

                // read content
                var _Content = await Windows.Storage.FileIO.ReadTextAsync(_File);
                //await new Windows.UI.Popups.MessageDialog(_Content).ShowAsync();
                keyString = _Content;

            }
            else
            {
                keyString = keyLocation;
            }

            //string strKeys = strPath;
            if (initial)
                this.SecretKeys = new List<TransportableSecretKey>();

            ArmorTypes atType = new ArmorTypes();
            string strKey = Armor.RemoveArmor(keyString, ref atType, ref keyString);
            while (strKey.Length > 0)
            {
                TransportableSecretKey[] transSecKeys = TransportableSecretKey.SplitKeys(strKey);
                foreach (TransportableSecretKey transSecKey in transSecKeys)
                {
                    this.SecretKeys.Add(transSecKey);
                }

                strKey = Armor.RemoveArmor(keyString, ref atType, ref keyString);
            }
            //bIsUpdated = false;

            Sort();

            //listSecretKeys.Sort(delegate(TransportableSecretKey sk1, TransportableSecretKey sk2)
            //{
            //    return sk1.PrimaryUserID.CompareTo(sk2.PrimaryUserID);
            //});

        }


        /// <summary>
        /// Sorts a keyring file
        /// </summary>
        public void Sort()
        {
            listSecretKeys.Sort(delegate(TransportableSecretKey sk1, TransportableSecretKey sk2)
            {
                return sk1.PrimaryUserID.CompareTo(sk2.PrimaryUserID);
            });

        }

        /// <summary>
        /// Saves the keyring to a specific location
        /// </summary>
        /// <param name="strPath">location to save to</param>
        /// <param name="keyLocation">The keyring file location (localstore, backup), 
        /// backup uses file system.</param>
        public async Task Save(string keyLocation)
        {

            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            var _outPut = "";
            IEnumerator ieKeys = this.SecretKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (ieKeys.Current is TransportableSecretKey)
                {
                    try
                    {
                        TransportableSecretKey transSecKey = (TransportableSecretKey)ieKeys.Current;
                        byte[] bKey = transSecKey.Generate();
                        string strKey = Armor.WrapPrivateKey(bKey);
                        _outPut += strKey;
                        //swOutput.Write(strKey);
                    }
                    catch (Exception e)
                    {
                        string msg = messageLoader.GetString("ErrorSavingSecKey") + e.Message;
                        throw new Exception(msg);
                        // MessageBox.Show("Error while trying to save a secret key: " + e.Message, "Error...", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }
                }
            }

            if (keyLocation == "localstore")
            {
                var _skrName = "secretkeyring.asc";
                var _Folder = Windows.Storage.ApplicationData.Current.LocalFolder;
                var _Option = Windows.Storage.CreationCollisionOption.ReplaceExisting;

                var _skrFile = await _Folder.CreateFileAsync(_skrName, _Option);
                await Windows.Storage.FileIO.WriteTextAsync(_skrFile, _outPut);
                //var _WriteThis = "Hello world!";
                //swOutput.Close();
                //bIsUpdated = false;
            }
            else if (keyLocation == "backup")
            {
                string strKey;
                string myUserName = string.Empty;
                string myKeyID = string.Empty;
                string myUserID = string.Empty;
                string myTime = DateTime.Now.ToString();
                string fileNameToUse = string.Empty;

                fileNameToUse = "SKR_Backup_" + myTime;
                fileNameToUse = fileNameToUse.Replace(":", "_");
                fileNameToUse = fileNameToUse.Replace(" ", "_");
                fileNameToUse = fileNameToUse.Replace("/", "_");
                fileNameToUse = fileNameToUse.Replace(".", "_");
                fileNameToUse += ".asc";
                var savePicker = new FileSavePicker
                {
                    SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                    SuggestedFileName = fileNameToUse
                };

                savePicker.FileTypeChoices.Add("Armored", new List<string>() { ".asc" });
                savePicker.FileTypeChoices.Add("Text", new List<string>() { ".txt" });

                StorageFile file2 = await savePicker.PickSaveFileAsync();

                if (file2 != null)
                    try
                    {
                        await FileIO.WriteTextAsync(file2, _outPut);
                    }
                    catch (Exception e)
                    {
                        string msg = messageLoader.GetString("ErrorSavingSecKey") + e.Message;
                        throw new Exception(msg);
                        //throw new Exception("Error while trying to save the secret key ring: " + e.Message); ;
                    }
                else
                {
                    string msg = messageLoader.GetString("ErrorSavingSecKeyOutputFile");
                    throw new Exception(msg);
                    //throw new Exception("Error while trying to save the secret key ring: unable to allocate output file"); ;
                }
            }
            else
            {
                string msg = messageLoader.GetString("ErrorSavingSecKeyInvalidKeyLoc");
                throw new Exception(msg);
                //throw new Exception("Error while trying to save the public key ring: Invalid keyLocation");
            }
        }

        /// <summary>
        /// Exports a key to a location
        /// </summary>
        /// <param name="KeyID">key to save</param>
        public async Task Export(ulong KeyID)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            string strKey;
            string myUserName = string.Empty;
            string myKeyID = string.Empty;
            string myUserID = string.Empty;
            string myTime = DateTime.Now.ToString();

            try
            {
                TransportableSecretKey transSecKey = this.Find(KeyID);
                byte[] bKey = transSecKey.Generate();
                strKey = Armor.WrapPrivateKey(bKey);

                myUserName = transSecKey.PrimaryUserID.Trim();
                myKeyID = transSecKey.KeyID();
                myKeyID = transSecKey.PrimaryKey.PublicKey.KeyID.ToString("X");
            }
            catch (Exception e)
            {
                throw new Exception("Error while trying to save a private key: " + e.Message);
            }

            string fileNameToUse = myUserName + "_" + myKeyID + "_" + myTime + "_skey";
            fileNameToUse = fileNameToUse.Replace(":", "_");
            fileNameToUse = fileNameToUse.Replace(" ", "_");
            fileNameToUse = fileNameToUse.Replace("/", "_");
            fileNameToUse = fileNameToUse.Replace(".", "_");
            fileNameToUse = fileNameToUse.Replace("<", "_");
            fileNameToUse = fileNameToUse.Replace(">", "_");
            fileNameToUse += ".asc";
            fileNameToUse = fileNameToUse.Replace("__", "_");
            var savePicker = new FileSavePicker
            {
                SuggestedStartLocation = PickerLocationId.DocumentsLibrary,
                SuggestedFileName = fileNameToUse
            };

            savePicker.FileTypeChoices.Add("Armored", new List<string>() { ".asc" });
            savePicker.FileTypeChoices.Add("Text", new List<string>() { ".txt" });

            StorageFile file = await savePicker.PickSaveFileAsync();

            if (file != null)
                try
                {
                    await FileIO.WriteTextAsync(file, strKey);
                }
                catch (Exception e)
                {
                    string msg = messageLoader.GetString("ErrorSavingSecKey") + e.Message;
                    throw new Exception(msg);
                    //throw new Exception("Error while trying to save a secret key: " + e.Message); ;
                }
            else
            {
                string msg = messageLoader.GetString("ErrorSavingSecKeyOutputFile");
                throw new Exception(msg);
                //throw new Exception("Error while trying to save a secret key: unable to allocate output file"); 
            }
        }

        /// <summary>
        /// Private method to add a key with checks
        /// </summary>
        /// <param name="transSecKey">key to be added</param>
        public void AddSecretKey(TransportableSecretKey transSecKey)
        {
            if (transSecKey != null)
            {
                if (this.Find(transSecKey.PrimaryKey.PublicKey.KeyID) == null)
                {
                    this.Add(transSecKey);
                }
            }
        }

        /// <summary>
        /// Add a key to the keyring
        /// </summary>
        /// <param name="transSecKey">the key to be added</param>
        private void Add(TransportableSecretKey transSecKey)
        {
            //bIsUpdated = true;
            SecretKeys.Add(transSecKey);
        }

        /// <summary>
        /// Removes the specified key from the ring
        /// </summary>
        /// <param name="transSecKey">the key to remove</param>
        public void Delete(TransportableSecretKey transSecKey)
        {
            //bIsUpdated = true;
            SecretKeys.Remove(transSecKey);
        }

        /// <summary>
        /// Removes the specified key from the ring
        /// </summary>
        /// <param name="lKeyID">the key to remove</param>
        public void Delete(ulong lKeyID)
        {
            //bIsUpdated = true;
            SecretKeys.Remove(Find(lKeyID));
        }

        /// <summary>
        /// Finds a key 
        /// </summary>
        /// <param name="lKeyID"> the key to be found></param>
        /// <returns>the found key</returns>
        public TransportableSecretKey Find(ulong lKeyID)
        {
            var messageLoader = new Windows.ApplicationModel.Resources.ResourceLoader();
            IEnumerator ieKeys = SecretKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                TransportableSecretKey transSecKey = (TransportableSecretKey)ieKeys.Current;
                if (transSecKey.PrimaryKey.PublicKey.KeyID == lKeyID)
                {
                    return transSecKey;
                }

                IEnumerator ieSubkeys = transSecKey.SubKeys.GetEnumerator();
                while (ieSubkeys.MoveNext())
                {
                    if (!(ieSubkeys.Current is SecretKeyPacket))
                    {
                        string msg = messageLoader.GetString("NoSecKeyPacket");
                        throw new Exception(msg);
                        //throw new Exception("Expected a secret key packet, but did not find one.");
                    }


                    SecretKeyPacket secKeyPkt = (SecretKeyPacket)ieSubkeys.Current;
                    if (secKeyPkt.PublicKey.KeyID == lKeyID)
                    {
                        return transSecKey;
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Finds a list of keys doing a query on the userIDs in the ring
        /// </summary>
        /// <param name="userID">the userID to find</param>
        /// <returns>the list of keys containing such user id</returns>
        public List<TransportableSecretKey> FindSecretKeysByID(string userID)
        {
            List<TransportableSecretKey> pkList = new List<TransportableSecretKey>();
            IEnumerator ieKeys = SecretKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (!(ieKeys.Current is TransportableSecretKey))
                {
                    continue;
                }
                foreach (UserIDPacket userIDPkt in ((TransportableSecretKey)ieKeys.Current).UserIDs)
                {
                    if (userIDPkt.UserID.IndexOf(userID) >= 0)
                    {
                        pkList.Add((TransportableSecretKey)ieKeys.Current);
                    }
                }
            }
            return pkList;
        }

        /// <summary>
        /// Finds the key related to the specified fingerprint
        /// </summary>
        /// <param name="fingerprint">fingerprint</param>
        /// <returns>a key</returns>
        public TransportableSecretKey FindSecretKey(string fingerprint)
        {
            IEnumerator ieKeys = SecretKeys.GetEnumerator();
            while (ieKeys.MoveNext())
            {
                if (!(ieKeys.Current is TransportableSecretKey))
                {
                    continue;
                }
                if (((TransportableSecretKey)ieKeys.Current).PrimaryKey.PublicKey.Fingerprint.ToString() == fingerprint)
                {
                    return ((TransportableSecretKey)ieKeys.Current);
                }
            }
            return null;
        }

    }

}
